/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Interface to the RTLinux scheduler.
 * Based on rtl_sched.h from the standard RTLinux distribution
 * written by Michael Barabanov <baraban@fsmlabs.com> and
 * Victor Yodaiken <yodaiken@fsmlabs.com>.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifndef _rtl_sched_h
#define _rtl_sched_h

#include <rtl/rtl_core.h>
#include <rtl/rtl_conf.h>
#include <rtl/rtl_time.h>

#define SCHED_OTHER  0
#define SCHED_FIFO   1
#define SCHED_RR     2

#define RTL_MAX_SIGNAL    31

#define RTL_SIGNAL_NULL    0
#define RTL_SIGNAL_DELETE  1
#define RTL_SIGNAL_WAKEUP  2
#define RTL_SIGNAL_SUSPEND 3
#define RTL_SIGNAL_FIRST   RTL_SIGNAL_DELETE	/* Ck-specific */
#define RTL_SIGNAL_LAST    RTL_SIGNAL_SUSPEND	/* Ck-specific */

#ifdef PTHREAD_STACK_MIN
#undef PTHREAD_STACK_MIN
#endif
#define PTHREAD_STACK_MIN   512
#define PTHREAD_STACK_MAX   1048576
#define PTHREAD_CANCELED    ((void *)-1)

#define PTHREAD_CANCEL_ENABLE       1
#define PTHREAD_CANCEL_DISABLE      0
#define PTHREAD_CANCEL_DEFERRED     1
#define PTHREAD_CANCEL_ASYNCHRONOUS 0
#define PTHREAD_CREATE_JOINABLE     0
#define PTHREAD_CREATE_DETACHED     1

#ifndef HAVE_SYS_PTHREAD_T

/* Warning: Check/Update fixincludes if any of the following
    types or definitions are changed. */

struct rtl_cleanup_struct {

    void (*routine)(void *arg);
    void *arg;
    struct rtl_cleanup_struct *next;
};

struct rtl_thread_struct {

    ckhandle_t handle;
    struct rtl_cleanup_struct *cleanup;
};

typedef struct rtl_thread_struct RTL_THREAD_STRUCT;

typedef struct rtl_thread_struct *pthread_t;

struct rtl_sched_param {

    int sched_priority;
};

#define sched_param rtl_sched_param

enum rtl_task_states {

    RTL_THREAD_READY =1,
    RTL_THREAD_DELAYED =2,
    RTL_THREAD_DORMANT =4,
    RTL_THREAD_SUSPENDED =8,
    RTL_THREAD_ACTIVE =16,
    RTL_THREAD_ZOMBIE =32,
    RTL_THREAD_WAIT_NEEDED =64
};

typedef struct STRUCT_PTHREAD_ATTR {

    size_t stack_size;
    struct rtl_sched_param sched_param;
    int cpu;			/* Ck: unused */
    int use_fp;			/* Ck: unused */
    int detachstate;
    enum rtl_task_states initial_state;

} pthread_attr_t;

#endif /* !HAVE_SYS_PTHREAD_T */

#ifdef __cplusplus
extern "C" {
#endif

/* RTL-specific */

#define pthread_delete_np ck_pthread_delete_np
int ck_pthread_delete_np(pthread_t thread);

#define pthread_wakeup_np ck_pthread_wakeup_np
int ck_pthread_wakeup_np(pthread_t thread);

#define pthread_suspend_np ck_pthread_suspend_np
int ck_pthread_suspend_np(pthread_t thread);

#define pthread_setfp_np ck_pthread_setfp_np
int ck_pthread_setfp_np(pthread_t thread,
			int flag);

#define pthread_attr_getcpu_np ck_pthread_attr_getcpu_np
int ck_pthread_attr_getcpu_np(const pthread_attr_t *attr,
			      int *cpu);

#define pthread_attr_setcpu_np ck_pthread_attr_setcpu_np
int ck_pthread_attr_setcpu_np(pthread_attr_t *attr,
			   int cpu);

#define pthread_wait_np ck_pthread_wait_np
int ck_pthread_wait_np(void);

#define pthread_make_periodic_np ck_pthread_make_periodic_np
int ck_pthread_make_periodic_np(pthread_t pthread,
				hrtime_t start_time,
				hrtime_t period);
#define usleep ck_usleep
int ck_usleep(useconds_t interval);

#define nanosleep ck_nanosleep
int ck_nanosleep(const struct timespec *rqtp,
		 struct timespec *rmtp);

clockid_t rtl_getschedclock(void);

int rtl_setclockmode(clockid_t clock,
		     int mode,
		     hrtime_t param);

int rtl_sched_lock(pthread_t pthread);

/* POSIX */

#define pthread_cleanup_push(p_routine, p_arg) { \
	rtl_irqstate_t __flags; \
	struct rtl_cleanup_struct __cleanup; \
	__cleanup.routine = (p_routine); \
	__cleanup.arg = (p_arg); \
	rtl_no_interrupts (__flags); \
	__cleanup.next = ck_pthread_self()->cleanup; \
	ck_pthread_self()->cleanup = &__cleanup; \
	rtl_restore_interrupts (__flags);


#define pthread_cleanup_pop(execute) \
	rtl_no_interrupts (__flags); \
	ck_pthread_self()->cleanup = ck_pthread_self()->cleanup->next; \
	if (execute) \
	    __cleanup.routine(__cleanup.arg); \
	rtl_restore_interrupts (__flags); \
}

int sched_get_priority_max(int policy);

int sched_get_priority_min(int policy);

#define pthread_self ck_pthread_self
pthread_t ck_pthread_self(void);

#define pthread_equal ck_pthread_equal
static inline int ck_pthread_equal(pthread_t thread1, pthread_t thread2) {
	return thread1 == thread2;
}

#define pthread_attr_init ck_pthread_attr_init
int ck_pthread_attr_init(pthread_attr_t *attr);

#define pthread_attr_destroy ck_pthread_attr_destroy
int ck_pthread_attr_destroy(pthread_attr_t *attr);

#define pthread_attr_setstacksize ck_pthread_attr_setstacksize
int ck_pthread_attr_setstacksize(pthread_attr_t *attr,
				 size_t stacksize);

#define pthread_yield ck_pthread_yield
void ck_pthread_yield(void);

#define pthread_setschedparam ck_pthread_setschedparam
int ck_pthread_setschedparam(pthread_t thread,
			     int policy,
			     const struct rtl_sched_param *param);
 
#define pthread_getschedparam ck_pthread_getschedparam
int ck_pthread_getschedparam(pthread_t thread,
			     int *policy,
			     struct rtl_sched_param *param);

#define pthread_attr_setschedparam ck_pthread_attr_setschedparam
int ck_pthread_attr_setschedparam(pthread_attr_t *attr,
				  const struct rtl_sched_param *param);

#define pthread_attr_getschedparam ck_pthread_attr_getschedparam
int ck_pthread_attr_getschedparam(const pthread_attr_t *attr,
				  struct rtl_sched_param *param);

#define pthread_create ck_pthread_create
int ck_pthread_create(pthread_t *thread,
		      pthread_attr_t *attr,
		      void *(*start_routine)(void *),
		      void *arg);

#define pthread_exit ck_pthread_exit
void ck_pthread_exit(void *retval);

#define pthread_kill ck_pthread_kill
int ck_pthread_kill(pthread_t thread,
		    int signo);

#define pthread_setcancelstate ck_pthread_setcancelstate
int ck_pthread_setcancelstate(int state,
			      int *oldstate);

#define pthread_setcanceltype ck_pthread_setcanceltype
int ck_pthread_setcanceltype(int type,
			     int *oldtype);

#define pthread_cancel ck_pthread_cancel
int ck_pthread_cancel(pthread_t thread);

#define pthread_testcancel ck_pthread_testcancel
void ck_pthread_testcancel(void);

#define pthread_join ck_pthread_join
int ck_pthread_join(pthread_t thread,
		    void **value_ptr);

#ifdef __cplusplus
};
#endif

/* Deprecated call */
#define pthread_setperiod_np(pthread,value) ( \
    !timespec_nz(&value->it_value) ? \
	ck_pthread_make_periodic_np(pthread,HRTIME_INFINITY,0) : \
	ck_pthread_make_periodic_np(pthread,timespec_to_ns(&value->it_value), \
				 timespec_to_ns (&value->it_interval)) \
)

#endif /* !_rtl_sched_h */
