/* -*- C++ -*-
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Interface to the RTLinux target abstraction.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifndef _rtlNode_h
#define _rtlNode_h

#if defined(__GNUG__) && !defined(__OBSCAN__)
#pragma interface
#endif // __GNUG__ && !__OBSCAN__

#include "CkNode.h"
#include "rtlClock.h"
#include "rtl/rtl.h"
#include "rtl/rtl_time.h"
#include "rtl/rtl_sema.h"
#include "rtl/rtl_fifo.h"
#include "rtl/rt_ipc.h"

#ifdef IMPLEMENTING_SO
#undef IMPLEMENTING_SO
#endif // IMPLEMENTING_SO
#define IMPLEMENTING_SO IMPLEMENT_RTL
#include "sodecl.h"

#define RTLINUX_MAGIC  0x52544c58 // RTLinux magic number for nodeType

#define RTLINUX_THREAD_INODE   0
#define RTLINUX_CLOCK_INODE    1
#define RTLINUX_MUTEX_INODE    2
#define RTLINUX_SEMA4_INODE    3
#define RTLINUX_MESGQ_INODE    4
#define RTLINUX_CONDV_INODE    5
#define RTLINUX_FIFO_INODE     6

// Configuration options flags
#define RTL_WARN_MISSED_DEADLINE  0x1
#define RTL_WARN_REMOVED_SYNCH    0x2
#define RTL_WARN_PRIO_INVERSION   0x4
#define RTL_LOOSE_CALLCHECK       0x8

// RTLinux specific node flags
#define RTL_FIFO_PEND_HANDLER  CK_NODE_BIT0

#define rtlRunningNode   ((rtlNode *)CkNode::This)
#define rtlRunningThread ((rtlThread *)rtlRunningNode->getRunningThread())

class rtlThread;
class rtlClock;
class rtlMutex;
class rtlSema4;
class rtlMq;
class rtlCondVar;
class rtlFifo;
class rtlLinuxThread;

class rtlNode : public CkNode {

    friend class rtlConfig;

protected:

    // Configuration parameters

    int (*startAddr)(void);

    void (*cleanupAddr)(void);

    unsigned customOptions;

    unsigned maxFifos,
	maxMutexWaiters;

    int maxPrio,
	minPrio;

    rtlSystemTimer *sysTimer;

    rtlClock *clockRT;

    struct rtl_clock clock_realtime;

    rtlClock *clockUST;

    struct rtl_clock clock_ust;

    rtlFifo *fifoTable[RTF_NO];

    rtlInterrupt *irqTable[RTL_IRQ_MAX];

    rtlLinuxThread *linuxThread;

public:

    rtlNode(const char *nodeName,
	    CkNodeParms *nparms);

    int testOption(int option) const {
	return !!(customOptions & option);
    }

    int getMinPrio() const {
	return minPrio;
    }

    int getMaxPrio() const {
	return maxPrio;
    }

    clockid_t getUstClock() {
	return &clock_ust;
    }

    clockid_t getRealtimeClock() {
	return &clock_realtime;
    }

    hrtime_t getHrTime() const {
	return clockRT->getTime();
    }

    ckticks_t getHrTicks() const {
	return clockRT->getTicks();
    }

    hrtime_t getHrRes() const {
	return clockRT->getRes();
    }

    ckticks_t ns2ticks(hrtime_t ns) {
	// Always round the result to the next full tick.
	hrtime_t r = getHrRes();
	return (ckticks_t)((ns + r - 1) / r);
    }

    void registerIrq(unsigned irq,
		     rtlInterrupt *interrupt) {
	irqTable[irq] = interrupt;
    }

    void cleanup();

    void nameObject(CkObject *object,
		    int type);

    void unnameObject(CkObject *object);

    CkObject *findActiveObject(CkId id);

    virtual ckticks_t rrGetCredit();

    virtual void (*getRootEntry())(...);

    virtual void protoInit();

    virtual void kinit(warmInit)();

    virtual void deleteThread(CkThread *thread);

    virtual void kdoor(tick)();

    virtual CkInterrupt *createInterrupt(ckvector_t vector,
					 cklevel_t level,
					 ckisr_t *isr,
					 ckdsr_t *dsr,
					 void *cookie,
					 const char *symbolicName);
    void mustBeInit();

    void mustBeInitOrSynch();

    void pendReschedule() {
	// In fact, this method prevents the rescheduling to take place
	// immediately before the next source statement is executed, which
	// would be too soon, by clearing the CKSCHED informative flag
	// which may have been raised while altering the scheduler's state.
	// Expect the next clock tick occurring on this node to trigger the
	// rescheduling procedure.
	clrStatus(CKSCHED);
    }

    void signalLinux(int flags);

    rtlThread *findThreadByPtr(pthread_t pthread);

    rtlMutex *findMutexByPtr(pthread_mutex_t *mutex);

    rtlSema4 *findSema4ByPtr(rt_sem_t *sem);

    rtlMq *findMqByPtr(rt_mq_t *mq);

    rtlCondVar *findCondVarByPtr(pthread_cond_t *cond);

    rtlClock *findClockByPtr(clockid_t clock);

    int kdoor(threadCreate)(pthread_t *pthread,
			    const pthread_attr_t *attr,
			    void *(*startAddr)(void *arg),
			    void *startArg);

    void kdoor(threadExit)(void *retval);

    int kdoor(threadKill)(pthread_t pthread,
			  int signo);

    int kdoor(threadSetCancelState)(int state,
				    int *oldstate);

    int kdoor(threadSetCancelType)(int type,
				   int *oldtype);

    int kdoor(threadJoin)(pthread_t pthread,
			  void **value_ptr);

    void kdoor(threadYield)();

    pthread_t kdoor(threadSelf)();

    int kdoor(threadSetSchedParam)(pthread_t pthread,
				   int policy,
				   const rtl_sched_param *param);

    int kdoor(threadGetSchedParam)(pthread_t pthread,
				   int *policy,
				   rtl_sched_param *param);

    void kdoor(threadTestCancel)(void);

    void kdoor(threadAttrInit)(pthread_attr_t *attr);

    int kdoor(threadMakePeriodic)(pthread_t pthread,
				  hrtime_t start_time,
				  hrtime_t period);
    int kdoor(threadWait)();

    int kdoor(threadSuspend)(pthread_t pthread);

    int kdoor(threadWakeup)(pthread_t pthread);

    int kdoor(threadSetFp)(pthread_t pthread,
			   int flag);

    int kdoor(setClockMode)(clockid_t clock,
			    int mode,
			    hrtime_t param);

    int kdoor(setClockHandler)(clockid_t clock,
			       clock_irq_handler_t fn);

    int kdoor(unsetClockHandler)(clockid_t clock);

    int kdoor(clockGetTime)(clockid_t clock,
			    struct timespec *ts);

    int kdoor(clockSetTime)(clockid_t clock,
			    const struct timespec *ts);

    int kdoor(clockGetRes)(clockid_t clock,
			   struct timespec *ts);

    hrtime_t kdoor(clockGetHrTime)(clockid_t clock);

    time_t kdoor(clockTime)(void);

    int kdoor(usleep)(useconds_t interval);

    int kdoor(nanosleep)(const struct timespec *rqtp,
			 struct timespec *rmtp);

    int kdoor(mutexLock)(pthread_mutex_t *mutex);

    int kdoor(mutexTryLock)(pthread_mutex_t *mutex);

    int kdoor(mutexUnlock)(pthread_mutex_t *mutex);

    int kdoor(mutexInit)(pthread_mutex_t *mutex,
			 const pthread_mutexattr_t *attr);

    int kdoor(mutexDestroy)(pthread_mutex_t *mutex);

    int kdoor(mutexSetPrioCeiling)(pthread_mutex_t *mutex,
				   int prioceiling);

    int kdoor(mutexGetPrioCeiling)(pthread_mutex_t *mutex);

    int kdoor(condInit)(pthread_cond_t *cond,
			const pthread_condattr_t *attr);

    int kdoor(condDestroy)(pthread_cond_t *cond);

    int kdoor(condWait)(pthread_cond_t *cond,
			pthread_mutex_t *mutex);

    int kdoor(condSignal)(pthread_cond_t *cond);

    int kdoor(condBroadcast)(pthread_cond_t *cond);

    int kdoor(condTimedWait)(pthread_cond_t *cond,
			     pthread_mutex_t *mutex,
			     const struct timespec *abstime);

    int kdoor(schedLock)(pthread_t pthread);

    int kdoor(ipcSemInit)(rt_sem_t *sem,
			  RT_SEMTYPE type,
			  int init_val);

    int kdoor(ipcSemDestroy)(rt_sem_t *sem);

    int kdoor(ipcSemPost)(rt_sem_t *sem);

    int kdoor(ipcSemWait)(rt_sem_t *sem,
			  RTIME timeout);

    int kdoor(ipcSemTryWait)(rt_sem_t *sem);

    int kdoor(ipcTaskDelay)(RTIME duration);

    int kdoor(ipcMqInit)(rt_mq_t *mq,
			 int max_msgs,
			 int msg_size);

    int kdoor(ipcMqDestroy)(rt_mq_t *mq);

    int kdoor(ipcMqSend)(rt_mq_t *mq,
			 char *msg,
			 RT_MQ_PRIO prio,
			 RTIME wait);

    int kdoor(ipcMqReceive)(rt_mq_t *mq,
			    char *msg,
			    RTIME wait);

    int kdoor(fifoCreate)(int fifo,
			  int size);

    int kdoor(fifoDestroy)(int fifo);

    int kdoor(fifoPut)(int fifo,
		       const void *buf,
		       int count);

    int kdoor(fifoGet)(int fifo,
		       void *buf,
		       int count);

    int kdoor(fifoCreateUserHandler)(int fifo,
				     int (*handler)(unsigned fifo));

    int kdoor(fifoCreateRtHandler)(int fifo,
				   int (*handler)(unsigned fifo));

    void kdoor(hardEnableIrq)(unsigned irq);

    void kdoor(hardDisableIrq)(unsigned irq);

    int kdoor(requestGlobalIrq)(unsigned irq,
				rtl_irq_handler_t handler);

    int kdoor(freeGlobalIrq)(unsigned irq);

    int kdoor(getSoftIrq)(rtl_soft_irq_handler_t handler,
			  const char *devname);

    void kdoor(freeSoftIrq)(unsigned irq);

    void kdoor(globalPendIrq)(unsigned irq);

    int kdoor(stdSemInit)(sem_t *sem,
			  unsigned init_val);

    int kdoor(stdSemDestroy)(sem_t *sem);

    int kdoor(stdSemPost)(sem_t *sem);

    int kdoor(stdSemTimedWait)(sem_t *sem,
			       const struct timespec *abstime);

    int kdoor(stdSemWait)(sem_t *sem);

    int kdoor(stdSemTryWait)(sem_t *sem);

    unsigned kdoor(stdSemGetValue)(sem_t *sem);
};

MakeGList(rtlNode);

#endif // !_rtlNode_h
