/* -*- C++ -*-
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Interface to the RTLinux kernel synchronization objects.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifndef _rtlSynch_h
#define _rtlSynch_h

#if defined(__GNUG__) && !defined(__OBSCAN__)
#pragma interface
#endif // __GNUG__ && !__OBSCAN__

#include "CkSynch.h"
#include "rtl/rtl_sync.h"
#include "rtl/rtl_mutex.h"
#include "rtl/rt_ipc.h"

#ifdef IMPLEMENTING_SO
#undef IMPLEMENTING_SO
#endif // IMPLEMENTING_SO
#define IMPLEMENTING_SO IMPLEMENT_RTL
#include "sodecl.h"

class rtlNode;
class rtlThread;
class rtlMutex;

class rtlSynch : public CkSynchro {

public:

    rtlSynch(const char *name,
	      InsertMode mode,
	      CkStateExportObject *synchState,
	      rtlNode *node);

    rtlNode *getNode() {
	return (rtlNode *)CkSynchro::getNode();
    }

    void blockThread(rtlThread *thread,
		     ckticks_t timeout);

    void unblockThread(rtlThread *thread);

    virtual void forget(CkThread *thread);
};

enum rtlSema4States {

    rtlSema4IDLE =0,
    rtlSema4PENDED =1,
    rtlSema4POSTED =2
};

class rtlSema4 : public rtlSynch {

private:

    unsigned count;

    unsigned ceiling;

protected:

    virtual void protoInit();

    virtual void dynamicExpose();

    virtual void dynamicConfigure(StProtoInfoMessage *mbuf);

    static const char *stateArray[3];

public:

    rtlSema4(unsigned count,
	     unsigned ceiling,
	     InsertMode mode,
	     rtlNode *node);

    virtual ~rtlSema4();

    unsigned peek() {
    	return count;
    }

    int acquire(rtlThread *thread,
		ckticks_t timeout);

    int release();

    virtual const char *getTypedName();

    virtual void forget(CkThread *thread);
};

enum rtlMutexStates {

    rtlMutexFREE =0,
    rtlMutexLOCKED =1,
    rtlMutexPENDED =2
};

class rtlMutex : public rtlSynch {

private:

    rtlThread *owner;

    int prioCeiling;

protected:

    virtual void protoInit();

    virtual void dynamicExpose();

    virtual void dynamicConfigure(StProtoInfoMessage *mbuf);

    static const char *stateArray[3];

public:

    rtlMutex(int prioCeiling,
	     rtlNode *node);

    virtual ~rtlMutex();

    int lock(rtlThread *thread);

    int tryLock(rtlThread *thread);

    int unlock();

    int isLocked() const {
	return synchState->getState() == rtlMutexLOCKED;
    }

    int getPrioCeiling() const {
	return prioCeiling;
    }

    void setPrioCeiling (int ceiling) {
	prioCeiling = ceiling;
    }

    virtual const char *getTypedName();

    virtual void forget(CkThread *thread);
};

enum rtlMqStates {

    rtlMqEMPTY =0,
    rtlMqPENDED =1,
    rtlMqPOSTED =2,
    rtlMqFULL =3
};

struct rtlMessage : public Link {

    char *src;
    char *copy;
    RT_MQ_PRIO prio;
    ITime time;

    rtlMessage(char *_ptr, int _size, RT_MQ_PRIO _prio) {
	src = _ptr;	// Original source pointer
	prio = _prio;
	time = SxClock;

	if (_size > 0 && _ptr)
	    {
	    copy = new char[_size];
	    memcpy(copy,_ptr,_size);
	    }
	else
	    copy = NULL;
    }

    ~rtlMessage() {

	if (copy)
	    delete[] copy;
    }

    void *operator new(size_t sz);

    void operator delete(void *ptr);
};

MakeList(rtlMessage);

class rtlMq : public rtlSynch {

    friend struct rtlMessage;

protected:

    int maxMsg;

    int msgSize;

    int sendBlockCount;

    char *copyBuf;

    rtlMessageList storage;

    virtual void protoInit();

    virtual void dynamicExpose();

    virtual void dynamicConfigure(StProtoInfoMessage *mbuf);

    static const char *stateArray[4];

    static rtlMessageList messageTank;
	
public:

    rtlMq(int maxMsg,
	  int msgSize,
	  rtlNode *node);

    virtual ~rtlMq();

    unsigned getMsgCount() const {
	return storage.getCount();
    }

    int getSendBlockCnt() const {
	return sendBlockCount;
    }

    int getRecvBlockCnt() const {
	return getCount() - sendBlockCount;
    }

    int getMaxMsg() const {
	return maxMsg;
    }

    int getMsgSize() const {
	return msgSize;
    }

    int send(rtlThread *sender,
	     char *ptr,
	     RT_MQ_PRIO prio,
	     ckticks_t timeout);

    int receive(rtlThread *receiver,
		ckticks_t timeout);

    rtlMessage *peek() const {
	return storage.first();
    }

    virtual const char *getTypedName();

    virtual void forget(CkThread *thread);
};

enum rtlCondVarStates {

    rtlCondVarIDLE =0,
    rtlCondVarPENDED =1,
    rtlCondVarSIG =2,
    rtlCondVarBRD =3,
};

class rtlCondVar : public rtlSynch {

protected:

    virtual void protoInit();

    virtual void dynamicExpose();

    virtual void dynamicConfigure(StProtoInfoMessage *mbuf);

    static const char *stateArray[4];

public:

    rtlCondVar(rtlNode *node);

    virtual ~rtlCondVar();

    void wait(rtlThread *thread,
	      ckticks_t timeout);

    int signal();

    int broadcast();

    virtual const char *getTypedName();

    virtual void forget(CkThread *thread);
};

MakeGList(rtlCondVar);

#endif // !_rtlSynch_h
