/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: The RTLinux API this model exports to applications.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#include <config.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include "CkManager.h"
#include "rtlNode.h"
#include "rtlTrace.h"
#include <rtl/rtl_sched.h>
#include <rtl/rtl_mutex.h>
#include <rtl/rtl_sync.h>
#include <rtl/rtl_sema.h>
#include <rtl/rtl_fifo.h>
#include <rtl/posix/time.h>

int pthread_create (pthread_t *pthread,
		    pthread_attr_t *attr,
		    void *(*start_routine)(void *),
		    void *arg)
{
    int err;

    CkInAddr p1("thread",pthread);
    CkInAddr p2("attr",attr);
    CkInAddr p3("start_routine",(void *)start_routine);
    CkInAddr p4("arg",arg);

    CkTracePoint t(0,&err,&p1,&p2,&p3,&p4,NULL);

    if (!pthread || !TEST_PARAM(pthread,sizeof(*pthread),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_create(thread=%p): invalid memory",
			       pthread);

    if (attr && !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_create(attr=%p): invalid memory",
			       attr);

    if (!start_routine || !TEST_PARAM(start_routine,1,MEMADDR_TEXT))
	CkManager::This->fatal("pthread_create(start_routine=%p): invalid routine",
			       start_routine);

    if (attr && attr->cpu < 0)
	CkManager::This->fatal("pthread_create(attr->cpu=%d): invalid CPU #",
			       attr->cpu);

    err = rtlRunningNode->kdoor(threadCreate)(pthread,
					      attr,
					      start_routine,
					      arg);
    t.kdoor(emit)();

    return err;
}

void pthread_exit (void *retval)

{
    CkInAddr p1("retval",retval);

    CkTracePoint t(1,NULL,&p1,NULL);

    // Must emit now since we won't return from
    // rtlNode::threadExit().

    t.kdoor(emit)();

    rtlRunningNode->kdoor(threadExit)(retval);
}

int pthread_kill (pthread_t pthread,
		  int signo)
{
    int err;

    CkInAddr p1("thread",pthread);
    CkInInt p2("signo",signo);

    CkTracePoint t(2,&err,&p1,&p2,NULL);

    if (!pthread || !TEST_PARAM(pthread,sizeof(*pthread),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_kill(thread=%p): invalid memory",
			       pthread);

    err = rtlRunningNode->kdoor(threadKill)(pthread,signo);

    t.kdoor(emit)();

    return err;
}

int pthread_setcancelstate (int state,
			    int *oldstate)
{
    int err;

    CkInInt p1("state",state);
    CkOutInt p2("oldstate",oldstate);

    CkTracePoint t(3,&err,&p1,&p2,NULL);

    // oldstate == NULL is allowed

    if (oldstate && !TEST_PARAM(oldstate,sizeof(*oldstate),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_setcancelstate(oldstate=%p): invalid memory",
			       oldstate);

    err = rtlRunningNode->kdoor(threadSetCancelState)(state,oldstate);

    t.kdoor(emit)();

    return err;
}

int pthread_setcanceltype (int type,
			   int *oldtype)
{
    int err;

    CkInInt p1("type",type);
    CkOutInt p2("oldtype",oldtype);

    CkTracePoint t(4,&err,&p1,&p2,NULL);

    // oldtype == NULL is allowed

    if (oldtype && !TEST_PARAM(oldtype,sizeof(*oldtype),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_setcanceltype(oldtype=%p): invalid memory",
			       oldtype);

    err = rtlRunningNode->kdoor(threadSetCancelType)(type,oldtype);

    t.kdoor(emit)();

    return err;
}

int pthread_cancel (pthread_t pthread)

{
    int err;

    CkInAddr p1("thread",pthread);

    CkTracePoint t(5,&err,&p1,NULL);

    if (!pthread || !TEST_PARAM(pthread,sizeof(*pthread),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_cancel(thread=%p): invalid memory",
			       pthread);

    err = rtlRunningNode->kdoor(threadKill)(pthread,RTL_SIGNAL_DELETE);

    t.kdoor(emit)();

    return err;
}

void pthread_testcancel (void)

{
    CkTracePoint t(6,NULL);

    rtlRunningNode->kdoor(threadTestCancel)();

    t.kdoor(emit)();
}

int pthread_join (pthread_t pthread,
		  void **value_ptr)
{
    int err;

    CkInAddr p1("thread",pthread);
    CkOutAddr p2("value_ptr",value_ptr);

    CkTracePoint t(7,&err,&p1,&p2,NULL);

    err = rtlRunningNode->kdoor(threadJoin)(pthread,value_ptr);

    t.kdoor(emit)();

    return err;
}

void pthread_yield (void)

{
    CkTracePoint t(8,NULL);

    rtlRunningNode->kdoor(threadYield)();

    t.kdoor(emit)();
}

pthread_t pthread_self (void)

{
    pthread_t self;

    CkRetAddr rv((void **)&self);

    CkTracePoint t(9,NULL,&rv,NULL);

    self = rtlRunningNode->kdoor(threadSelf)();
    
    t.kdoor(emit)();

    return self;
}

int pthread_setschedparam (pthread_t pthread,
			   int policy,
			   const rtl_sched_param *param)
{
    int err;

    CkInAddr p1("thread",pthread);
    CkInInt p2("policy",policy);
    CkInAddr p3("param",param);

    CkTracePoint t(10,&err,&p1,&p2,&p3,NULL);

    if (!param || !TEST_PARAM(param,sizeof(*param),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_setschedparam(param=%p): invalid memory",
			       param);

    err = rtlRunningNode->kdoor(threadSetSchedParam)(pthread,policy,param);

    t.kdoor(emit)();

    return err;
}

int pthread_getschedparam (pthread_t pthread,
			   int *policy,
			   rtl_sched_param *param)
{
    int err;

    CkInAddr p1("thread",pthread);
    CkOutInt p2("policy",policy);
    CkInAddr p3("param",param);

    CkTracePoint t(11,&err,&p1,&p2,&p3,NULL);

    if (!param || !TEST_PARAM(param,sizeof(*param),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_getschedparam(param=%p): invalid memory",
			       param);

    if (!policy || !TEST_PARAM(policy,sizeof(*policy),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_getschedparam(policy=%p): invalid memory",
			       policy);

    err = rtlRunningNode->kdoor(threadGetSchedParam)(pthread,policy,param);

    t.kdoor(emit)();

    return err;
}

int pthread_attr_init (pthread_attr_t *attr)

{
    int err = 0;

    CkInAddr p1("attr",attr);

    CkTracePoint t(12,&err,&p1,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_attr_init(attr=%p): invalid memory",
			       attr);

    rtlRunningNode->kdoor(threadAttrInit)(attr);

    t.kdoor(emit)();

    return err;
}

int pthread_attr_getschedparam (const pthread_attr_t *attr,
				rtl_sched_param *param)
{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkInAddr p2("param",param);

    CkTracePoint t(13,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_attr_getschedparam(attr=%p): invalid memory",
			       attr);

    if (!param || !TEST_PARAM(param,sizeof(*param),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_attr_getschedparam(param=%p): invalid memory",
			       param);

    *param = attr->sched_param;

    t.kdoor(emit)();

    return err;
}

int pthread_attr_setschedparam (pthread_attr_t *attr,
				const rtl_sched_param *param)
{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkInAddr p2("param",param);

    CkTracePoint t(14,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_attr_setschedparam(attr=%p): invalid memory",
			       attr);

    if (!param || !TEST_PARAM(param,sizeof(*param),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_attr_setschedparam(param=%p): invalid memory",
			       param);

    if (param->sched_priority >= rtlRunningNode->getMinPrio() &&
	param->sched_priority <= rtlRunningNode->getMaxPrio())
	attr->sched_param = *param;
    else
	err = EINVAL;

    t.kdoor(emit)();

    return err;
}

int pthread_attr_getdetachstate (const pthread_attr_t *attr,
				 int *detachstate)
{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkOutInt p2("detachstate",detachstate);

    CkTracePoint t(15,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_attr_getdetachstate(attr=%p): invalid memory",
			       attr);

    if (!detachstate || !TEST_PARAM(detachstate,sizeof(*detachstate),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_attr_getdetachstate(detachstate=%p): invalid memory",
			       detachstate);

    *detachstate = attr->detachstate;

    t.kdoor(emit)();

    return err;
}

int pthread_attr_setdetachstate (pthread_attr_t *attr,
				 int detachstate)
{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkInInt p2("detachstate",detachstate);

    CkTracePoint t(16,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_attr_setdetachstate(attr=%p): invalid memory",
			       attr);

    if (detachstate != PTHREAD_CREATE_JOINABLE &&
	detachstate != PTHREAD_CREATE_DETACHED)
	return EINVAL;

    attr->detachstate = detachstate;

    t.kdoor(emit)();

    return err;
}

int pthread_attr_setstacksize (pthread_attr_t *attr,
			       size_t stack_size)
{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkInInt p2("stack_size",stack_size);

    CkTracePoint t(17,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_attr_setschedparam(attr=%p): invalid memory",
			       attr);

    // Note: the simulator will silently enforce an internal bare minimum
    // of RTLTHREAD_STACKSIZE (8k) for each RTLinux threads it creates.

    if (stack_size > 0)
	attr->stack_size = stack_size;
    else
	err = EINVAL;

    t.kdoor(emit)();

    return err;
}

int pthread_attr_destroy (pthread_attr_t *attr)

{
    int err = 0;

    CkInAddr p1("attr",attr);

    CkTracePoint t(18,&err,&p1,NULL);

    // No-op

    t.kdoor(emit)();

    return err;
}

int pthread_make_periodic_np (pthread_t pthread,
			      hrtime_t start_time,
			      hrtime_t period)
{
    int err;

    CkInAddr p1("thread",pthread);
    CkInQuad p2("start_time",start_time);
    CkInQuad p3("period",period);

    CkTracePoint t(19,&err,&p1,&p2,&p3,NULL);

    if (!pthread || !TEST_PARAM(pthread,sizeof(*pthread),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_make_periodic_np(thread=%p): invalid memory",
			       pthread);

    // RTLinux doesn't check for parameter sanity regarding
    // start_time and period -- but Ck must do it in order to
    // report problems as soon as possible.

    if (start_time < 0)
	CkManager::This->fatal("pthread_make_periodic_np(start_time=%lld): invalid start time",
			       (gnuquad_t)start_time);

    if (period < 0)
	CkManager::This->fatal("pthread_make_periodic_np(period=%lld): invalid period",
			       (gnuquad_t)period);

    err = rtlRunningNode->kdoor(threadMakePeriodic)(pthread,
						    start_time,
						    period);
    t.kdoor(emit)();

    return err;
}

int pthread_wait_np (void)

{
    int err;

    CkTracePoint t(20,&err,NULL);

    err = rtlRunningNode->kdoor(threadWait)();

    t.kdoor(emit)();

    return err;
}

int pthread_suspend_np (pthread_t pthread)

{
    int err;

    CkInAddr p1("thread",pthread);

    CkTracePoint t(21,&err,&p1,NULL);

    if (!pthread || !TEST_PARAM(pthread,sizeof(*pthread),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_suspend_np(thread=%p): invalid memory",
			       pthread);

    err = rtlRunningNode->kdoor(threadSuspend)(pthread);

    t.kdoor(emit)();

    return err;
}

int pthread_wakeup_np (pthread_t pthread)

{
    int err;

    CkInAddr p1("thread",pthread);

    CkTracePoint t(22,&err,&p1,NULL);

    if (!pthread || !TEST_PARAM(pthread,sizeof(*pthread),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_wakeup_np(thread=%p): invalid memory",
			       pthread);

    err = rtlRunningNode->kdoor(threadWakeup)(pthread);

    t.kdoor(emit)();

    return err;
}

int pthread_delete_np (pthread_t pthread)

{
    int err;

    CkInAddr p1("thread",pthread);

    CkTracePoint t(23,&err,&p1,NULL);

    if (!pthread || !TEST_PARAM(pthread,sizeof(*pthread),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_delete_np(thread=%p): invalid memory",
			       pthread);

    err = rtlRunningNode->kdoor(threadKill)(pthread,RTL_SIGNAL_DELETE);

    if (!err)
	err = rtlRunningNode->kdoor(threadJoin)(pthread,NULL);

    t.kdoor(emit)();

    return err;
}

int pthread_setfp_np (pthread_t pthread,
		      int flag)
{
    int err = 0;

    CkInAddr p1("thread",pthread);
    CkInInt p2("flag",flag);

    CkTracePoint t(24,&err,&p1,&p2,NULL);

    if (!pthread || !TEST_PARAM(pthread,sizeof(*pthread),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_delete_np(thread=%p): invalid memory",
			       pthread);

    err = rtlRunningNode->kdoor(threadSetFp)(pthread,flag);

    t.kdoor(emit)();

    return err;
}

int pthread_attr_getcpu_np (const pthread_attr_t *attr,
			    int *pcpu)
{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkOutInt p2("cpu",pcpu);

    CkTracePoint t(25,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_attr_getcpu_np(attr=%p): invalid memory",
			       attr);

    if (!pcpu || !TEST_PARAM(pcpu,sizeof(*pcpu),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_attr_getcpu_np(pcpu=%p): invalid memory",
			       pcpu);

    *pcpu = attr->cpu;

    t.kdoor(emit)();

    return err;
}

int pthread_attr_setcpu_np (pthread_attr_t *attr,
			    int cpu)
{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkInInt p2("cpu",cpu);

    CkTracePoint t(26,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_attr_getcpu_np(attr=%p): invalid memory",
			       attr);

    if (cpu < 0)
	CkManager::This->fatal("pthread_attr_setcpu_np(cpu=%d): invalid CPU #",
			       cpu);

    // Note: Ck does not currently support SMP nodes.
    // Pretend that any number of CPUs exist in the system.

    attr->cpu = cpu;

    t.kdoor(emit)();

    return err;
}

int pthread_mutexattr_init (pthread_mutexattr_t *attr)

{
    int err = 0;

    CkInAddr p1("attr",attr);

    CkTracePoint t(27,&err,&p1,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_mutexattr_init(attr=%p): invalid memory",
			       attr);

    attr->type = PTHREAD_MUTEX_DEFAULT;
    attr->pshared = PTHREAD_PROCESS_SHARED;
    attr->protocol = PTHREAD_PRIO_NONE;
    attr->prioceiling = -1;

    t.kdoor(emit)();

    return err;
}

int pthread_mutexattr_settype (pthread_mutexattr_t *attr, int type)

{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkInInt p2("type",type);

    CkTracePoint t(28,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_mutexattr_settype(attr=%p): invalid memory",
			       attr);

    if (type != PTHREAD_MUTEX_DEFAULT)
	// Behave strictly like RTLinux does.
	err = EINVAL;
    else
	attr->type = PTHREAD_MUTEX_DEFAULT;

    t.kdoor(emit)();

    return err;
}

int pthread_mutexattr_setpshared (pthread_mutexattr_t *attr, int pshared)

{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkInInt p2("pshared",pshared);

    CkTracePoint t(29,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_mutexattr_setpshared(attr=%p): invalid memory",
			       attr);

    if (pshared != PTHREAD_PROCESS_SHARED &&
	pshared != PTHREAD_PROCESS_PRIVATE)
	err = EINVAL;
    else
	attr->pshared = pshared;

    t.kdoor(emit)();

    return err;
}

int pthread_mutexattr_gettype (const pthread_mutexattr_t *attr, int *ptype)

{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkOutInt p2("type",ptype);

    CkTracePoint t(30,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_mutexattr_gettype(attr=%p): invalid memory",
			       attr);

    if (!ptype || !TEST_PARAM(ptype,sizeof(*ptype),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_mutexattr_gettype(ptype=%p): invalid memory",
			       ptype);

    *ptype = attr->type;

    t.kdoor(emit)();

    return err;
}

int pthread_mutexattr_getpshared (const pthread_mutexattr_t *attr, int *ppshared)

{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkOutInt p2("pshared",ppshared);

    CkTracePoint t(31,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_mutexattr_getpshared(attr=%p): invalid memory",
			       attr);

    if (!ppshared || !TEST_PARAM(ppshared,sizeof(*ppshared),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_mutexattr_getpshared(ppshared=%p): invalid memory",
			       ppshared);

    *ppshared = attr->pshared;

    t.kdoor(emit)();

    return err;
}

int pthread_mutexattr_destroy (pthread_mutexattr_t *attr)

{
    int err = 0;

    CkInAddr p1("attr",attr);

    CkTracePoint t(32,&err,&p1,NULL);

    // No-op

    t.kdoor(emit)();

    return err;
}

int pthread_mutexattr_setprotocol (pthread_mutexattr_t *attr,
				   int protocol)
{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkInInt p2("protocol",protocol);

    CkTracePoint t(33,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_mutexattr_setprotocol(attr=%p): invalid memory",
			       attr);

    if (protocol != PTHREAD_PRIO_PROTECT &&
	protocol != PTHREAD_PRIO_NONE)
	err = EOPNOTSUPP;
    else
	attr->protocol = protocol;

    t.kdoor(emit)();

    return err;
}

int pthread_mutexattr_setprioceiling (pthread_mutexattr_t *attr,
				      int prioceiling)
{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkInInt p2("prioceiling",prioceiling);

    CkTracePoint t(34,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_mutexattr_setprioceiling(attr=%p): invalid memory",
			       attr);

    attr->prioceiling = prioceiling;

    t.kdoor(emit)();

    return err;
}

int pthread_mutexattr_getprotocol (const pthread_mutexattr_t *attr,
				   int *protocol)
{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkOutInt p2("protocol",protocol);

    CkTracePoint t(35,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_mutexattr_getprotocol(attr=%p): invalid memory",
			       attr);

    if (!protocol || !TEST_PARAM(protocol,sizeof(*protocol),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_mutexattr_getprotocol(ptype=%p): invalid memory",
			       protocol);

    *protocol = attr->protocol;

    t.kdoor(emit)();

    return err;
}

int pthread_mutexattr_getprioceiling (const pthread_mutexattr_t *attr,
				      int *prioceiling)
{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkOutInt p2("prioceiling",prioceiling);

    CkTracePoint t(36,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_mutexattr_getprioceiling(attr=%p): invalid memory",
			       attr);

    if (!prioceiling || !TEST_PARAM(prioceiling,sizeof(*prioceiling),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_mutexattr_getprioceiling(ptype=%p): invalid memory",
			       prioceiling);

    *prioceiling = attr->prioceiling;

    t.kdoor(emit)();

    return err;
}

int pthread_mutex_lock (pthread_mutex_t *mutex)

{
    int err = 0;

    CkInAddr p1("mutex",mutex);

    CkTracePoint t(37,&err,&p1,NULL);

    if (!mutex || !TEST_PARAM(mutex,sizeof(*mutex),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_mutex_lock(mutex=%p): invalid memory",
			       mutex);

    err = rtlRunningNode->kdoor(mutexLock)(mutex);

    t.kdoor(emit)();

    return err;
}

int pthread_mutex_trylock (pthread_mutex_t *mutex)

{
    int err = 0;

    CkInAddr p1("mutex",mutex);

    CkTracePoint t(38,&err,&p1,NULL);

    if (!mutex || !TEST_PARAM(mutex,sizeof(*mutex),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_mutex_trylock(mutex=%p): invalid memory",
			       mutex);

    err = rtlRunningNode->kdoor(mutexTryLock)(mutex);

    t.kdoor(emit)();

    return err;
}

int pthread_mutex_unlock (pthread_mutex_t *mutex)

{
    int err = 0;

    CkInAddr p1("mutex",mutex);

    CkTracePoint t(39,&err,&p1,NULL);

    if (!mutex || !TEST_PARAM(mutex,sizeof(*mutex),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_mutex_unlock(mutex=%p): invalid memory",
			       mutex);

    err = rtlRunningNode->kdoor(mutexUnlock)(mutex);

    t.kdoor(emit)();

    return err;
}

int pthread_mutex_init (pthread_mutex_t *mutex,
			const pthread_mutexattr_t *attr)
{
    int err = 0;

    CkInAddr p1("mutex",mutex);
    CkInAddr p2("attr",attr);

    CkTracePoint t(40,&err,&p1,&p2,NULL);

    if (!mutex || !TEST_PARAM(mutex,sizeof(*mutex),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_mutex_init(mutex=%p): invalid memory",
			       mutex);

    // "attr" may be NULL.

    if (attr && !TEST_PARAM(attr,sizeof(*attr),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_mutex_init(attr=%p): invalid memory",
			       attr);

    err = rtlRunningNode->kdoor(mutexInit)(mutex,attr);

    t.kdoor(emit)();

    return err;
}

int pthread_mutex_destroy (pthread_mutex_t *mutex)

{
    int err = 0;

    CkInAddr p1("mutex",mutex);

    CkTracePoint t(41,&err,&p1,NULL);

    if (!mutex || !TEST_PARAM(mutex,sizeof(*mutex),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_mutex_destroy(mutex=%p): invalid memory",
			       mutex);

    err = rtlRunningNode->kdoor(mutexDestroy)(mutex);

    t.kdoor(emit)();

    return err;
}

int pthread_mutex_setprioceiling (pthread_mutex_t *mutex,
				  int prioceiling,
				  int *oldceiling)
{
    int err = 0;

    CkInAddr p1("mutex",mutex);
    CkInInt p2("prioceiling",prioceiling);
    CkOutInt p3("oldceiling",oldceiling);

    CkTracePoint t(42,&err,&p1,&p2,&p3,NULL);

    if (!mutex || !TEST_PARAM(mutex,sizeof(*mutex),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_mutex_setprioceiling(mutex=%p): invalid memory",
			       mutex);

    if (!oldceiling || !TEST_PARAM(oldceiling,sizeof(*oldceiling),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_mutex_setprioceiling(oldceiling=%p): invalid memory",
			       oldceiling);

    *oldceiling = rtlRunningNode->kdoor(mutexSetPrioCeiling)(mutex,prioceiling);

    return err;
}

int pthread_mutex_getprioceiling (pthread_mutex_t *mutex,
				  int *prioceiling)
{
    int err = 0;

    CkInAddr p1("mutex",mutex);
    CkOutInt p2("prioceiling",prioceiling);

    CkTracePoint t(43,&err,&p1,&p2,NULL);

    if (!mutex || !TEST_PARAM(mutex,sizeof(*mutex),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_mutex_getprioceiling(mutex=%p): invalid memory",
			       mutex);

    if (!prioceiling || !TEST_PARAM(prioceiling,sizeof(*prioceiling),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_mutex_getprioceiling(prioceiling=%p): invalid memory",
			       prioceiling);

    *prioceiling = rtlRunningNode->kdoor(mutexGetPrioCeiling)(mutex);

    return err;
}

int pthread_condattr_init (pthread_condattr_t *attr)

{
    int err = 0;

    CkInAddr p1("attr",attr);

    CkTracePoint t(44,&err,&p1,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_condattr_init(attr=%p): invalid memory",
			       attr);

    attr->pshared = PTHREAD_PROCESS_SHARED;

    t.kdoor(emit)();

    return err;
}

int pthread_condattr_setpshared (pthread_condattr_t *attr, int pshared)

{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkInInt p2("pshared",pshared);

    CkTracePoint t(45,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_condattr_setpshared(attr=%p): invalid memory",
			       attr);

    if (pshared != PTHREAD_PROCESS_SHARED &&
	pshared != PTHREAD_PROCESS_PRIVATE)
	err = EINVAL;
    else
	attr->pshared = pshared;

    t.kdoor(emit)();

    return err;
}

int pthread_condattr_getpshared (const pthread_condattr_t *attr, int *ppshared)

{
    int err = 0;

    CkInAddr p1("attr",attr);
    CkOutInt p2("pshared",ppshared);

    CkTracePoint t(46,&err,&p1,&p2,NULL);

    if (!attr || !TEST_PARAM(attr,sizeof(*attr),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_condattr_getpshared(attr=%p): invalid memory",
			       attr);

    if (!ppshared || !TEST_PARAM(ppshared,sizeof(*ppshared),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_condattr_getpshared(ppshared=%p): invalid memory",
			       ppshared);

    *ppshared = attr->pshared;

    t.kdoor(emit)();

    return err;
}

int pthread_condattr_destroy (pthread_condattr_t *attr)

{
    int err = 0;

    CkInAddr p1("attr",attr);

    CkTracePoint t(47,&err,&p1,NULL);

    // No-op

    t.kdoor(emit)();

    return err;
}

int pthread_cond_init (pthread_cond_t *cond,
		       const pthread_condattr_t *attr)
{
    int err;

    CkInAddr p1("cond",cond);
    CkInAddr p2("attr",attr);

    CkTracePoint t(48,&err,&p1,&p2,NULL);

    if (!cond || !TEST_PARAM(cond,sizeof(*cond),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_cond_init(cond=%p): invalid memory",
			       cond);

    // "attr" may be NULL.

    if (attr && !TEST_PARAM(attr,sizeof(*attr),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_cond_init(attr=%p): invalid memory",
			       attr);

    err = rtlRunningNode->kdoor(condInit)(cond,attr);

    t.kdoor(emit)();

    return err;
}

int pthread_cond_destroy (pthread_cond_t *cond)

{
    int err;

    CkInAddr p1("cond",cond);

    CkTracePoint t(49,&err,&p1,NULL);

    if (!cond || !TEST_PARAM(cond,sizeof(*cond),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_cond_destroy(cond=%p): invalid memory",
			       cond);

    err = rtlRunningNode->kdoor(condDestroy)(cond);

    t.kdoor(emit)();

    return err;
}

int pthread_cond_wait (pthread_cond_t *cond,
		       pthread_mutex_t *mutex)
{
    int err;

    CkInAddr p1("cond",cond);
    CkInAddr p2("mutex",mutex);

    CkTracePoint t(50,&err,&p1,&p2,NULL);

    if (!cond || !TEST_PARAM(cond,sizeof(*cond),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_cond_wait(cond=%p): invalid memory",
			       cond);

    if (!mutex || !TEST_PARAM(mutex,sizeof(*mutex),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_cond_wait(mutex=%p): invalid memory",
			       mutex);

    err = rtlRunningNode->kdoor(condWait)(cond,mutex);

    t.kdoor(emit)();

    return err;
}

int pthread_cond_broadcast (pthread_cond_t *cond)

{
    int err;

    CkInAddr p1("cond",cond);

    CkTracePoint t(51,&err,&p1,NULL);

    if (!cond || !TEST_PARAM(cond,sizeof(*cond),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_cond_broadcast(cond=%p): invalid memory",
			       cond);

    err = rtlRunningNode->kdoor(condBroadcast)(cond);

    t.kdoor(emit)();

    return err;
}

int pthread_cond_signal (pthread_cond_t *cond)

{
    int err;

    CkInAddr p1("cond",cond);

    CkTracePoint t(52,&err,&p1,NULL);

    if (!cond || !TEST_PARAM(cond,sizeof(*cond),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_cond_signal(cond=%p): invalid memory",
			       cond);

    err = rtlRunningNode->kdoor(condSignal)(cond);

    t.kdoor(emit)();

    return err;
}

int pthread_cond_timedwait (pthread_cond_t *cond,
			    pthread_mutex_t *mutex,
			    const struct timespec *abstime)
{
    int err;

    CkInAddr p1("cond",cond);
    CkInAddr p2("mutex",mutex);
    CkInAddr p3("abstime",abstime);

    CkTracePoint t(53,&err,&p1,&p2,&p3,NULL);

    if (!cond || !TEST_PARAM(cond,sizeof(*cond),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_cond_timedwait(cond=%p): invalid memory",
			       cond);

    if (!mutex || !TEST_PARAM(mutex,sizeof(*mutex),MEMADDR_WRITABLE))
	CkManager::This->fatal("pthread_cond_timedwait(mutex=%p): invalid memory",
			       mutex);

    if (!abstime || !TEST_PARAM(abstime,sizeof(*abstime),MEMADDR_READABLE))
	CkManager::This->fatal("pthread_cond_timedwait(abstime=%p): invalid memory",
			       abstime);

    err = rtlRunningNode->kdoor(condTimedWait)(cond,mutex,abstime);

    t.kdoor(emit)();

    return err;
}

int sched_get_priority_min (int policy)

{
    int minPrio;

    CkInInt p1("policy",policy);
    CkRetInt rv(&minPrio);

    CkTracePoint t(54,NULL,&p1,&rv,NULL);

    // "policy" is currently ignored by RTLinux

    minPrio = rtlRunningNode->getMinPrio();

    t.kdoor(emit)();

    return minPrio;
}

int sched_get_priority_max (int policy)

{
    int maxPrio;

    CkInInt p1("policy",policy);
    CkRetInt rv(&maxPrio);

    CkTracePoint t(55,NULL,&p1,&rv,NULL);

    // "policy" is currently ignored by RTLinux

    maxPrio = rtlRunningNode->getMaxPrio();

    t.kdoor(emit)();

    return maxPrio;
}

void rtl_schedule (void)

{
    CkTracePoint t(56,NULL,NULL);

    rtlRunningNode->resched();

    t.kdoor(emit)();
}

int rtl_sched_lock (pthread_t pthread)

{
    int err;

    CkInAddr p1("thread",pthread);

    CkTracePoint t(57,&err,&p1,NULL);

    err = rtlRunningNode->kdoor(schedLock)(pthread);

    t.kdoor(emit)();

    return err;
}

clockid_t rtl_getschedclock (void)

{
    clockid_t clock;

    CkRetAddr rv((void **)&clock);

    CkTracePoint t(58,NULL,&rv,NULL);
    
    clock = rtlRunningNode->getRealtimeClock();

    t.kdoor(emit)();

    return clock;
}

int rtl_setclockmode (clockid_t clock,
		      int mode,
		      hrtime_t param)
{
    int err;

    CkInAddr p1("clock",clock);
    CkInInt p2("mode",mode);
    CkInQuad p3("param",param);

    CkTracePoint t(59,NULL,&p1,&p2,&p3,NULL);
    
    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_READABLE))
	CkManager::This->fatal("rtl_setclockmode(clock=%p): invalid memory",
			       clock);

    err = rtlRunningNode->kdoor(setClockMode)(clock,mode,param);

    t.kdoor(emit)();

    return err;
}

int rtl_setclockhandler (clockid_t clock,
			 clock_irq_handler_t fn)
{
    int err;

    CkInAddr p1("clock",clock);
    CkInAddr p2("fn",(void *)fn);

    CkTracePoint t(60,NULL,&p1,&p2,NULL);
    
    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_READABLE))
	CkManager::This->fatal("rtl_setclockhandler(clock=%p): invalid memory",
			       clock);

    if (!fn || !TEST_PARAM(fn,1,MEMADDR_TEXT))
	CkManager::This->fatal("rtl_setclockhandler(fn=%p): invalid handler",
			       fn);

    err = rtlRunningNode->kdoor(setClockHandler)(clock,fn);

    t.kdoor(emit)();

    return err;
}

int rtl_unsetclockhandler (clockid_t clock)

{
    int err;

    CkInAddr p1("clock",clock);

    CkTracePoint t(61,NULL,&p1,NULL);
    
    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_READABLE))
	CkManager::This->fatal("rtl_unsetclockhandler(clock=%p): invalid memory",
			       clock);

    err = rtlRunningNode->kdoor(unsetClockHandler)(clock);

    t.kdoor(emit)();

    return err;
}

clockid_t rtl_getbestclock (unsigned cpu)

{
    clockid_t clock;

    CkInInt p1("cpu",cpu);	// Ck: Ignored (UP targets only).
    CkRetAddr rv((void **)&clock);

    CkTracePoint t(62,NULL,&p1,&rv,NULL);
    
    clock = rtlRunningNode->getRealtimeClock();

    t.kdoor(emit)();

    return clock;
}

int clock_gettime (clockid_t clock,
		   struct timespec *ts)
{
    int err;

    CkInAddr p1("clock",clock);
    CkInAddr p2("ts",ts);

    CkTracePoint t(63,&err,&p1,&p2,NULL);
    
    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_READABLE))
	CkManager::This->fatal("clock_gettime(clock=%p): invalid memory",
			       clock);

    if (!ts || !TEST_PARAM(ts,sizeof(*ts),MEMADDR_WRITABLE))
	CkManager::This->fatal("clock_gettime(ts=%p): invalid memory",
			       ts);

    err = rtlRunningNode->kdoor(clockGetTime)(clock,ts);

    t.kdoor(emit)();

    return err;
}

int clock_settime (clockid_t clock,
		   const struct timespec *ts)
{
    int err;

    CkInAddr p1("clock",clock);
    CkInAddr p2("ts",ts);

    CkTracePoint t(64,&err,&p1,&p2,NULL);
    
    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_READABLE))
	CkManager::This->fatal("clock_settime(clock=%p): invalid memory",
			       clock);

    if (!ts || !TEST_PARAM(ts,sizeof(*ts),MEMADDR_READABLE))
	CkManager::This->fatal("clock_settime(ts=%p): invalid memory",
			       ts);

    err = rtlRunningNode->kdoor(clockSetTime)(clock,ts);

    t.kdoor(emit)();

    return err;
}

int clock_getres (clockid_t clock,
		  struct timespec *ts)
{
    int err;

    CkInAddr p1("clock",clock);
    CkInAddr p2("ts",ts);

    CkTracePoint t(65,&err,&p1,&p2,NULL);
    
    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_READABLE))
	CkManager::This->fatal("clock_getres(clock=%p): invalid memory",
			       clock);

    if (ts && !TEST_PARAM(ts,sizeof(*ts),MEMADDR_WRITABLE)) // May be null (?)
	CkManager::This->fatal("clock_getres(ts=%p): invalid memory",
			       ts);

    err = rtlRunningNode->kdoor(clockGetRes)(clock,ts);

    t.kdoor(emit)();

    return err;
}

hrtime_t clock_gethrtime (clockid_t clock)

{
    hrtime_t tval;

    CkInAddr p1("clock",clock);
    CkRetQuad rv(&tval);

    CkTracePoint t(66,NULL,&p1,&rv,NULL);
    
    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_READABLE))
	CkManager::This->fatal("clock_gethrtime(clock=%p): invalid memory",
			       clock);

    tval = rtlRunningNode->kdoor(clockGetHrTime)(clock);

    t.kdoor(emit)();

    return tval;
}

time_t time (time_t *tloc)

{
    CkInAddr p1("tloc",tloc);
    CkRetLong rv(tloc);

    CkTracePoint t(67,NULL,&p1,&rv,NULL);
    
    if (tloc && !TEST_PARAM(tloc,sizeof(*tloc),MEMADDR_READABLE))
	CkManager::This->fatal("time(tloc=%p): invalid memory",
			       tloc);

    time_t tval = rtlRunningNode->kdoor(clockTime)();

    if (tloc)
	*tloc = tval;

    t.kdoor(emit)();

    return tval;
}

int usleep (useconds_t interval)

{
    int err = 0;

    CkInUInt p1("interval",interval);
    
    CkTracePoint t(68,&err,&p1,NULL);

    err = rtlRunningNode->kdoor(usleep)(interval);

    t.kdoor(emit)();

    return err;
}

int nanosleep (const struct timespec *rqtp,
	       struct timespec *rmtp)
{
    int err = 0;

    CkInAddr p1("rqtp",rqtp);
    CkInAddr p2("rmtp",rmtp);
    
    CkTracePoint t(69,&err,&p1,&p2,NULL);

    if (!rqtp || !TEST_PARAM(rqtp,sizeof(*rqtp),MEMADDR_READABLE))
	CkManager::This->fatal("nanosleep(rqtp=%p): invalid memory",
			       rqtp);

    if (rmtp && !TEST_PARAM(rmtp,sizeof(*rmtp),MEMADDR_WRITABLE))
	CkManager::This->fatal("nanosleep(rmtp=%p): invalid memory",
			       rmtp);

    err = rtlRunningNode->kdoor(nanosleep)(rqtp,rmtp);

    t.kdoor(emit)();

    return err;
}

void rtl_delay (long ns)

{
    CkInLong p1("ns",ns);

    CkTracePoint t(70,NULL,&p1,NULL);

    ITime endSpinTime(SxClock);
    endSpinTime += (((double)ns) / 1000.0);

    // This busy loop won't be really accurate unless the node's Warp
    // factor is high enough for the SIMEX per-instruction time
    // increment to have the proper resolution. Anyway, this call
    // should be mainly useful in inserting delays between REAL
    // hardware accesses, so we don't care much of it under Ck.

    while (SxClock < endSpinTime)
	rtlRunningNode->khook(advanceClock)();
    
    t.kdoor(emit)();
}

hrtime_t gethrtime (void)

{
    hrtime_t now;

    CkRetQuad rv(&now);

    CkTracePoint t(71,NULL,&rv,NULL);
    
    now = rtlRunningNode->getHrTime();

    t.kdoor(emit)();

    return now;
}

hrtime_t gethrtimeres (void)

{
    hrtime_t resolution;

    CkRetQuad rv(&resolution);

    CkTracePoint t(72,NULL,&rv,NULL);
    
    resolution = rtlRunningNode->getHrRes();

    t.kdoor(emit)();

    return resolution;
}

int rtf_create (unsigned fifo,
		int size)
{
    int err;

    CkInInt p1("fifo",fifo);
    CkInInt p2("size",size);

    CkTracePoint t(73,&err,&p1,&p2,NULL);

    err = rtlRunningNode->kdoor(fifoCreate)(fifo,size);

    t.kdoor(emit)();

    return err;
}

int rtf_destroy (unsigned fifo)

{
    int err;

    CkInInt p1("fifo",fifo);

    CkTracePoint t(74,&err,&p1,NULL);

    err = rtlRunningNode->kdoor(fifoDestroy)(fifo);

    t.kdoor(emit)();

    return err;
}

int rtf_put (unsigned fifo,
	     void *buf,
	     int count)	
{
    int err;

    CkInInt p1("fifo",fifo);
    CkInAddr p2("buf",buf);
    CkInInt p3("count",count);

    CkTracePoint t(75,&err,&p1,&p2,&p3,NULL);

    if (!buf || !TEST_PARAM(buf,count,MEMADDR_READABLE))
	CkManager::This->fatal("rtf_put(buf=%p, count=%d): invalid memory",
			       buf,
			       count);

    err = rtlRunningNode->kdoor(fifoPut)(fifo,buf,count);

    t.kdoor(emit)();

    return err;
}

int rtf_get (unsigned fifo,
	     void *buf,
	     int count)	
{
    int err;

    CkInInt p1("fifo",fifo);
    CkInAddr p2("buf",buf);
    CkInInt p3("count",count);

    CkTracePoint t(76,&err,&p1,&p2,&p3,NULL);

    if (!buf || !TEST_PARAM(buf,count,MEMADDR_WRITABLE))
	CkManager::This->fatal("rtf_get(buf=%p, count=%d): invalid memory",
			       buf,
			       count);

    err = rtlRunningNode->kdoor(fifoGet)(fifo,buf,count);

    t.kdoor(emit)();

    return err;
}

int rtf_create_handler (unsigned fifo,
			int (*handler)(unsigned fifo))
{
    int err;

    CkInInt p1("fifo",fifo);
    CkInAddr p2("handler",(void *)handler);

    CkTracePoint t(77,&err,&p1,&p2,NULL);

    if (!handler || !TEST_PARAM(handler,1,MEMADDR_TEXT))
	CkManager::This->fatal("rtf_create_handler(handler=%p): invalid handler",
			       handler);

    err = rtlRunningNode->kdoor(fifoCreateUserHandler)(fifo,handler);

    t.kdoor(emit)();

    return err;
}

int rtf_create_rt_handler (unsigned fifo,
			   int (*handler)(unsigned fifo))
{
    int err;

    CkInInt p1("fifo",fifo);
    CkInAddr p2("handler",(void *)handler);

    CkTracePoint t(78,&err,&p1,&p2,NULL);

    if (!handler || !TEST_PARAM(handler,1,MEMADDR_TEXT))
	CkManager::This->fatal("rtf_create_rt_handler(handler=%p): invalid handler",
			       handler);

    err = rtlRunningNode->kdoor(fifoCreateRtHandler)(fifo,handler);

    t.kdoor(emit)();

    return err;
}

int rt_sem_init (rt_sem_t *sem, RT_SEMTYPE type, int init_val)

{
    int err = 0;

    CkInAddr p1("sem",sem);
    CkInInt p2("type",(int)sem);
    CkInInt p3("init_val",init_val);
    
    CkTracePoint t(79,&err,&p1,&p2,&p3,NULL);

    if (!sem || !TEST_PARAM(sem,sizeof(*sem),MEMADDR_WRITABLE))
	CkManager::This->fatal("rt_sem_init(sem=%p): invalid memory",
			       sem);

    if (init_val < 0 ||
	(type != RT_SEM_BINARY && type != RT_SEM_COUNTING) ||
	(type == RT_SEM_BINARY && init_val > 1))
	err = -EINVAL;
    else
	err = rtlRunningNode->kdoor(ipcSemInit)(sem,type,init_val);

    t.kdoor(emit)();

    return err;
}

int rt_sem_destroy (rt_sem_t *sem)

{
    int err = 0;

    CkInAddr p1("sem",sem);
    
    CkTracePoint t(80,&err,&p1,NULL);

    if (!sem || !TEST_PARAM(sem,sizeof(*sem),MEMADDR_WRITABLE))
	CkManager::This->fatal("rt_sem_destroy(sem=%p): invalid memory",
			       sem);

    err = rtlRunningNode->kdoor(ipcSemDestroy)(sem);

    t.kdoor(emit)();

    return err;
}

int rt_sem_post (rt_sem_t *sem)

{
    int err = 0;

    CkInAddr p1("sem",sem);
    
    CkTracePoint t(81,&err,&p1,NULL);

    if (!sem || !TEST_PARAM(sem,sizeof(*sem),MEMADDR_WRITABLE))
	CkManager::This->fatal("rt_sem_post(sem=%p): invalid memory",
			       sem);

    err = rtlRunningNode->kdoor(ipcSemPost)(sem);

    t.kdoor(emit)();

    return err;
}

int rt_sem_wait (rt_sem_t *sem, RTIME timeout)

{
    int err = 0;

    CkInAddr p1("sem",sem);
    CkInQuad p2("timeout",timeout);
    
    CkTracePoint t(82,&err,&p1,&p2,NULL);

    if (!sem || !TEST_PARAM(sem,sizeof(*sem),MEMADDR_WRITABLE))
	CkManager::This->fatal("rt_sem_wait(sem=%p): invalid memory",
			       sem);

    err = rtlRunningNode->kdoor(ipcSemWait)(sem,timeout);

    t.kdoor(emit)();

    return err;
}

int rt_sem_trywait (rt_sem_t *sem)

{
    int err = 0;

    CkInAddr p1("sem",sem);
    
    CkTracePoint t(83,&err,&p1,NULL);

    if (!sem || !TEST_PARAM(sem,sizeof(*sem),MEMADDR_WRITABLE))
	CkManager::This->fatal("rt_sem_trywait(sem=%p): invalid memory",
			       sem);

    err = rtlRunningNode->kdoor(ipcSemTryWait)(sem);

    t.kdoor(emit)();

    return err;
}

int rt_task_delay (RTIME duration)

{
    int err = 0;

    CkInQuad p1("duration",duration);
    
    CkTracePoint t(84,&err,&p1,NULL);

    err = rtlRunningNode->kdoor(ipcTaskDelay)(duration);

    t.kdoor(emit)();

    return err;
}

int rt_task_ipc_init (RT_TASK_IPC *task,
		      void (*fn)(int data),
		      int data,
		      int stack_size,
		      int priority)
{
    int err;

    CkInAddr p1("task",task);
    CkInAddr p2("fn",(void *)fn);
    CkInInt p3("data",data);
    CkInInt p4("stack_size",stack_size);
    CkInInt p5("priority",priority);

    CkTracePoint t(85,&err,&p1,&p2,&p3,&p4,&p5,NULL);

    if (!task || !TEST_PARAM(task,sizeof(*task),MEMADDR_WRITABLE))
	CkManager::This->fatal("rt_task_ipc_init(task=%p): invalid memory",
			       task);

    if (!fn || !TEST_PARAM(fn,1,MEMADDR_TEXT))
	CkManager::This->fatal("rt_task_ipc_init(fn=%p): invalid routine",
			       fn);

    if (stack_size >= PTHREAD_STACK_MIN && stack_size <= PTHREAD_STACK_MAX)
	{
	pthread_attr_t attr;
	rtlRunningNode->kdoor(threadAttrInit)(&attr);
	attr.initial_state = RTL_THREAD_DORMANT;
	attr.stack_size = stack_size;
	attr.sched_param.sched_priority = rtlRunningNode->getMaxPrio() - priority;
	err = -rtlRunningNode->kdoor(threadCreate)(MAKE_RT_TASK(task),
						   &attr,
						   (void *(*)(void *))fn,
						   (void *)data);
	}
    else
	err = -EINVAL;

    t.kdoor(emit)();

    return err;
}

int rt_mq_init (rt_mq_t *mq,
		int max_msgs,
		int msg_size)
{
    int err;

    CkInAddr p1("mq",mq);
    CkInInt p2("max_msgs",max_msgs);
    CkInInt p3("msg_size",msg_size);

    CkTracePoint t(86,&err,&p1,&p2,&p3,NULL);

    if (!mq || !TEST_PARAM(mq,sizeof(*mq),MEMADDR_WRITABLE))
	CkManager::This->fatal("rt_mq_init(mq=%p): invalid memory",
			       mq);

    if (max_msgs > 0 && msg_size >= 0)
	err = rtlRunningNode->kdoor(ipcMqInit)(mq,max_msgs,msg_size);
    else
	err = -EINVAL;

    t.kdoor(emit)();

    return err;
}

int rt_mq_destroy (rt_mq_t *mq)

{
    int err;

    CkInAddr p1("mq",mq);

    CkTracePoint t(87,&err,&p1,NULL);

    if (!mq || !TEST_PARAM(mq,sizeof(*mq),MEMADDR_WRITABLE))
	CkManager::This->fatal("rt_mq_destroy(mq=%p): invalid memory",
			       mq);

    err = rtlRunningNode->kdoor(ipcMqDestroy)(mq);

    t.kdoor(emit)();

    return err;
}

int rt_mq_send (rt_mq_t *mq,
		char *msg,
		RT_MQ_PRIO prio,
		RTIME wait)
{
    int err;

    CkInAddr p1("mq",mq);
    CkInAddr p2("msg",msg);
    CkInInt p3("prio",prio);
    CkInQuad p4("wait",wait);

    CkTracePoint t(88,&err,&p1,&p2,&p3,&p4,NULL);

    if (!mq || !TEST_PARAM(mq,sizeof(*mq),MEMADDR_WRITABLE))
	CkManager::This->fatal("rt_mq_send(mq=%p): invalid memory",
			       mq);

    if (prio != RT_MQ_NORMAL && prio != RT_MQ_URGENT)
	CkManager::This->fatal("rt_mq_send(prio=%d): invalid priority",
			       prio);

    err = rtlRunningNode->kdoor(ipcMqSend)(mq,msg,prio,wait);

    t.kdoor(emit)();

    return err;
}

int rt_mq_receive (rt_mq_t *mq,
		   char *msg,
		   RTIME wait)
{
    int err;

    CkInAddr p1("mq",mq);
    CkInAddr p2("msg",msg);
    CkInQuad p3("wait",wait);

    CkTracePoint t(89,&err,&p1,&p2,&p3,NULL);

    if (!mq || !TEST_PARAM(mq,sizeof(*mq),MEMADDR_WRITABLE))
	CkManager::This->fatal("rt_mq_receive(mq=%p): invalid memory",
			       mq);

    err = rtlRunningNode->kdoor(ipcMqReceive)(mq,msg,wait);

    t.kdoor(emit)();

    return err;
}

void rtl_hard_enable_irq (unsigned irq)

{
    CkInUInt p1("irq",irq);

    CkTracePoint t(90,NULL,&p1,NULL);

    rtlRunningNode->kdoor(hardEnableIrq)(irq);

    t.kdoor(emit)();
}

void rtl_hard_disable_irq (unsigned irq)

{
    CkInUInt p1("irq",irq);

    CkTracePoint t(91,NULL,&p1,NULL);

    rtlRunningNode->kdoor(hardDisableIrq)(irq);

    t.kdoor(emit)();
}

int rtl_request_global_irq (unsigned irq,
			    rtl_irq_handler_t handler)
{
    int err;

    CkInUInt p1("irq",irq);
    CkInAddr p2("handler",(void *)handler);

    CkTracePoint t(92,&err,&p1,&p2,NULL);

    if (!handler || !TEST_PARAM(handler,1,MEMADDR_TEXT))
	CkManager::This->fatal("rtl_request_global_irq(handler=%p): invalid handler",
			       handler);

    err = rtlRunningNode->kdoor(requestGlobalIrq)(irq,handler);

    t.kdoor(emit)();

    return err;
}

int rtl_free_global_irq (unsigned irq)

{
    int err;

    CkInUInt p1("irq",irq);

    CkTracePoint t(93,&err,&p1,NULL);

    err = rtlRunningNode->kdoor(freeGlobalIrq)(irq);

    t.kdoor(emit)();

    return err;
}

int rtl_get_soft_irq (rtl_soft_irq_handler_t handler,
		      const char *devname)
{
    int err = 0, irq;

    CkInAddr p1("handler",(void *)handler);
    CkInString p2("devname",devname);
    CkRetInt p3(&irq);

    CkTracePoint t(94,&err,&p1,&p2,&p3,NULL);

    if (!handler || !TEST_PARAM(handler,1,MEMADDR_TEXT))
	CkManager::This->fatal("rtl_get_soft_irq(handler=%p): invalid handler",
			       handler);

    if (!devname || !TEST_PARAM(devname,1,MEMADDR_READABLE))
	CkManager::This->fatal("rtl_get_soft_irq(devname=%p): invalid memory",
			       devname);

    irq = rtlRunningNode->kdoor(getSoftIrq)(handler,devname);

    if (irq < 0)
	{
	err = irq;
	irq = -1;
	}

    t.kdoor(emit)();

    return irq;
}

void rtl_free_soft_irq (unsigned irq)

{
    CkInUInt p1("irq",irq);

    CkTracePoint t(95,NULL,&p1,NULL);

    rtlRunningNode->kdoor(freeSoftIrq)(irq);

    t.kdoor(emit)();
}

void rtl_global_pend_irq (unsigned irq)

{
    CkInUInt p1("irq",irq);

    CkTracePoint t(96,NULL,&p1,NULL);

    rtlRunningNode->kdoor(globalPendIrq)(irq);

    t.kdoor(emit)();
}

void ck_rtl_no_interrupts (rtl_irqstate_t *statep)

{
    CkInAddr p1("statep",statep);

    CkTracePoint t(97,NULL,&p1,NULL);

    if (statep && !TEST_PARAM(statep,sizeof(*statep),MEMADDR_WRITABLE))
	CkManager::This->fatal("rtl_no_interrupts(statep=%p): invalid memory",
			       statep);

    cklevel_t olevel = rtlRunningNode->kdoor(imask)(RTL_IRQ_MAX);

    if (statep)
	*statep = olevel;

    t.kdoor(emit)();
}

void ck_rtl_restore_interrupts (rtl_irqstate_t *statep)

{
    CkTracePoint t(98,NULL,NULL);

    rtlRunningNode->kdoor(imask)(statep ? *statep : 0);

    t.kdoor(emit)();
}

int sem_init (sem_t *sem, int pshared, unsigned init_val)

{
    int err = 0;

    CkInAddr p1("sem",sem);
    CkInInt p2("pshared",pshared);
    CkInUInt p3("init_val",init_val);
    
    CkTracePoint t(99,&err,&p1,&p2,&p3,NULL);

    if (!sem || !TEST_PARAM(sem,sizeof(*sem),MEMADDR_WRITABLE))
	CkManager::This->fatal("sem_init(sem=%p): invalid memory",
			       sem);

    if (pshared != PTHREAD_PROCESS_SHARED &&
	 pshared != PTHREAD_PROCESS_PRIVATE)
	err = EINVAL;
    else
	err = rtlRunningNode->kdoor(stdSemInit)(sem,init_val);

    t.kdoor(emit)();

    return err;
}

int sem_destroy (sem_t *sem)

{
    int err = 0;

    CkInAddr p1("sem",sem);
    
    CkTracePoint t(100,&err,&p1,NULL);

    if (!sem || !TEST_PARAM(sem,sizeof(*sem),MEMADDR_WRITABLE))
	CkManager::This->fatal("sem_destroy(sem=%p): invalid memory",
			       sem);

    err = rtlRunningNode->kdoor(stdSemDestroy)(sem);

    t.kdoor(emit)();

    return err;
}

int sem_post (sem_t *sem)

{
    int err = 0;

    CkInAddr p1("sem",sem);
    
    CkTracePoint t(101,&err,&p1,NULL);

    if (!sem || !TEST_PARAM(sem,sizeof(*sem),MEMADDR_WRITABLE))
	CkManager::This->fatal("sem_post(sem=%p): invalid memory",
			       sem);

    err = rtlRunningNode->kdoor(stdSemPost)(sem);

    t.kdoor(emit)();

    return err;
}

int sem_timedwait (sem_t *sem, const struct timespec *abstime)

{
    int err = 0;

    CkInAddr p1("sem",sem);
    CkInAddr p2("abstime",abstime);
    
    CkTracePoint t(102,&err,&p1,&p2,NULL);

    if (!sem || !TEST_PARAM(sem,sizeof(*sem),MEMADDR_WRITABLE))
	CkManager::This->fatal("sem_timedwait(sem=%p): invalid memory",
			       sem);

    err = rtlRunningNode->kdoor(stdSemTimedWait)(sem,abstime);

    t.kdoor(emit)();

    return err;
}

int sem_wait (sem_t *sem)

{
    int err = 0;

    CkInAddr p1("sem",sem);
    
    CkTracePoint t(103,&err,&p1,NULL);

    if (!sem || !TEST_PARAM(sem,sizeof(*sem),MEMADDR_WRITABLE))
	CkManager::This->fatal("sem_wait(sem=%p): invalid memory",
			       sem);

    err = rtlRunningNode->kdoor(stdSemWait)(sem);

    t.kdoor(emit)();

    return err;
}

int sem_trywait (sem_t *sem)

{
    int err = 0;

    CkInAddr p1("sem",sem);
    
    CkTracePoint t(104,&err,&p1,NULL);

    if (!sem || !TEST_PARAM(sem,sizeof(*sem),MEMADDR_WRITABLE))
	CkManager::This->fatal("sem_trywait(sem=%p): invalid memory",
			       sem);

    err = rtlRunningNode->kdoor(stdSemTryWait)(sem);

    t.kdoor(emit)();

    return err;
}

int sem_getvalue (sem_t *sem, int *sval)

{
    int err = 0;

    CkInAddr p1("sem",sem);
    CkOutInt p2("sval",sval);
    
    CkTracePoint t(105,&err,&p1,&p2,NULL);

    if (!sem || !TEST_PARAM(sem,sizeof(*sem),MEMADDR_WRITABLE))
	CkManager::This->fatal("sem_getvalue(sem=%p): invalid memory",
			       sem);

    if (!sval || !TEST_PARAM(sval,sizeof(*sval),MEMADDR_WRITABLE))
	CkManager::This->fatal("sem_getvalue(sval=%p): invalid memory",
			       sval);

    *sval = (int)rtlRunningNode->kdoor(stdSemGetValue)(sem);

    t.kdoor(emit)();

    return err;
}

// The following routines are untraced convenience service calls.

clockid_t ck_clock_getrealtime (void) {

    return rtlRunningNode->getUstClock();
}

clockid_t ck_clock_getust (void) {

    return rtlRunningNode->getRealtimeClock();
}

void conpr (const char *s) {

    if (!s || !TEST_PARAM(s,1,MEMADDR_READABLE))
	CkManager::This->fatal("conpr(s=%p): invalid memory",
			       s);
    puts(s);
    fflush(stdout);
}

void conprn (unsigned int hexnum) {
    conpr(CString().format("%8x",hexnum));
}
