/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Implementation of the RTLinux clock and timer support.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation
#endif // __GNUG__
#include <config.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include "CkManager.h"
#include "CkInterrupt.h"
#include "rtlNode.h"
#include "rtlClock.h"

#define RTL_CLOCK_INITIALIZER { \
	rtl_clock_wrap_init, \
	rtl_clock_wrap_uninit, \
	rtl_clock_wrap_gethrtime, \
	rtl_clock_wrap_sethrtime, \
	rtl_clock_wrap_settimer, \
	rtl_clock_wrap_settimermode, \
	rtl_clock_wrap_handler, \
	RTL_CLOCK_MODE_UNINITIALIZED, \
	0, \
	0, \
	0, \
}

extern "C" {

static int rtl_clock_wrap_init (struct rtl_clock *clock) {

    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_WRITABLE))
	CkManager::This->fatal("clock_init(clock=%p): invalid memory",
			       clock);
    return 0;
}

static void rtl_clock_wrap_uninit (struct rtl_clock *clock)

{
    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_READABLE))
	CkManager::This->fatal("clock_uninit(clock=%p): invalid memory",
			       clock);

    rtlClock *iclock = rtlRunningNode->findClockByPtr(clock);

    if (!iclock)
	CkManager::This->fatal("clock_uninit(clock=%p): invalid clock",
			       clock);
    iclock->stop();
}

static hrtime_t rtl_clock_wrap_gethrtime (struct rtl_clock *clock)

{
    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_READABLE))
	CkManager::This->fatal("clock_gethrtime(clock=%p): invalid memory",
			       clock);

    rtlClock *iclock = rtlRunningNode->findClockByPtr(clock);

    if (!iclock)
	CkManager::This->fatal("clock_gethrtime(clock=%p): invalid clock",
			       clock);
    return iclock->getTime();
}

static int rtl_clock_wrap_sethrtime (struct rtl_clock *clock,
				     hrtime_t t)
{
    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_WRITABLE))
	CkManager::This->fatal("clock_sethrtime(clock=%p): invalid memory",
			       clock);

    rtlClock *iclock = rtlRunningNode->findClockByPtr(clock);

    if (!iclock)
	CkManager::This->fatal("clock_sethrtime(clock=%p): invalid clock",
			       clock);

    iclock->setTime(t);

    return 0;
}

static int rtl_clock_wrap_settimer (struct rtl_clock *clock,
				    hrtime_t interval)
{
    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_READABLE))
	CkManager::This->fatal("clock_settimer(clock=%p): invalid memory",
			       clock);

    rtlClock *iclock = rtlRunningNode->findClockByPtr(clock);

    if (!iclock)
	CkManager::This->fatal("clock_settimer(clock=%p): invalid clock",
			       clock);

    enum CfSourceType srcType;

    if (clock->mode == RTL_CLOCK_MODE_PERIODIC)
	srcType = CfEventPeriodical;
    else if (clock->mode == RTL_CLOCK_MODE_ONESHOT)
	srcType = CfEventTimer;
    else
	return -1;

    iclock->setTimer(srcType,interval);

    return 0;
}

static int rtl_clock_wrap_settimermode (struct rtl_clock *clock, int mode) {

    if (!clock || !TEST_PARAM(clock,sizeof(*clock),MEMADDR_WRITABLE))
	CkManager::This->fatal("clock_settimermode(clock=%p): invalid memory",
			       clock);

    if (mode != RTL_CLOCK_MODE_ONESHOT && mode != RTL_CLOCK_MODE_PERIODIC)
	return -EINVAL;

    clock->mode = mode;

    return 0;
}

static void rtl_clock_wrap_handler (struct pt_regs *regs) {
}

static struct rtl_clock clock_default_init = RTL_CLOCK_INITIALIZER;

static int rtl_timer_tick_trampoline (ckvector_t vector, void *cookie) {

    rtlSystemTimer *sysTimer = (rtlSystemTimer *)cookie;
    sysTimer->kisrt(tick)();

    return CK_ISR_DEFERRED;	// Schedule the DSR for soft handling.
}

};

rtlSystemTimer::rtlSystemTimer (hrtime_t _resolution,
				unsigned _irq,
				rtlNode *_node) :
    rtlInterrupt(_irq,
		 &rtl_timer_tick_trampoline,
		 NULL,
		 CK_RTC_NAME,
		 this,
		 _node)
{
    now = 0;
    tickHandler = NULL;
    setTimer(CfEventPeriodical,_resolution);
    clrStatus(CK_INTR_MASKED);
}

void rtlSystemTimer::kisrt(tick) ()

{
    now += resolution;
    setStatus(CK_INTR_ACKED);
    getNode()->kdoor(tick)();

    if (tickHandler)
	tickHandler(NULL);
}

rtlClock::rtlClock (rtlSystemTimer *_sysTimer,
		    struct rtl_clock *_uclock, // User-side clock struct ptr.
		    rtlNode *_node) :
    CkObject(_node)
{
    sysTimer = _sysTimer;
    uclock = _uclock;

    *uclock = clock_default_init;
    uclock->resolution = getRes();
    uclock->mode = RTL_CLOCK_MODE_PERIODIC;

    _node->nameObject(this,RTLINUX_CLOCK_INODE);
    uclock->handle = getCkId();
}

rtlClock::~rtlClock () {

    getNode()->unnameObject(this);
}

rtlClockRealtime::rtlClockRealtime (rtlSystemTimer *_sysTimer,
				    struct rtl_clock *_uclock,
				    rtlNode *_node) :
    rtlClock(_sysTimer,
	     _uclock,
	     _node) {
}

rtlClockUnadjusted::rtlClockUnadjusted (rtlSystemTimer *_sysTimer,
					struct rtl_clock *_uclock,
					rtlNode *_node) :
    rtlClock(_sysTimer,
	     _uclock,
	     _node) {
}
