/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Implementation of the RTLinux node configuration classes.
 * This code is based on the CfClass package.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation
#endif // __GNUG__
#include <config.h>
#include <stdlib.h>
#include "rtlConfig.h"
#include "rtlNode.h"

BEGIN_RINFO
RDEFINE("RTLinux","RTLinux target",CKDB_OSNODE,rtlConfig,CF_NONE,0,9)
END_RINFO

const struct CkInodeType rtlInodeTypes[] = {

    { "thread", RTLINUX_THREAD_INODE },
    { "clock", RTLINUX_CLOCK_INODE },
    { "mutex", RTLINUX_MUTEX_INODE },
    { "semaphore", RTLINUX_SEMA4_INODE },
    { "mqueue", RTLINUX_MESGQ_INODE },
    { "condvar", RTLINUX_CONDV_INODE },
    { "fifo", RTLINUX_FIFO_INODE },
    { NULL, -1 }
};

static void rtlCleanup ()

{
    CkNodeIterator it(CkNodeList);
    rtlNode *node;

    while ((node = (rtlNode *)it.next()) != NULL)
	node->cleanup();
}

rtlConfig::rtlConfig (const char *_instanceName,
			CfContext *_context) :
    CfClass(_instanceName,_context)
{
    // Kernel configuration group
    
    HandlerAttributePointer startAddrPtr =
	new HandlerAttribute("Init routine",this,"RTLinux");
    startAddrPtr->setValue("init_module");

    HandlerAttributePointer cleanupAddrPtr =
	new HandlerAttribute("Cleanup routine",this,"RTLinux"); // No default

    IntegerAttributePointer maxFifoPtr =
	new IntegerAttribute("Max. FIFOs",this,0,256,"RTLinux");
    maxFifoPtr->setValue(64);

    IntegerAttributePointer maxWaitersPtr =
	new IntegerAttribute("Max. mutex waiters",this,1,1024,"RTLinux");
    maxWaitersPtr->setValue(20);

    IntegerAttributePointer prioMinPtr =
	new IntegerAttribute("Min. thread priority",this,0,1000000,"RTLinux");
    prioMinPtr->setValue(0);

    IntegerAttributePointer prioMaxPtr =
	new IntegerAttribute("Max. thread priority",this,0,1000000,"RTLinux");
    prioMaxPtr->setValue(1000000);

    CheckAttributePointer simTogglePtr =
	new CheckAttribute("Options",this,"RTLinux");
    simTogglePtr->addChoice("Warn upon missing periodic deadlines");
    simTogglePtr->addChoice("Warn upon deleting pended resources");
    simTogglePtr->addChoice("Warn upon detecting priority inversion");
    simTogglePtr->addChoice("Loose call context checking");
    simTogglePtr->setValue(RTL_WARN_REMOVED_SYNCH|
			   RTL_WARN_PRIO_INVERSION);

    // VRTOS configuration group

    ScaleAttributePointer warpPtr =
	new ScaleAttribute("Target Warp",this,0.0,10.0,0.1,"CarbonKernel");
    warpPtr->setValue(5.0);

    IntegerAttributePointer rtcIrqPtr =
	new IntegerAttribute("Timer IRQ",this,0,RTL_IRQ_MAX - 1,"CarbonKernel");
    rtcIrqPtr->setValue(0);

    RealAttributePointer rtcFreqPtr =
	new RealAttribute("Timer Frequency (Hz)",this,1,infinity(),"CarbonKernel");
    rtcFreqPtr->setValue(1193180.0);

    ResourceAttributePointer interruptPtr =
	new ResourceAttribute("Interrupts",this,CKDB_INTR,"CarbonKernel");

    ResourceAttributePointer panelPtr =
	new ResourceAttribute("Control panels",this,CKDB_PANEL,"CarbonKernel");
}

rtlConfig::rtlConfig (DBapl *apl) :
    CfClass(apl) {
}

rtlConfig::rtlConfig (rtlConfig& src) :
    CfClass(src) {
}

void rtlConfig::instantiate (CfContext *context)

{
    HandlerAttributePointer startAddrPtr(get("Init routine"));
    HandlerAttributePointer cleanupAddrPtr(get("Cleanup routine"));
    IntegerAttributePointer maxFifoPtr(get("Max. FIFOs"));
    IntegerAttributePointer maxWaitersPtr(get("Max. mutex waiters"));
    IntegerAttributePointer prioMinPtr(get("Min. thread priority"));
    IntegerAttributePointer prioMaxPtr(get("Max. thread priority"));
    ScaleAttributePointer warpPtr(get("Target Warp"));
    IntegerAttributePointer rtcIrqPtr(get("Timer IRQ"));
    RealAttributePointer rtcFreqPtr(get("Timer Frequency (Hz)"));
    CheckAttributePointer simTogglePtr(get("Options"));
    ResourceAttributePointer interruptPtr(get("Interrupts"));
    ResourceAttributePointer panelPtr(get("Control panels"));

    // Console configuration block
    
    CkConsoleParms cparms;
    cparms.iclistMax = CK_CLIST_INFINITE;
    cparms.oclistMax = CK_CLIST_INFINITE;

    // VRTOS configuration block
    
    CkNodeParms nparms;

    nparms.nodeSerial = getSerialNumber(context);
    nparms.warp = warpPtr->getValue();
    nparms.priMode = PRDN; // i.e. The higher the value, the higher the priority.
    nparms.maxIntrLevel = RTL_IRQ_MAX;
    nparms.rtcIntrVector = 0;	// Unused.
    nparms.rtcIntrLevel = (cklevel_t)rtcIrqPtr->getValue();
    nparms.rtcIntrHandler = NULL; // Not used for RTLinux.
    nparms.rtcHz = rtcFreqPtr->getValue();
    nparms.cparms = cparms;
    nparms.inodeTypes = rtlInodeTypes;

    // Initialize RTLinux node using current configuration

    rtlNode *node = new rtlNode(getSerialName(context),&nparms);

    node->customOptions = simTogglePtr->getValue();
    node->startAddr = (int (*)(void))startAddrPtr->resolveAddr(context);
    node->cleanupAddr = cleanupAddrPtr->resolveAddrIfAny(context);
    node->maxFifos = maxFifoPtr->getValue();
    node->maxMutexWaiters = maxWaitersPtr->getValue();
    node->minPrio = prioMinPtr->getValue();
    node->maxPrio = prioMaxPtr->getValue();
    
    CfContextList subContextList;

    if (interruptPtr.valid())
	{
	interruptPtr->instantiate(context,subContextList,node);
	subContextList.destroy();
	}
    
    if (panelPtr.valid())
	{
	panelPtr->instantiate(context,subContextList,node);
	subContextList.destroy();
	}

    atexit(rtlCleanup);
}
