/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Implementation of the RTLinux FIFO management routines.
 * Since a non-privileged user process cannot make nodes in /dev, simulated
 * RTL FIFOs are created in a publicly accessible directory under /tmp
 * (or whichever directory is referred to by stdio's P_tmpdir macro).
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation
#endif // __GNUG__
#include <config.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include "CkManager.h"
#include "rtlNode.h"
#include "rtlFifo.h"

rtlFifoGList allFifos;

rtlFifo::rtlFifo (unsigned _minor,
		  rtlNode *_node) :
    CkObject(_node), SxCallout(NULL)
{
    minor = _minor;
    fd = -1;
    written = 0;
    rtHandler = NULL;
    userHandler = NULL;

    _node->nameObject(this,RTLINUX_FIFO_INODE);
    SxThread::runChain.addCallout(this);

    create();
    allFifos.append(this);
}

rtlFifo::~rtlFifo ()

{
    if (fd != -1)
	{
	close(fd);
	unlink(fifoPath);
	}

    SxThread::runChain.removeCallout(this);
    getNode()->unnameObject(this);
    allFifos.remove(this);
}

void rtlFifo::create ()

{
    CString tmpDevDir;
    struct stat sbuf;

    tmpDevDir = P_tmpdir + CString("/ckdev");

    // Tmpdev directory permission bits are forced using chmod(2) to
    // get rid of process' current umask value.

    if (stat(tmpDevDir,&sbuf) &&
	(mkdir(tmpDevDir,0777) < 0 || chmod(tmpDevDir,0777) < 0))
	CkManager::This->fatal("cannot create directory %s",
			       (const char *)tmpDevDir);

    fifoPath = tmpDevDir + "/rtf";
    fifoPath += CString(minor);

    unlink(fifoPath);

    if (mknod(fifoPath,S_IFIFO|0666,0) < 0)
	CkManager::This->fatal("cannot create FIFO %s, errno %d",
			       (const char *)fifoPath,errno);

    fd = open(fifoPath,O_RDWR|O_NONBLOCK,0);

    if (fd < 0)
	CkManager::This->fatal("cannot open FIFO %s, errno %d",
			       (const char *)fifoPath,errno);

    // Return to blocking mode.
    fcntl(fd,F_SETFL,fcntl(fd,F_GETFL,0) & ~O_NONBLOCK);
    // Always set the close-on-exec bit for fifos.
    fcntl(fd,F_SETFD,FD_CLOEXEC);
}

int rtlFifo::put (CkRunnable *runnable, const void *buf, int count)

{
    int nbytes = 0;

    while (nbytes < count)
	{
	int n = ::write(fd,(char *)buf + nbytes,count - nbytes);

	if (n <= 0)
	    {
	    runnable->setErrno(-errno);
	    return -1;
	    }

	nbytes += n;
	}

    written = 1;
    
    if (rtHandler)
	rtHandler(minor);

    return count;
}

int rtlFifo::get (CkRunnable *runnable, void *buf, int count)

{
    struct timeval tv = { 0, 0 };
    fd_set rset;

    FD_ZERO(&rset);
    FD_SET(fd,&rset);

    if (select(fd + 1,&rset,NULL,NULL,&tv) != 1)
	return 0;

    int nbytes = ::read(fd,(char *)buf,count);

    if (nbytes < 0)
	{
	runnable->setErrno(-errno);
	return -1;
	}

    if (rtHandler)
	rtHandler(minor);

    return nbytes;
}

void rtlFifo::process ()

{
    if (!userHandler)
	return;

    struct timeval tv = { 0, 0 };
    fd_set rset, wset;

    FD_ZERO(&rset);
    FD_ZERO(&wset);
    FD_SET(fd,&rset);
    FD_SET(fd,&wset);
    select(fd + 1,&rset,&wset,NULL,&tv);

    // If the fifo is readable, the user-side process must have
    // written to it, therefore schedule the handler. If the fifo is writable
    // after we have just wrote to it, assume the user-side process
    // have consumed our data, and therefore schedule the handler too.
    // Ok, the latter case is quite discutable since a FIFO has a system-
    // maintained buffer, but user handlers seem used as means to ensure
    // that user-side processes are able to receive more data from the
    // RT-side. Let us assume that being able to pass those data to the
    // system with no loss is what the RT-side needs most of the time.

    readable = FD_ISSET(fd,&rset);
    writable = FD_ISSET(fd,&wset);

    // RTLinux fires the FIFO user handler on behalf of the
    // low-priority Linux thread from the file ops routines. We map
    // this context to the rtlLinuxThread object.

    if (readable || (written && writable))
	{
	// WARNING: do *not* call any instrumented code on behalf of
	// the current callout method, otherwise, strange behaviour will
	// certainly take place! Since a FIFO handler is usually instrumented
	// by ckcc, we just tell the node to wake up the low-priority
	// thread standing for the Linux context which will fire
	// the handler in turn.
	written = 0;
	getNode()->signalLinux(RTL_FIFO_PEND_HANDLER);
	}
}

void rtlFifo::kcout(runHandlers) (rtlNode *node)

{
    rtlFifoIterator it(allFifos);
    rtlFifo *fifo;
    
    while ((fifo = it.next()) != NULL)
	{
	if (fifo->getNode() == node &&
	    (fifo->readable || fifo->writable))
	    {
	    fifo->readable = 0;
	    fifo->writable = 0;

	    if (fifo->userHandler)
		fifo->userHandler(fifo->minor);
	    }
	}
}
