/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Implementation of the RTLinux target abstraction.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation "rtlNode.h"
#pragma implementation "rtlProto.h"
#endif // __GNUG__
#include <config.h>
#include <sys/types.h>
#include <signal.h>
#include <string.h>
#include <memory.h>
#include <errno.h>
#include "toolshop.h"
#include "CkManager.h"
#include "rtlNode.h"
#include "rtlProto.h"
#include "rtlThread.h"
#include "rtlSynch.h"
#include "rtlClock.h"
#include "rtlTrace.h"
#include "rtlFifo.h"
#include "rtlInterrupt.h"

extern "C" {

static RETSIGTYPE rtl_sigpipe_handler (int sig)

{
#if RETSIGTYPE != void
    return (RETSIGTYPE)0;
#endif
}

}

rtlNode::rtlNode (const char *_nodeName,
		  CkNodeParms *_nparms) :
    CkNode("RTLinux",
	   _nodeName,
	   RTLINUX_MAGIC,
	   _nparms)
{
    // Create the standard RTLinux clock objects.

    hrtime_t timerRes = (hrtime_t)(1000000000.0 / getTicks2Sec()); // nanoseconds

    sysTimer = new rtlSystemTimer(timerRes,
				  _nparms->rtcIntrLevel, // IRQ# actually.
				  this);

    clockRT = new rtlClockRealtime(sysTimer,
				   &clock_realtime,
				   this);

    clockUST = new rtlClockUnadjusted(sysTimer,
				      &clock_ust,
				      this);

    // Export the RTL API tracepoints to the ISE.

    setTraceInterface(new rtlTraceInterface(this));

    // Clear IRQ table.

    for (unsigned irq = 0; irq < RTL_IRQ_MAX; irq++)
	irqTable[irq] = NULL;
    
    // FIFO settings.

    for (unsigned n = 0; n < RTF_NO; n++)
	fifoTable[n] = NULL;

    struct sigaction sa;
    sa.sa_handler = (SIGHANDLER_TYPE)rtl_sigpipe_handler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sigaction(SIGPIPE,&sa,NULL);
}

// warmInit -- proceeds to the warm initialization chores of an RTLinux
// node. This mainly consists of starting the pseudo-Linux thread which
// will call init_module() in turn. The scheduling is automatically started
// by the VRTOS after this method returns.

void rtlNode::kinit(warmInit) ()

{
    CkNode::kinit(warmInit)();
    
    // Create a placeholder for the low-priority Linux thread.

    linuxThread = new rtlLinuxThread(this);
    declareThread(linuxThread);
    startThread(linuxThread);
}

void rtlNode::protoInit ()

{
    rtlNodeExportMessage nem(this);
    protoExport(&nem,sizeof(nem));
}

ckticks_t rtlNode::rrGetCredit () {

    return 0;	// i.e. No round-robin scheduling.
}

void (*rtlNode::getRootEntry ()) (...) {

    return (void (*)(...))startAddr;
}

void rtlNode::cleanup ()

{
    for (int fifo = 0; fifo < RTF_NO; fifo++)
	{
	if (fifoTable[fifo])
	    {
	    delete fifoTable[fifo];
	    fifoTable[fifo] = NULL;
	    }
	}
}

CkInterrupt *rtlNode::createInterrupt (ckvector_t _vector,
				       cklevel_t _level,
				       ckisr_t *_isr,
				       ckdsr_t *_dsr,
				       void *_cookie,
				       const char *_symbolicName)
{
    unsigned irq = _level - 1;

    if (irqTable[irq])
	CkManager::This->fatal("IRQ%u multiply defined",irq);

    return new rtlInterrupt(irq,
			    NULL, // Use default trampoline.
			    (rtl_irq_handler_t)_isr,
			    _symbolicName,
			    NULL,
			    this);
}

void rtlNode::nameObject (CkObject *object, int type)

{
    int seqno = allocID();

    if (seqno < 0)
	CkManager::This->fatal("cannot generate unique identifiers");

    CkId id(getNid(),type,0,seqno);
    mapObject((CkHandle)id,type,CkLocalDomain,object);
    
    object->setCkId(id);
    object->setCkExtId(seqno);
    tellNameServer(id,CK_MSGPORT_NAMEOBJ);
}

void rtlNode::unnameObject (CkObject *object)

{
    unmapObject((CkHandle)object->getCkId(),
		object->getCkType(),
		CkLocalDomain);
    releaseID(object->getCkId().getSeq());
    tellNameServer(object->getCkId(),CK_MSGPORT_UNNAMEOBJ);
}

CkObject *rtlNode::findActiveObject (CkId id) {

    return findObject((CkHandle)id,id.getType(),CkLocalDomain);
}

rtlThread *rtlNode::findThreadByPtr (pthread_t pthread)

{
    CkId threadid(pthread->handle);
    rtlThread *thread = NULL;

    if (threadid.getType() == RTLINUX_THREAD_INODE)
	thread = (rtlThread *)findActiveObject(threadid);

    return thread;
}

rtlMutex *rtlNode::findMutexByPtr (pthread_mutex_t *mutex)

{
    CkId mutexid(mutex->handle);
    rtlMutex *imutex = NULL;

    if (mutexid.getType() == RTLINUX_MUTEX_INODE)
	imutex = (rtlMutex *)findActiveObject(mutexid);

    return imutex;
}

rtlSema4 *rtlNode::findSema4ByPtr (rt_sem_t *sem)

{
    CkId semid(sem->handle);
    rtlSema4 *sema4 = NULL;

    if (semid.getType() == RTLINUX_SEMA4_INODE)
	sema4 = (rtlSema4 *)findActiveObject(semid);

    return sema4;
}

rtlMq *rtlNode::findMqByPtr (rt_mq_t *mq)

{
    CkId mqid(mq->handle);
    rtlMq *imq = NULL;

    if (mqid.getType() == RTLINUX_MESGQ_INODE)
	imq = (rtlMq *)findActiveObject(mqid);

    return imq;
}

rtlCondVar *rtlNode::findCondVarByPtr (pthread_cond_t *cond)

{
    CkId cvid(cond->handle);
    rtlCondVar *cv = NULL;

    if (cvid.getType() == RTLINUX_CONDV_INODE)
	cv = (rtlCondVar *)findActiveObject(cvid);

    return cv;
}

rtlClock *rtlNode::findClockByPtr (clockid_t clock)

{
    CkId ckid(clock->handle);
    rtlClock *iclock = NULL;

    if (ckid.getType() == RTLINUX_CLOCK_INODE)
	iclock = (rtlClock *)findActiveObject(ckid);

    return iclock;
}

void rtlNode::mustBeInit ()

{
    if (getActiveThread() != linuxThread)
	CkManager::This->fatal("service is available from init context only");
}

void rtlNode::mustBeInitOrSynch ()

{
    if (isAsynch()) // Init is performed on behalf of the linux thread.
	CkManager::This->fatal("illegal asynchronous call");
}

void rtlNode::deleteThread (CkThread *thread)

{
    ((rtlThread *)thread)->cleanup();
    CkNode::deleteThread(thread);
}

// rtlNode::tick() -- Receives each timer tick notification from the realtime
// clock object. It should be noted that since an interrupt context always
// underlies the clock handler execution, there is no need for an explicit
// rescheduling call since Ck automatically reschedules when the (outer)
// interrupt handler returns.

void rtlNode::kdoor(tick) ()

{
    hrtime_t now = sysTimer->getTime();
    clock_realtime.value = now;
    clock_ust.value = now;

    CkNode::kdoor(tick)();
}

int rtlNode::kdoor(threadCreate)(pthread_t *ppthread,
				 const pthread_attr_t *attr,
				 void *(*startAddr)(void *arg),
				 void *startArg)
{
    pthread_attr_t attr_default;

    if (!testOption(RTL_LOOSE_CALLCHECK))
	mustBeInit();

    if (!attr)
	{
	kdoor(threadAttrInit)(&attr_default);
	attr = &attr_default;
	}
    
    if (attr->sched_param.sched_priority < minPrio ||
	attr->sched_param.sched_priority > maxPrio)
	return EINVAL;

    *ppthread = new rtl_thread_struct;

    rtlThread *thread = new rtlThread(attr->sched_param.sched_priority,
				      startAddr,
				      startArg,
				      attr->stack_size,
				      *ppthread,
				      NULL,
				      this);
    thread->protoInit();
    thread->setCpu(attr->cpu);
    (*ppthread)->handle = thread->getCkId();
    (*ppthread)->cleanup = NULL;
    declareThread(thread);

    if (attr->detachstate == PTHREAD_CREATE_JOINABLE)
	thread->setStatus(RTL_THREAD_JOINABLE);
    else
	thread->clrStatus(RTL_THREAD_JOINABLE);

    if (attr->initial_state == RTL_THREAD_READY)
	startThread(thread);

    return 0;
}

// threadExit -- This must remain the unique way to delete a thread.
// Note that we need to lock the scheduling until the thread is
// actually destroyed as Ck's own mutexes and condvars
// used in joining call the rescheduling procedure internally (If the
// destroyed thread is the running one, then the scheduling will be
// unlocked automagically by Ck).
// This way, threadExit() looks atomic from the application's standpoint.

void rtlNode::kdoor(threadExit)(void *retval)

{
    mustBeSynch();
    rtlRunningThread->lock();
    rtlRunningThread->signalJoin(retval);
    deleteThread(rtlRunningThread);
    rtlRunningThread->unlock();
    resched();
}

// threadKill() -- sends a signal to a thread.
// RTLinux's signaling feature is directly mapped onto Ck's ASR
// mechanism.

int rtlNode::kdoor(threadKill)(pthread_t pthread, int signo)

{
    rtlThread *thread = findThreadByPtr(pthread);

    if (!thread)
	return ESRCH;

    if (signo < RTL_SIGNAL_NULL || signo > RTL_SIGNAL_LAST)
	return EINVAL;

    // Note: RTLinux behaviour is emulated by waiting for the next
    // rescheduling to dispatch the pending signals.

    if (signo != RTL_SIGNAL_NULL)
	{
	// If the killed thread is currently blocked, it may be unblocked
	// forcibly to allow it to run its ASR, depending on the signal/state
	// pair. The way Ck handles the ASR mecanism guarantees the thread
	// enters the signal handler before returning to the user code.
	// It should be also noted that suspensive conditions are not really
	// additive under RTLinux (they are under Ck) but threads pending
	// for a resource rather enter a wait loop ensuring they sleep until
	// the claimed resource is actually available, even if a signal wakes
	// them up in the meantime.

	thread->setSignals(1 << (signo - 1));

	if (!thread->isReady())
	    {
	    switch (signo)
		{
		case RTL_SIGNAL_DELETE:

		    if (thread->tstStatus(RTL_CANCEL_ENABLED))
			{
			if (thread->tstStatus(CKPEND))
			    unblockThread(thread);

			if (thread->tstStatus(CKSUSP))
			    readyThread(thread,CKSUSP);
			}

		    thread->setStatus(RTL_CANCEL_PENDING);
		    break;

		case RTL_SIGNAL_WAKEUP:

		    if (!thread->tstStatus(CKPEND))
			unblockThread(thread);

		    if (thread->tstStatus(CKSUSP))
			readyThread(thread,CKSUSP);

		    break;

		case RTL_SIGNAL_SUSPEND:

		    if (!thread->tstStatus(CKSUSP))
			suspendThread(thread,CKSUSP);

		    break;
		}
	    } // Otherwise, the ASR will do the job.

	pendReschedule();
	}

    return 0;
}

int rtlNode::kdoor(threadSetSchedParam) (pthread_t pthread,
					 int policy,
					 const rtl_sched_param *param)
{
    rtlThread *thread = findThreadByPtr(pthread);

    if (!thread)
	return ESRCH;

    if (thread->getPriNoPip() != param->sched_priority)
	reniceThread(thread,param->sched_priority);

    return 0;
}

int rtlNode::kdoor(threadGetSchedParam) (pthread_t pthread,
					 int *policy,
					 rtl_sched_param *param)
{
    rtlThread *thread = findThreadByPtr(pthread);

    if (!thread)
	return ESRCH;

    param->sched_priority = thread->getPri(); // Consider PIP boost too.
    *policy = SCHED_FIFO;

    return 0;
}

int rtlNode::kdoor(threadSetCancelState) (int state,
					  int *oldstate)
{
    mustBeSynch();

    if (state != PTHREAD_CANCEL_ENABLE && state != PTHREAD_CANCEL_DISABLE)
	return EINVAL;

    if (oldstate)
	{
	if (rtlRunningThread->tstStatus(RTL_CANCEL_ENABLED))
	    *oldstate = PTHREAD_CANCEL_ENABLE;
	else
	    *oldstate = PTHREAD_CANCEL_DISABLE;
	}

    if (state == PTHREAD_CANCEL_ENABLE)
	rtlRunningThread->setStatus(RTL_CANCEL_ENABLED);
    else
	rtlRunningThread->clrStatus(RTL_CANCEL_ENABLED);

    // Note: Some implementations may immediately dispatch a pending
    // cancellation request when the state is reenabled, but RTLinux
    // doesn't (since the OpenGroup standards don't require it anyway).
    // So don't we.

    return 0;
}

int rtlNode::kdoor(threadSetCancelType) (int type,
					 int *oldtype)
{
    mustBeSynch();

    if (type != PTHREAD_CANCEL_DEFERRED && type != PTHREAD_CANCEL_ASYNCHRONOUS)
	return EINVAL;

    if (oldtype)
	{
	if (rtlRunningThread->tstStatus(RTL_CANCEL_DEFERRED))
	    *oldtype = PTHREAD_CANCEL_DEFERRED;
	else
	    *oldtype = PTHREAD_CANCEL_ASYNCHRONOUS;
	}

    if (type == PTHREAD_CANCEL_DEFERRED)
	rtlRunningThread->setStatus(RTL_CANCEL_DEFERRED);
    else
	rtlRunningThread->clrStatus(RTL_CANCEL_DEFERRED);

    return 0;
}

void rtlNode::kdoor(threadTestCancel) (void)

{
    mustBeSynch();

    if (rtlRunningThread->tstStatus(RTL_CANCEL_ENABLED) &&
	rtlRunningThread->tstStatus(RTL_CANCEL_PENDING))
	kdoor(threadExit)(PTHREAD_CANCELED);
}

int rtlNode::kdoor(threadJoin) (pthread_t pthread,
				void **value_ptr)
{
    mustBeSynch();

    rtlThread *thread = findThreadByPtr(pthread);

    if (!thread)
	return ESRCH;

    if (thread == rtlRunningThread)
	return EDEADLK;

    if (!thread->tstStatus(RTL_THREAD_JOINABLE))
	return EINVAL;

    void *exitValue = rtlRunningThread->waitJoin(thread);

    if (value_ptr)
	*value_ptr = exitValue;

    return 0;
}

void rtlNode::kdoor(threadYield)()

{
    mustBeSynch();
    readyThread(rtlRunningThread); // This causes a round-robin effect
    resched();
}

// rtlNode::threadSelf() -- returns the thread ID of the calling
// thread. The OpenGroup standards do not define error condition
// for this call. So do our best when no user thread is running
// but an asynchonous activity by returning the linux pseudo-thread handle.

pthread_t rtlNode::kdoor(threadSelf)()

{
    if (rtlRunningThread)
	return rtlRunningThread->getUserHandle();

    return isRunning() ? NULL : linuxThread->getUserHandle();
}

void rtlNode::kdoor(threadAttrInit)(pthread_attr_t *attr)

{
    attr->stack_size = RTLTHREAD_STACKSIZE;
    attr->initial_state = RTL_THREAD_READY;
    attr->sched_param.sched_priority = minPrio;
    attr->cpu = 0;
    attr->use_fp = 0;
    attr->detachstate = 0;
}

int rtlNode::kdoor(threadMakePeriodic)(pthread_t pthread,
				       hrtime_t start_time,
				       hrtime_t period)
{
    if (!testOption(RTL_LOOSE_CALLCHECK))
	mustBeInitOrSynch();

    rtlThread *thread = findThreadByPtr(pthread);

    if (!thread)
	return ESRCH;

    // Note: the following case allows calling pthread_make_periodic_np()
    // in order to change a delayed thread's wake up time. In the real world,
    // doing this changes the resume_time property of a thread. I don't think
    // this is a good idea to depend on the RTLinux internals this way, but
    // this case is handled in an effort to be as close as possible from the
    // real world.

    if (thread->tstStatus(CKDELAY) && !thread->tstStatus(CKPEND))
	{
	hrtime_t now = getHrTime();

	if (start_time - now > getHrRes())
	    {
	    ckticks_t remTicks = ns2ticks(start_time - now);
	    thread->setTimerCount(remTicks);
	    }
	else
	    readyThread(thread,CKDELAY);
	}

    thread->setPeriodic(start_time,period);
    resched();

    return 0;
}

int rtlNode::kdoor(threadWait)()

{
    mustBeSynch();

    rtlRunningThread->schedulePeriodic();
    resched();

    return 0;
}

int rtlNode::kdoor(threadSuspend)(pthread_t pthread)

{
    if (!testOption(RTL_LOOSE_CALLCHECK))
	mustBeSynch();

    rtlThread *thread = findThreadByPtr(pthread);

    if (!thread)
	return ESRCH;

    if (!thread->tstStatus(CKSUSP))
	{
	suspendThread(thread,CKSUSP);

	if (thread == rtlRunningThread)
	    resched();
	}

    return 0;
}

int rtlNode::kdoor(threadWakeup)(pthread_t pthread)

{
    rtlThread *thread = findThreadByPtr(pthread);

    if (!thread)
	return ESRCH;

    // RTLinux does not enforce FIFO ordering inside a given
    // priority group; so don't apply round-robin when waking
    // up a non-suspended thread to leave the scheduling
    // order unchanged.

    if (thread->tstStatus(CKWAIT))
	startThread(thread);
    else if (thread->tstStatus(CKSUSP))
	{
	readyThread(thread,CKSUSP);
	resched();
	}

    return 0;
}

int rtlNode::kdoor(threadSetFp)(pthread_t pthread,
				int flag)
{
    rtlThread *thread = findThreadByPtr(pthread);

    if (!thread)
	return ESRCH;

    // No-op -- EINVAL (i.e. CPU mismatch) is currently never
    // returned.

    return 0;
}

int rtlNode::kdoor(setClockMode)(clockid_t clock,
				 int mode,
				 hrtime_t param)
{
    rtlClock *iclock = findClockByPtr(clock);

    if (!iclock)
	CkManager::This->fatal("rtl_setclockmode(clock=%p): invalid clock",
			       clock);

    switch (mode)
	{
	case RTL_CLOCK_MODE_ONESHOT:

	    iclock->setTimer(CfEventTimer,HRTIME_INFINITY);
	    break;

	case RTL_CLOCK_MODE_PERIODIC:

	    if (param <= 0)
		return -EINVAL;

	    iclock->setTimer(CfEventPeriodical,param);
	    break;

	default:

	    return -EINVAL;
	}

    return 0;
}

int rtlNode::kdoor(setClockHandler) (clockid_t clock,
				     clock_irq_handler_t fn)
{
    rtlClock *iclock = findClockByPtr(clock);

    if (!iclock)
	CkManager::This->fatal("rtl_setclockhandler(clock=%p): invalid clock",
			       clock);

    if (iclock->getTickHandler())
	return -EBUSY;

    iclock->setTickHandler(fn);

    return 0;
}

int rtlNode::kdoor(unsetClockHandler) (clockid_t clock)

{
    rtlClock *iclock = findClockByPtr(clock);

    if (!iclock)
	CkManager::This->fatal("rtl_unsetclockhandler(clock=%p): invalid clock",
			       clock);

    if (!iclock->getTickHandler())
	return -EINVAL;

    iclock->setTickHandler(NULL);

    return 0;
}

int rtlNode::kdoor(clockGetTime) (clockid_t clock,
				  struct timespec *ts)
{
    rtlClock *iclock = findClockByPtr(clock);

    if (!iclock)
	return EINVAL;

    hrtime_t t = iclock->getTime();
    *ts = timespec_from_ns(t);

    return 0;
}

int rtlNode::kdoor(clockSetTime) (clockid_t clock,
				  const struct timespec *ts)
{
    rtlClock *iclock = findClockByPtr(clock);
    hrtime_t t = timespec_to_ns(ts);

    if (!iclock || t < 0)
	return EINVAL;

    iclock->setTime(t);

    return 0;
}

int rtlNode::kdoor(clockGetRes) (clockid_t clock,
				 struct timespec *ts)
{
    rtlClock *iclock = findClockByPtr(clock);

    if (!iclock)
	return EINVAL;

    if (ts)
	*ts = timespec_from_ns(iclock->getRes());

    return 0;
}

hrtime_t rtlNode::kdoor(clockGetHrTime) (clockid_t clock)

{
    rtlClock *iclock = findClockByPtr(clock);

    if (!iclock)
	CkManager::This->fatal("clock_gethrtime(clock=%p): invalid clock",
			       clock);

    return iclock->getTime();
}

time_t rtlNode::kdoor(clockTime) ()

{
    struct timespec ts = timespec_from_ns(getHrTime());
    return ts.tv_sec;
}

int rtlNode::kdoor(usleep) (useconds_t interval)

{
    mustBeSynch();

    ckticks_t ticks = ns2ticks(interval * 1000);

    if (ticks > 0)
	{
	suspendThread(rtlRunningThread,CKDELAY,ticks);
	resched();
	}

    return 0;
}

int rtlNode::kdoor(nanosleep) (const struct timespec *rqtp,
			       struct timespec *rmtp)
{
    mustBeSynch();

    hrtime_t tsleep = timespec_to_ns(rqtp), tstart = getHrTime();
    ckticks_t ticks = ns2ticks(tsleep);

    if (ticks > 0)
	{
	suspendThread(rtlRunningThread,CKDELAY,ticks);
	resched();

	if (rtlRunningThread->tstStatus(CKBREAK))
	    {
	    if (rmtp)
		{
		hrtime_t now = getHrTime();

		if (now < tstart + tsleep)
		    *rmtp = timespec_from_ns(tstart + tsleep - now);
		}

	    rtlRunningThread->setErrno(EINTR);
	    return -1;
	    }
	}

    return 0;
}

int rtlNode::kdoor(mutexInit) (pthread_mutex_t *mutex,
			       const pthread_mutexattr_t *attr)
{
    int prioCeiling = -1;

    if (attr)
	{
	if (attr->protocol == PTHREAD_PRIO_PROTECT)
	    {
	    prioCeiling = attr->prioceiling;

	    if (prioCeiling < minPrio ||
		prioCeiling > maxPrio)
		return EINVAL;
	    }
	else if (attr->protocol != PTHREAD_PRIO_NONE)
	    return EINVAL;
	}

    rtlMutex *imutex = new rtlMutex(prioCeiling,this);

    mutex->handle = imutex->getCkId();

    return 0;
}

int rtlNode::kdoor(mutexDestroy) (pthread_mutex_t *mutex)

{
    rtlMutex *imutex = findMutexByPtr(mutex);

    if (!imutex)
	return EINVAL;

    if (imutex->getCount() > 0 && testOption(RTL_WARN_REMOVED_SYNCH))
	CkManager::This->warning("pthread_mutex_destroy(mutex=%p): %d threads still pending on deleted mutex",
				 mutex,
				 imutex->getCount());
    delete imutex;
    pendReschedule();

    return 0;
}

int rtlNode::kdoor(mutexLock) (pthread_mutex_t *mutex)

{
    mustBeInitOrSynch();

    rtlMutex *imutex = findMutexByPtr(mutex);

    if (!imutex)
	return EINVAL;

    int prioCeiling = imutex->getPrioCeiling();

    if (prioCeiling != -1 && testPri(rtlRunningThread->getPri(),prioCeiling) < 0)
	return EINVAL;

    int s = imutex->lock(rtlRunningThread);

    // Allow the mutex to be locked from the init routine unless
    // the caller should block to gain access.

    if (s == CK_SYNCH_WAIT)
	{
	mustBeSynch();

	if (imutex->getCount() >= maxMutexWaiters)
	    CkManager::This->warning("pthread_mutex_lock(mutex=%p): more than %d threads are already pending on this mutex",
				     mutex,
				     imutex->getCount());
	resched();

	if (rtlRunningThread->tstStatus(CKRMID) &&
	    testOption(RTL_WARN_REMOVED_SYNCH)) // Ck-specific
	    CkManager::This->warning("pthread_mutex_lock(mutex=%p): thread woken up due to mutex deletion",
				     mutex);
	}
    else if (s == CK_SYNCH_ERROR)
	     CkManager::This->warning("pthread_mutex_lock(mutex=%p): recursive lock attempted",
				      mutex);
    return 0;
}

int rtlNode::kdoor(mutexTryLock) (pthread_mutex_t *mutex)

{
    mustBeInitOrSynch();

    rtlMutex *imutex = findMutexByPtr(mutex);

    if (!imutex)
	return EINVAL;
    
    int prioCeiling = imutex->getPrioCeiling();

    if (prioCeiling != -1 && rtlRunningThread->getPri() > prioCeiling)
	return EINVAL;

    if (imutex->tryLock(rtlRunningThread) != CK_SYNCH_DONE)
	return EBUSY;

    return 0;
}

int rtlNode::kdoor(mutexUnlock) (pthread_mutex_t *mutex)

{
    rtlMutex *imutex = findMutexByPtr(mutex);

    if (!imutex)
	return EINVAL;

    int s = imutex->unlock();

    if (s == CK_SYNCH_RESCHED)
	pendReschedule();
    else if (s == CK_SYNCH_ERROR)
	CkManager::This->warning("pthread_mutex_unlock(mutex=%p): mutex already unlocked",
				 mutex);

    return 0;
}

int rtlNode::kdoor(mutexSetPrioCeiling) (pthread_mutex_t *mutex,
					 int prioceiling)
{
    rtlMutex *imutex = findMutexByPtr(mutex);

    if (!imutex)
	return EINVAL;

    int oldceiling = imutex->getPrioCeiling();

    imutex->setPrioCeiling(prioceiling);

    return oldceiling;
}

int rtlNode::kdoor(mutexGetPrioCeiling) (pthread_mutex_t *mutex)

{
    rtlMutex *imutex = findMutexByPtr(mutex);

    if (!imutex)
	return EINVAL;

    return imutex->getPrioCeiling();
}

int rtlNode::kdoor(condInit) (pthread_cond_t *cond,
			      const pthread_condattr_t *attr)
{
    rtlCondVar *icond = new rtlCondVar(this);

    cond->handle = icond->getCkId();

    return 0;
}

int rtlNode::kdoor(condDestroy) (pthread_cond_t *cond)

{
    rtlCondVar *icond = findCondVarByPtr(cond);

    if (!icond)
	return EINVAL;

    if (icond->getCount() > 0 && testOption(RTL_WARN_REMOVED_SYNCH))
	CkManager::This->warning("pthread_cond_destroy(cond=%p): %d threads still pending on deleted condition variable",
				 cond,
				 icond->getCount());
    delete icond;
    pendReschedule();

    return 0;
}

int rtlNode::kdoor(condWait) (pthread_cond_t *cond,
			      pthread_mutex_t *mutex)
{
    mustBeSynch();

    rtlCondVar *icond = findCondVarByPtr(cond);

    if (!icond)
	return EINVAL;

    int s = kdoor(mutexUnlock)(mutex);

    if (s)
	return s;

    icond->wait(rtlRunningThread,CK_INFINITE);

    resched();

    if (rtlRunningThread->tstStatus(CKRMID) &&
	testOption(RTL_WARN_REMOVED_SYNCH)) // Ck-specific
	CkManager::This->warning("pthread_cond_wait(cond=%p): thread woken up due to condition variable deletion",
				 cond);

    return kdoor(mutexLock)(mutex);
}

int rtlNode::kdoor(condSignal) (pthread_cond_t *cond)

{
    rtlCondVar *icond = findCondVarByPtr(cond);

    if (!icond)
	return EINVAL;

    if (icond->signal() == CK_SYNCH_RESCHED)
	pendReschedule();

    return 0;
}

int rtlNode::kdoor(condBroadcast) (pthread_cond_t *cond)

{
    rtlCondVar *icond = findCondVarByPtr(cond);

    if (!icond)
	return EINVAL;

    if (icond->broadcast() == CK_SYNCH_RESCHED)
	pendReschedule();

    return 0;
}

int rtlNode::kdoor(condTimedWait) (pthread_cond_t *cond,
				   pthread_mutex_t *mutex,
				   const struct timespec *abstime)
{
    mustBeSynch();

    rtlCondVar *icond = findCondVarByPtr(cond);

    if (!icond)
	return EINVAL;

    int s = kdoor(mutexUnlock)(mutex);

    if (s)
	return s;

    // Convert absolute timespec to tick count.

    int err = ETIMEDOUT;

    ckticks_t absTicks = ns2ticks(timespec_to_ns(abstime));

    if (absTicks > getHrTicks())
	{
	icond->wait(rtlRunningThread,absTicks - getHrTicks());
	resched();

	if (rtlRunningThread->tstStatus(CKRMID) &&
	    testOption(RTL_WARN_REMOVED_SYNCH)) // Ck-specific
	    CkManager::This->warning("pthread_cond_timedwait(cond=%p): thread woken up due to condition variable deletion",
				     cond);

	if (!rtlRunningThread->tstStatus(CKTIMEO))
	    err = 0;
	} // Otherwise, too late.

    s = kdoor(mutexLock)(mutex);

    return err ? err : s;
}

int rtlNode::kdoor(schedLock) (pthread_t pthread)

{
    if (!getActiveThread())
	return 0;

    if (!pthread)
	{
	getActiveThread()->unlock();
	pendReschedule();
	return 0;
	}

    rtlThread *thread = findThreadByPtr(pthread);

    if (thread && thread == getActiveThread())
	thread->lock();

    return 0;
}

int rtlNode::kdoor(ipcSemInit) (rt_sem_t *sem,
				RT_SEMTYPE type,
				int init_val)
{
    // Sanity checks have already been done in the wrapper code.
    // Jerry Epplin's implementation for RTLinux exhibits a prioritized-wait
    // of threads pending on semaphores.

    rtlSema4 *sema4 = new rtlSema4((unsigned)init_val,
				   type == RT_SEM_BINARY ? 1 : ~0,
				   getPriMode(),
				   this);

    sem->handle = sema4->getCkId();

    return 0;
}

int rtlNode::kdoor(ipcSemDestroy) (rt_sem_t *sem)

{
    rtlSema4 *sema4 = findSema4ByPtr(sem);

    if (!sema4)
	return -EINVAL;

    if (sema4->getCount() > 0 && testOption(RTL_WARN_REMOVED_SYNCH))
	CkManager::This->warning("rt_sem_destroy(sem=%p): %d threads still pending on deleted semaphore",
				 sem,
				 sema4->getCount());
    delete sema4;
    pendReschedule();

    return 0;
}

int rtlNode::kdoor(ipcSemPost) (rt_sem_t *sem)

{
    rtlSema4 *sema4 = findSema4ByPtr(sem);

    if (!sema4)
	return -EINVAL;

    int s = sema4->release();

    if (s == CK_SYNCH_RESCHED)
	// Note: According to the IPC module implementation, the rescheduling
	// must be done immediately since pthread_wakeup_np() is used
	// to wake the thread(s) up.
	resched();

    return 0;
}

int rtlNode::kdoor(ipcSemWait) (rt_sem_t *sem,
				RTIME timeout)
{
    mustBeSynch();

    rtlSema4 *sema4 = findSema4ByPtr(sem);

    if (!sema4)
	return -EINVAL;

    // Convert RT_IPC timeout values to Ck's.

    ckticks_t _timeout = 0;

    if (timeout == RT_WAIT_FOREVER)
	_timeout = CK_INFINITE;
    else if (timeout == RT_NO_WAIT)
	_timeout = CK_NONBLOCK;
    else
	{
	timeout = HRT_FROM_8254(timeout);
	hrtime_t now = getHrTime();

	if (timeout > now)
	    _timeout = ns2ticks(timeout - now);

	if (_timeout == 0)
	    _timeout = CK_NONBLOCK;
	}

    int s = sema4->acquire(rtlRunningThread,_timeout);

    if (s == CK_SYNCH_WAIT)
	{
	resched();

	if (rtlRunningThread->tstStatus(CKRMID) &&
	    testOption(RTL_WARN_REMOVED_SYNCH)) // Ck-specific
	    CkManager::This->warning("rt_sem_wait(sem=%p): thread woken up due to semaphore deletion",
				     sem);

	if (rtlRunningThread->tstStatus(CKTIMEO))
	    return -ETIME;
	}
    else if (s == CK_SYNCH_NOWAIT)
	return -EAGAIN;

    return 0;
}

int rtlNode::kdoor(ipcSemTryWait) (rt_sem_t *sem)

{
    rtlSema4 *sema4 = findSema4ByPtr(sem);

    if (!sema4)
	return -EINVAL;

    if (sema4->acquire(rtlRunningThread,CK_NONBLOCK) != CK_SYNCH_DONE)
	return -EAGAIN;		// i.e. Not available.

    return 0;
}

int rtlNode::kdoor(ipcTaskDelay) (RTIME duration)

{
    mustBeSynch();

    // rtl_schedule() (rtlinux/schedulers/rtl_sched.c) does not really care
    // about priority groups, but the code rather elects the last enqueued
    // ready task to run among multiple candidates from the highest priority
    // level. Thus it seems there is no reason to perform any manual
    // round-robin for null delays.

    duration = HRT_FROM_8254(duration);
    hrtime_t now = getHrTime();

    if (duration > now)
	{
	ckticks_t ticks = ns2ticks(duration - now);

	if (ticks > 0)
	    {
	    suspendThread(rtlRunningThread,CKDELAY,ticks);
	    resched();
	    }
	}

    return 0;
}

int rtlNode::kdoor(ipcMqInit) (rt_mq_t *mq,
			       int max_msgs,
			       int msg_size)
{
    // Sanity checks have already been done in the wrapper code.
    rtlMq *imq = new rtlMq(max_msgs,msg_size,this);

    mq->handle = imq->getCkId();

    return 0;
}

int rtlNode::kdoor(ipcMqDestroy) (rt_mq_t *mq)

{
    rtlMq *imq = findMqByPtr(mq);

    if (!imq)
	return -EINVAL;

    if (imq->getCount() > 0 && testOption(RTL_WARN_REMOVED_SYNCH))
	CkManager::This->warning("rt_mq_destroy(mq=%p): %d threads still pending on deleted mqueue",
				 mq,
				 imq->getCount());
    delete imq;
    pendReschedule();

    return 0;
}

int rtlNode::kdoor(ipcMqSend) (rt_mq_t *mq,
			       char *msg,
			       RT_MQ_PRIO prio,
			       RTIME wait)
{
    mustBeSynch();

    rtlMq *imq = findMqByPtr(mq);

    if (!imq)
	return -EINVAL;

    if (msg && imq->getMsgSize() == 0)
	CkManager::This->warning("rt_mq_send(mq=%p): sending valid message to zero-sized mqueue",
				 mq);

    if (!msg || !TEST_PARAM(msg,imq->getMsgSize(),MEMADDR_READABLE))
	CkManager::This->fatal("rt_mq_send(msg=%p): invalid memory",
			       msg);

    // Convert RT_IPC timeout values to Ck's.

    ckticks_t timeout = 0;

    if (wait == RT_WAIT_FOREVER)
	timeout = CK_INFINITE;
    else if (wait == RT_NO_WAIT)
	timeout = CK_NONBLOCK;
    else
	{
	wait = HRT_FROM_8254(wait);
	hrtime_t now = getHrTime();

	if (wait > now)
	    timeout = ns2ticks(wait - now);

	if (timeout == 0)
	    timeout = CK_NONBLOCK;
	}

    int s = imq->send(rtlRunningThread,msg,prio,timeout);

    if (s == CK_SYNCH_NOWAIT)
	return -EAGAIN;		// No space available.

    if (s != CK_SYNCH_DONE)
	{
	resched();

	if (s == CK_SYNCH_WAIT)
	    {
	    if (rtlRunningThread->tstStatus(CKTIMEO))
		return -ETIME;

	    if (rtlRunningThread->tstStatus(CKRMID) &&
		testOption(RTL_WARN_REMOVED_SYNCH)) // Ck-specific
		CkManager::This->warning("rt_mq_send(mq=%p): thread woken up due to mqueue deletion",
					 mq);
	    }
	}

    return 0;
}

int rtlNode::kdoor(ipcMqReceive) (rt_mq_t *mq,
				  char *msg,
				  RTIME wait)
{
    mustBeSynch();

    rtlMq *imq = findMqByPtr(mq);

    if (!imq)
	return -EINVAL;

    // Convert RT_IPC timeout values to Ck's.

    ckticks_t timeout = 0;

    if (wait == RT_WAIT_FOREVER)
	timeout = CK_INFINITE;
    else if (wait == RT_NO_WAIT)
	timeout = CK_NONBLOCK;
    else
	{
	wait = HRT_FROM_8254(wait);
	hrtime_t now = getHrTime();

	if (wait > now)
	    timeout = ns2ticks(wait - now);

	if (timeout == 0)
	    timeout = CK_NONBLOCK;
	}

    int s = imq->receive(rtlRunningThread,timeout);

    if (s == CK_SYNCH_NOWAIT)
	return -EAGAIN;		// No data available.

    if (s != CK_SYNCH_DONE)
	{
	resched();

	if (s == CK_SYNCH_WAIT)
	    {
	    if (rtlRunningThread->tstStatus(CKTIMEO))
		return -ETIME;

	    if (rtlRunningThread->tstStatus(CKRMID) &&
		testOption(RTL_WARN_REMOVED_SYNCH)) // Ck-specific
		CkManager::This->warning("rt_mq_receive(mq=%p): thread woken up due to mqueue deletion",
					 mq);
	    }
	}

    rtlMessage *mblock = rtlRunningThread->msgGet();

    if (mblock)
	{
	if (msg)
	    {
	    if (!TEST_PARAM(msg,imq->getMsgSize(),MEMADDR_WRITABLE))
		CkManager::This->fatal("rt_mq_receive(msg=%p): invalid memory",
				       msg);

	    if (mblock->copy)
		memcpy(msg,mblock->copy,imq->getMsgSize());
	    else if (mblock->src == NULL && imq->getMsgSize() > 0)
		CkManager::This->warning("rt_mq_receive(msg=%p): receiving NULL message",
					 msg);
	    }

	delete mblock;
	}

    return 0;
}

int rtlNode::kdoor(fifoCreate) (int fifo,
				int size) // Ignored.
{
    if (!testOption(RTL_LOOSE_CALLCHECK))
	mustBeInit();

    if (fifo < 0 || fifo >= RTF_NO)
	return -ENODEV;

    if (fifoTable[fifo])
	return -EBUSY;

    fifoTable[fifo] = new rtlFifo(fifo,this);

    return 0;
}

int rtlNode::kdoor(fifoDestroy) (int fifo)

{
    if (!testOption(RTL_LOOSE_CALLCHECK))
	mustBeInit();

    if (fifo < 0 || fifo >= RTF_NO)
	return -ENODEV;

    if (!fifoTable[fifo])
	return -EINVAL;

    delete fifoTable[fifo];
    fifoTable[fifo] = NULL;

    return 0;
}

int rtlNode::kdoor(fifoPut) (int fifo,
			     const void *buf,
			     int count)
{
    if (fifo < 0 || fifo >= RTF_NO)
	return -ENODEV;

    if (!fifoTable[fifo])
	return -EINVAL;

    return fifoTable[fifo]->put(findCurrentRunnable(),buf,count);
}

int rtlNode::kdoor(fifoGet) (int fifo,
			     void *buf,
			     int count)
{
    if (fifo < 0 || fifo >= RTF_NO)
	return -ENODEV;

    if (!fifoTable[fifo])
	return -EINVAL;

    return fifoTable[fifo]->get(findCurrentRunnable(),buf,count);
}

int rtlNode::kdoor(fifoCreateUserHandler) (int fifo,
					   int (*handler)(unsigned int fifo))
{
    if (fifo < 0 || fifo >= RTF_NO)
	return -ENODEV;

    if (!fifoTable[fifo])
	return -EINVAL;

    fifoTable[fifo]->setUserHandler(handler);

    return 0;
}

int rtlNode::kdoor(fifoCreateRtHandler) (int fifo,
					 int (*handler)(unsigned int fifo))
{
    if (fifo < 0 || fifo >= RTF_NO)
	return -ENODEV;

    if (!fifoTable[fifo])
	return -EINVAL;

    fifoTable[fifo]->setRtHandler(handler);

    return 0;
}

// Note: CarbonKernel defines a split interrupt model (ISR/DSR) exhibiting
// prioritized IRQ lines identified by (vector,level) pairs. The
// RTLinux model maps the low-level handler to ISRs, and software
// interrupt handlers to DSRs. The interrupt level stands for the IRQ
// number. The vector is always 0.

void rtlNode::kdoor(hardEnableIrq) (unsigned irq)

{
    if (irq >= RTL_IRQ_MAX)
	CkManager::This->fatal("invalid IRQ number: %u",irq);

    rtlInterrupt *interrupt = irqTable[irq];

    if (interrupt)
	interrupt->clrStatus(CK_INTR_MASKED);
}

void rtlNode::kdoor(hardDisableIrq) (unsigned irq)

{
    if (irq >= RTL_IRQ_MAX)
	CkManager::This->fatal("invalid IRQ number: %u",irq);

    rtlInterrupt *interrupt = irqTable[irq];

    if (interrupt)
	interrupt->setStatus(CK_INTR_MASKED);
}

int rtlNode::kdoor(requestGlobalIrq) (unsigned irq,
				      rtl_irq_handler_t handler)
{
    if (!testOption(RTL_LOOSE_CALLCHECK))
	mustBeInit();

    if (irq >= RTL_IRQ_MAX)
	CkManager::This->fatal("invalid IRQ number: %u",irq);

    rtlInterrupt *interrupt = irqTable[irq];

    if (interrupt)
	{
	if (interrupt->getHardHandler())
	    return -EBUSY;

	interrupt->setHardHandler(handler);
	}

    return 0;
}

int rtlNode::kdoor(freeGlobalIrq) (unsigned irq)

{
    if (irq >= RTL_IRQ_MAX)
	CkManager::This->fatal("invalid IRQ number: %u",irq);

    rtlInterrupt *interrupt = irqTable[irq];

    if (interrupt)
	{
	if (!interrupt->getHardHandler())
	    return -EINVAL;

	interrupt->setHardHandler(NULL);
	interrupt->setStatus(CK_INTR_MASKED);
	}

    return 0;
}

// Note: Software interrupts are simulated by regular h/w interrupt objects
// with a DSR to execute the user-defined handler but with no ISR.

int rtlNode::kdoor(getSoftIrq) (rtl_soft_irq_handler_t handler,
				const char *devname)
{
    unsigned irq;

    for (irq = 0; irq < RTL_IRQ_MAX; irq++)
	{
	if (!irqTable[irq])
	    break;
	}

    if (irq >= RTL_IRQ_MAX)
	return -EBUSY;

    rtlInterrupt *interrupt = new rtlInterrupt(irq,
					       NULL,
					       NULL,
					       devname,
					       NULL,
					       this);
    interrupt->setSoftHandler(handler);

    return (int)irq;
}

void rtlNode::kdoor(freeSoftIrq) (unsigned irq)

{
    if (irq >= RTL_IRQ_MAX)
	CkManager::This->fatal("invalid IRQ number: %u",irq);

    rtlInterrupt *interrupt = irqTable[irq];

    if (interrupt && interrupt->getSoftHandler())
	{
	irqTable[irq] = NULL;
	getInterruptManager()->destroyInterrupt(interrupt);
	}
}

void rtlNode::kdoor(globalPendIrq) (unsigned irq)

{
    if (irq >= RTL_IRQ_MAX)
	CkManager::This->fatal("invalid IRQ number: %u",irq);

    rtlInterrupt *interrupt = irqTable[irq];

    if (interrupt)
	getInterruptManager()->postIrq(interrupt);
}

int rtlNode::kdoor(stdSemInit) (sem_t *sem,
				unsigned init_val)
{
    // Sanity checks have already been done in the wrapper code.
    // Standard implementation for RTLinux exhibits a FIFO-ordering
    // of threads pending on semaphores.

    rtlSema4 *sema4 = new rtlSema4(init_val,~0,FIFO,this);

    sem->handle = sema4->getCkId();

    return 0;
}

int rtlNode::kdoor(stdSemDestroy) (sem_t *sem)

{
    rtlSema4 *sema4 = findSema4ByPtr(sem);

    if (!sema4)
	return EINVAL;

    if (sema4->getCount() > 0 && testOption(RTL_WARN_REMOVED_SYNCH))
	CkManager::This->warning("sem_destroy(sem=%p): %d threads still pending on deleted semaphore",
				 sem,
				 sema4->getCount());
    delete sema4;
    pendReschedule();

    return 0;
}

int rtlNode::kdoor(stdSemPost) (sem_t *sem)

{
    rtlSema4 *sema4 = findSema4ByPtr(sem);

    if (!sema4)
	return EINVAL;

    int s = sema4->release();

    if (s == CK_SYNCH_RESCHED)
	resched();

    return 0;
}

int rtlNode::kdoor(stdSemWait) (sem_t *sem)

{
    mustBeSynch();

    rtlSema4 *sema4 = findSema4ByPtr(sem);

    if (!sema4)
	return EINVAL;

    if (sema4->acquire(rtlRunningThread,CK_INFINITE) == CK_SYNCH_WAIT)
	{
	resched();

	if (rtlRunningThread->tstStatus(CKRMID) &&
	    testOption(RTL_WARN_REMOVED_SYNCH)) // Ck-specific
	    CkManager::This->warning("sem_wait(sem=%p): thread woken up due to semaphore deletion",
				     sema4);
	}

    return 0;
}

int rtlNode::kdoor(stdSemTimedWait) (sem_t *sem, const struct timespec *abstime)

{
    mustBeSynch();

    rtlSema4 *sema4 = findSema4ByPtr(sem);

    if (!sema4)
	return EINVAL;

    // Convert absolute timespec to tick count.

    int err = ETIMEDOUT;

    ckticks_t absTicks = ns2ticks(timespec_to_ns(abstime));

    if (absTicks > getHrTicks())
	{
	if (sema4->acquire(rtlRunningThread,absTicks - getHrTicks()) == CK_SYNCH_WAIT)
	    {
	    resched();

	    if (rtlRunningThread->tstStatus(CKRMID) &&
		testOption(RTL_WARN_REMOVED_SYNCH)) // Ck-specific
		CkManager::This->warning("sem_timedwait(sem=%p): thread woken up due to semaphore deletion",
					 sema4);

	    if (!rtlRunningThread->tstStatus(CKTIMEO))
		err = 0;
	    }
	else
	    err = 0;
	} // Otherwise, too late.

    return err;
}

int rtlNode::kdoor(stdSemTryWait) (sem_t *sem)

{
    rtlSema4 *sema4 = findSema4ByPtr(sem);

    if (!sema4)
	return EINVAL;

    if (sema4->acquire(rtlRunningThread,CK_NONBLOCK) != CK_SYNCH_DONE)
	return EAGAIN;		// i.e. Not available.

    return 0;
}

unsigned rtlNode::kdoor(stdSemGetValue) (sem_t *sem)

{
    rtlSema4 *sema4 = findSema4ByPtr(sem);

    if (!sema4)
	return EINVAL;

    return sema4->peek();
}

void rtlNode::signalLinux (int flags)

{
    setStatus(flags);
    linuxThread->post();
}

rtlNodeExportMessage::rtlNodeExportMessage (rtlNode *node) :
    CkNodeExportMessage(node->getName())
{
    nid = node->getNid();
    scopy(osName,node->getOS(),sizeof(osName) - 1);
    strcpy(threadTypeName,"Thread");
    asciiXID = False;
    nThreadStates = 7;
    strcpy(threadStates[0].stateLabel,"Halted");
    threadStates[0].stateHandle = rtlThreadHALTED;
    strcpy(threadStates[1].stateLabel,"Dormant");
    threadStates[1].stateHandle = rtlThreadDORMANT;
    strcpy(threadStates[2].stateLabel,"Delayed");
    threadStates[2].stateHandle = rtlThreadDELAYED;
    strcpy(threadStates[3].stateLabel,"Sleeping");
    threadStates[3].stateHandle = rtlThreadSLEEPING;
    strcpy(threadStates[4].stateLabel,"Suspended");
    threadStates[4].stateHandle = rtlThreadSUSPENDED;
    strcpy(threadStates[5].stateLabel,"Ready");
    threadStates[5].stateHandle = rtlThreadREADY;
    strcpy(threadStates[6].stateLabel,"Active");
    threadStates[6].stateHandle = rtlThreadACTIVE;
}
