/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Implementation of the RTLinux kernel synchronization
 * objects. Defines objects from the IPC module written by
 * by Jerry Epplin for the RTLinux system.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation
#endif // __GNUG__
#include <config.h>
#include <stdlib.h>
#include "CkManager.h"
#include "CkSynch.h"
#include "rtlSynch.h"
#include "rtlNode.h"
#include "rtlThread.h"

rtlMessageList rtlMq::messageTank;

rtlSynch::rtlSynch (const char *_name,
		    InsertMode _mode,
		    CkStateExportObject *_synchState,
		    rtlNode *_node) :
    CkSynchro(_name,
	      _mode,
	      _synchState,
	      _node) {
}

void rtlSynch::blockThread (rtlThread *thread, ckticks_t timeout)

{
    insert(thread,thread->getPri());
    getNode()->suspendThread(thread,CKPEND,timeout,this);
}

void rtlSynch::unblockThread (rtlThread *thread) {

    getNode()->readyThread(thread,CKPEND);
}

void rtlSynch::forget (CkThread *thread) {

    CkSynchro::forget(thread);
}

// rtlSema4 -- Counting or binary semaphore.

const char *rtlSema4::stateArray[3] = {
    "IDLE", "PENDED", "POSTED"
};

rtlSema4::rtlSema4 (unsigned _count,
		    unsigned _ceiling,
		    InsertMode _mode,
		    rtlNode *_node) :
    rtlSynch(NULL,
	     _mode,
	     NULL,
	     _node)
{
    count = _count;
    ceiling = _ceiling;

    _node->nameObject(this,RTLINUX_SEMA4_INODE);
    protoSetName(CString().format("sema4#%.3X",getCkId().getSeq()));
    protoSetExportable();
    synchState = new CkStateExportObject(_node,
					 protoGetName(),
					 "Semaphores",
					 stateArray,3);
    synchState->setState(_count > 0 ? rtlSema4POSTED : rtlSema4IDLE);
    protoInit();
}

rtlSema4::~rtlSema4 ()

{
    rtlThread *thread;
    
    while ((thread = (rtlThread *)get()) != NULL)
	{
	thread->setStatus(CKRMID);
	getNode()->readyThread(thread,CKPEND);
	}

    getNode()->unnameObject(this);
}

void rtlSema4::protoInit ()

{
    // First, export sema4 as a node-based object
    CkObjectExportMessage em(getNode()->getNid(),
			     this,
			     getPlugInfo("rtl","rtlSema4"));
    protoExport(&em,sizeof(em));

    synchState->protoInit();
}

const char *rtlSema4::getTypedName ()

{
    CkSynchro::typedName.set("sema4");
    CkSynchro::typedName.append(protoGetName());
    return CkSynchro::typedName;
}

void rtlSema4::dynamicExpose ()

{
    TclList __tkl, _tkl, tkl;

    _tkl.set("waiters");
    buildPendList(__tkl);
    _tkl.append(__tkl);
    tkl.append(_tkl);

    _tkl.set("value");
    _tkl.append(count);
    
    tkl.append(_tkl);
    
    protoInfo(SXPROTO_OBJECT_INFO,tkl.get(),tkl.length());
}

void rtlSema4::dynamicConfigure (StProtoInfoMessage *mbuf)

{
    TclListParser parser(mbuf->data);
    const char *setting;

    while ((setting = parser.next()) != NULL)
	{
	if (strcmp(setting,"value") == 0)
	    {
	    unsigned newCount = (unsigned)atoi(parser.next());

	    if (newCount > count)
		{
		for (unsigned relCount = newCount - count;
		     relCount > 0; relCount--)
		    release();
		}
	    else
		count = newCount;
	    }
	}
}

int rtlSema4::acquire (rtlThread *thread, ckticks_t timeout)

{
    if (count > 0)
	{
	if (--count == 0)
	    synchState->setState(rtlSema4IDLE);

	return CK_SYNCH_DONE;
	}

    if (timeout == CK_NONBLOCK)
	return CK_SYNCH_NOWAIT;

    blockThread(thread,timeout);
    synchState->setState(rtlSema4PENDED);

    return CK_SYNCH_WAIT;
}

int rtlSema4::release ()

{
    rtlThread *thread = (rtlThread *)get();

    if (thread)
	{
	unblockThread(thread);

	if (getCount() == 0)
	    synchState->setState(rtlSema4IDLE);

	return CK_SYNCH_RESCHED;
	}

    if (count < ceiling)
	{
	count++;
	synchState->setState(rtlSema4POSTED);
	}
    else if (ceiling > 1)
	CkManager::This->warning("counting semaphore count %s would overflow (value=%u)",
				 protoGetName(),
				 count);

    return CK_SYNCH_DONE;
}

void rtlSema4::forget (CkThread *thread)

{
    rtlSynch::forget(thread);

    if (getCount() == 0)
	synchState->setState(rtlSema4IDLE);
}

// rtlMutex -- Mutex. Enforces priority ceiling upon request.

const char *rtlMutex::stateArray[3] = {
    "FREE", "LOCKED", "PENDED"
};

rtlMutex::rtlMutex (int _prioCeiling,
		    rtlNode *_node) :
    rtlSynch(NULL,
	     FIFO,
	     NULL,
	     _node)
{
    prioCeiling = _prioCeiling;	// -1 if no priority ceiling
    owner = NULL;
    _node->nameObject(this,RTLINUX_MUTEX_INODE);
    protoSetName(CString().format("mutex#%.3X",getCkId().getSeq()));
    protoSetExportable();
    synchState = new CkStateExportObject(_node,
					 protoGetName(),
					 "Mutexes",
					 stateArray,3);
    synchState->setState(rtlMutexFREE);
    protoInit();
}

rtlMutex::~rtlMutex ()

{
    rtlThread *thread;
    
    while ((thread = (rtlThread *)get()) != NULL)
	{
	thread->setStatus(CKRMID);
	getNode()->readyThread(thread,CKPEND);
	}

    getNode()->unnameObject(this);
}

void rtlMutex::protoInit ()

{
    // First, export mutex as a node-based object
    CkObjectExportMessage em(getNode()->getNid(),
			     this,
			     getPlugInfo("rtl","rtlMutex"));
    protoExport(&em,sizeof(em));

    synchState->protoInit();
}

void rtlMutex::dynamicExpose ()

{
    TclList __tclist, _tclist, tclist;

    _tclist.set("waiters");
    buildPendList(__tclist);
    _tclist.append(__tclist);
    tclist.append(_tclist);

    _tclist.set("state");

    if (synchState->getState() == rtlMutexFREE)
	_tclist.append("FREE");
    else
	_tclist.append("LOCKED");

    tclist.append(_tclist);
    
    _tclist.set("owner");
    _tclist.append(owner ? owner->protoGetName() : "(none)");
    tclist.append(_tclist);

    protoInfo(SXPROTO_OBJECT_INFO,tclist.get(),tclist.length());
}

void rtlMutex::dynamicConfigure (StProtoInfoMessage *mbuf)

{
    TclListParser parser(mbuf->data);
    const char *setting;

    while ((setting = parser.next()) != NULL)
	{
	if (strcmp(setting,"state") == 0)
	    {
	    const char *newState = parser.next();

	    if (!strcmp(newState,"FREE"))
		unlock();
	    else
		synchState->setState(rtlMutexLOCKED);
	    }
	}
}

const char *rtlMutex::getTypedName ()

{
    CkSynchro::typedName.set("mutex");
    CkSynchro::typedName.append(protoGetName());
    return CkSynchro::typedName;
}

int rtlMutex::lock (rtlThread *thread)

{
    if (synchState->getState() == rtlMutexFREE)
	{
	if (prioCeiling != -1)
	    pipDeclare(thread);

	owner = thread;
	synchState->setState(rtlMutexLOCKED);
	return CK_SYNCH_DONE;
	}
    else if (thread == owner)
	return CK_SYNCH_ERROR;	// Recursive lock.

    pipApply(thread);
    blockThread(thread,CK_INFINITE);
    synchState->setState(rtlMutexPENDED);

    return CK_SYNCH_WAIT;
}

int rtlMutex::tryLock (rtlThread *thread)

{
    if (synchState->getState() == rtlMutexFREE)
	{
	owner = thread;
	synchState->setState(rtlMutexLOCKED);
	return CK_SYNCH_DONE;
	}

    return CK_SYNCH_NOWAIT;
}

int rtlMutex::unlock ()

{
    if (synchState->getState() == rtlMutexFREE)
	return CK_SYNCH_ERROR;

    int s = pipExit();	// Returns CK_SYNCH_DONE if no prioCeiling

    owner = (rtlThread *)get();

    if (owner)
	{
	unblockThread(owner);

	if (getCount() == 0)
	    synchState->setState(rtlMutexLOCKED);

	return CK_SYNCH_RESCHED;
	}

    synchState->setState(rtlMutexFREE);

    return s;
}

void rtlMutex::forget (CkThread *thread)

{
    rtlSynch::forget(thread);

    if (getCount() == 0)
	synchState->setState(rtlMutexLOCKED);
}

const char *rtlMq::stateArray[4] = {
    "EMPTY", "PENDED", "POSTED", "FULL"
};

rtlMq::rtlMq (int _maxMsg,
	      int _msgSize,
	      rtlNode *_node) :
    rtlSynch(NULL,
	     _node->getPriMode(),
	     NULL,
	     _node)
{
    // The queue enters full state after one post if msgSize == 0.
    maxMsg = _msgSize > 0 ? _maxMsg : 1;
    msgSize = _msgSize;
    sendBlockCount = 0;

    _node->nameObject(this,RTLINUX_MESGQ_INODE);
    protoSetName(CString().format("mq#%.3X",getCkId().getSeq()));
    protoSetExportable();
    synchState = new CkStateExportObject(_node,
					 protoGetName(),
					 "MsgQueues",
					 stateArray,4);
    synchState->setState(rtlMqEMPTY);

    protoInit();
}

rtlMq::~rtlMq ()

{
    // Wake up all pending threads.

    rtlThread *thread;
    
    while ((thread = (rtlThread *)get()) != NULL)
	{
	thread->setStatus(CKRMID);
	getNode()->readyThread(thread,CKPEND);
	}

    storage.destroy();

    getNode()->unnameObject(this);
}

void rtlMq::protoInit ()

{
    CkObjectExportMessage em(getNode()->getNid(),
			     this,
			     getPlugInfo("rtl",
					 "rtlMq",
					 protoGetName()));
    protoExport(&em,sizeof(em));

    synchState->protoInit();
}

const char *rtlMq::getTypedName ()

{
    CkSynchro::typedName.set("mq");
    CkSynchro::typedName.append(protoGetName());
    return CkSynchro::typedName;
}

void rtlMq::dynamicExpose ()

{
    TclList __tclist, _tclist, tclist;

    _tclist.set("waiters");
    buildPendList(__tclist);
    _tclist.append(__tclist);
    tclist.append(_tclist);

    _tclist.set("messages");
    __tclist.clear();

    for (rtlMessage *msg = storage.first(); msg;
	 msg = (rtlMessage *)msg->next())
	{
	CString m((void *)msg->src);
	m += ", ";
	m += msg->time.format(NULL);
	if (msg->prio == RT_MQ_URGENT)
	    m += (" (URGENT)"); // Debugging help.
	__tclist.append(m);
	}

    _tclist.append(__tclist);
    tclist.append(_tclist);
    
    protoInfo(SXPROTO_OBJECT_INFO,tclist.get(),tclist.length());
}

void rtlMq::dynamicConfigure (StProtoInfoMessage *mbuf) {
}

int rtlMq::send (rtlThread *sender,
		 char *ptr,
		 RT_MQ_PRIO prio,
		 ckticks_t timeout)
{
    if (storage.getCount() < (unsigned)maxMsg)
	{
	// Enough space to send without waiting

	rtlMessage *msg = new rtlMessage(ptr,msgSize,prio);
	rtlThreadIterator it(this);
	rtlThread *thread;

	while ((thread = it.next()) != NULL)
	    {
	    // Search for a thread pending on the receive side of this queue.
	    if (!thread->tstStatus(RTL_PEND_SENDSIDE))
		{
		it.pop();
		break;
		}
	    }

	if (thread)
	    {
	    // A thread currently waits for a message -- pass it this
	    // message and wake it up.
	    thread->msgSet(msg);
	    unblockThread(thread);

	    if (getCount() == 0) // pended => empty?
		synchState->setState(rtlMqEMPTY);

	    return CK_SYNCH_RESCHED;
	    }

	// Nobody waits for this message -- just enqueue it according
	// to the requested priority.

	if (prio == RT_MQ_URGENT)
	    storage.prepend(msg);
	else
	    storage.append(msg);

	if (storage.getCount() == 1) // empty => posted?
	    synchState->setState(rtlMqPOSTED);

	return CK_SYNCH_DONE;
	}

    if (timeout == CK_NONBLOCK)
	return CK_SYNCH_NOWAIT;

    if (sendBlockCount++ == 0)
	synchState->setState(rtlMqFULL);

    sender->setStatus(RTL_PEND_SENDSIDE);
    sender->msgSet(new rtlMessage(ptr,msgSize,prio));
    blockThread(sender,timeout);

    return CK_SYNCH_WAIT;
}

int rtlMq::receive (rtlThread *receiver,
		    ckticks_t timeout)
{
    rtlMessage *msg = storage.get();

    if (msg)
	{
	receiver->msgSet(msg);

	if (sendBlockCount > 0)
	    {
	    rtlThreadIterator it(this);
	    rtlThread *thread;

	    while ((thread = (rtlThread *)it.next()) != NULL)
		{
		// Search for a thread waiting on the send-side

		if (thread->tstStatus(RTL_PEND_SENDSIDE))
		    {
		    it.pop();
		    break;
		    }
		}

	    // Transfer the message waiting on the send-side to
	    // the receive-side of the queue, then release the sender.

	    msg = thread->msgGet();

	    if (msg->prio == RT_MQ_URGENT)
		storage.prepend(msg);
	    else
		storage.append(msg);

	    thread->clrStatus(RTL_PEND_SENDSIDE);
	    unblockThread(thread);

	    if (--sendBlockCount == 0 && storage.getCount() > 0) // full => posted?
		synchState->setState(rtlMqPOSTED);

	    return CK_SYNCH_RESCHED;
	    }
	else if (storage.getCount() == 0) // posted => empty?
	    synchState->setState(rtlMqEMPTY);

	return CK_SYNCH_DONE;
	}

    receiver->msgSet(NULL);

    if (timeout == CK_NONBLOCK)
	return CK_SYNCH_NOWAIT;

    blockThread(receiver,timeout);

    if (getCount() == 0) // empty => pended?
	synchState->setState(rtlMqPENDED);

    return CK_SYNCH_WAIT;
}

void rtlMq::forget (CkThread *thread)

{
    CkSynchro::forget(thread);

    if (thread->tstStatus(RTL_PEND_SENDSIDE))
	{
	delete ((rtlThread *)thread)->msgGet();
	((rtlThread *)thread)->msgSet(NULL);
	thread->clrStatus(RTL_PEND_SENDSIDE);
	sendBlockCount--;
	}

    if (getCount() == 0) // pended => empty?
	synchState->setState(rtlMqEMPTY);
}

void *rtlMessage::operator new (size_t sz)

{
    register void *ptr = NULL;

    if (sz <= sizeof(rtlMessage))
	ptr = rtlMq::messageTank.get();

    return ptr ? ptr : new char[sz];
}

void rtlMessage::operator delete (void *ptr)

{
    register rtlMessage *m = (rtlMessage *)ptr;
    rtlMq::messageTank.append(m);
}

const char *rtlCondVar::stateArray[4] = {
    "IDLE", "PENDED", "SIGNALED", "BROADCAST"
};

rtlCondVar::rtlCondVar (rtlNode *_node) :
    rtlSynch(NULL,
	     FIFO,
	     NULL,
	     _node)
{
    _node->nameObject(this,RTLINUX_CONDV_INODE);
    protoSetName(CString().format("cond#%.3X",getCkId().getSeq()));
    protoSetExportable();
    synchState = new CkStateExportObject(_node,
					 protoGetName(),
					 "CondVars",
					 stateArray,4);
    synchState->setState(rtlCondVarIDLE);
    protoInit();
}

rtlCondVar::~rtlCondVar ()

{
    rtlThread *thread;
    
    while ((thread = (rtlThread *)get()) != NULL)
	{
	thread->setStatus(CKRMID);
	getNode()->readyThread(thread,CKPEND);
	}

    getNode()->unnameObject(this);
}

void rtlCondVar::protoInit ()

{
    CkObjectExportMessage em(getNode()->getNid(),
			     this,
			     getPlugInfo("rtl","rtlCondVar"));
    protoExport(&em,sizeof(em));

    synchState->protoInit();
}

void rtlCondVar::dynamicExpose ()

{
    TclList __tclist, _tclist, tclist;

    _tclist.set("waiters");
    buildPendList(__tclist);
    _tclist.append(__tclist);
    tclist.append(_tclist);

    _tclist.set("state");

    // Only report non-transient states
    // (i.e. all but SIGNALED and BROADCAST).

    if (getCount() == 0)
	_tclist.append("IDLE");
    else
	_tclist.append("PENDED");

    tclist.append(_tclist);
    
    protoInfo(SXPROTO_OBJECT_INFO,tclist.get(),tclist.length());
}

void rtlCondVar::dynamicConfigure (StProtoInfoMessage *mbuf)

{
    TclListParser parser(mbuf->data);
    const char *setting;

    while ((setting = parser.next()) != NULL)
	{
	if (strcmp(setting,"state") == 0)
	    {
	    const char *cmd = parser.next();

	    if (!strcmp(cmd,"SIGNALED"))
		signal();
	    else if (!strcmp(cmd,"BROADCAST"))
		broadcast();

	    // Ignore requests to switch to IDLE or PENDED states.
	    }
	}
}

const char *rtlCondVar::getTypedName ()

{
    CkSynchro::typedName.set("condvar");
    CkSynchro::typedName.append(protoGetName());
    return CkSynchro::typedName;
}

void rtlCondVar::wait (rtlThread *thread,
		       ckticks_t timeout)
{
    if (getCount() == 0)
	synchState->setState(rtlCondVarPENDED);

    blockThread(thread,timeout);
}

int rtlCondVar::signal () {

    // RTLinux behaviour.
    return broadcast();
}

int rtlCondVar::broadcast ()

{
    int nwakes = getCount();
    rtlThread *thread;

    synchState->setState(rtlCondVarBRD);

    while ((thread = (rtlThread *)get()) != NULL)
	unblockThread(thread);

    synchState->setState(rtlCondVarIDLE);

    return nwakes > 0 ? CK_SYNCH_RESCHED : CK_SYNCH_DONE;
}

void rtlCondVar::forget (CkThread *thread)

{
    rtlSynch::forget(thread);

    if (getCount() == 0)
	synchState->setState(rtlCondVarIDLE);
}
