/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Implementation of the RTLinux thread class.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation
#endif // __GNUG__
#include <config.h>
#include <stdlib.h>
#include "CkManager.h"
#include "rtlNode.h"
#include "rtlThread.h"
#include "rtlFifo.h"
#include "rtlSynch.h"

rtlThreadGList allRtlThreads;

const char *rtlThread::stateArray[7] = {
    "HALTED",
    "DORMANT",
    "DELAYED",
    "SLEEPING",
    "SUSPENDED",
    "READY",
    "ACTIVE"
};

rtlThread::rtlThread (int _priority,
		      void *(*_startAddr)(void *arg),
		      void *_startArg,
		      size_t _stackSize,
		      pthread_t _userHandle,
		      const char *_name,
		      rtlNode *_node) :
    CkThread(_node,
	     NULL,
	     _priority,
	     NULL,
	     Min(_stackSize,RTLTHREAD_STACKSIZE)),
    joinMutex(NULL,_node),
    joinCv(NULL,_node)
{
    startAddr = _startAddr;
    startArg = _startArg;
    stackSize = _stackSize;
    userHandle = _userHandle;
    lastStateChangeTime = 0;
    joiningThread = NULL;
    period = 0;
    periodStart = HRTIME_INFINITY;
    cpuAffinity = 0;
    setStatus(RTL_CANCEL_ENABLED|RTL_CANCEL_DEFERRED|RTL_THREAD_JOINABLE);

    // Ensure that APEXproc::startAsr() is called when a signal
    // is sent to the thread.
    catchAsr(CK_INTERNAL_ASR,0,0);

    _node->nameObject(this,RTLINUX_THREAD_INODE);
    CString threadName(_name);

    if (threadName.isEmpty())
	threadName.format("rt-thread#%.3X",getCkId().getSeq());

    tstate = new CkThreadState(threadName,
			       "Threads",
			       rtlThread::stateArray,7,
			       _node,
			       this),
    protoSetName(threadName);
    protoSetExportable();

    stateDistribs[0] = new CkHistogram(threadName,"Delayed state",getNode());
    stateDistribs[0]->protoInit();
    stateDistribs[1] = new CkHistogram(threadName,"Sleeping state",getNode());
    stateDistribs[1]->protoInit();
    stateDistribs[2] = new CkHistogram(threadName,"Suspended state",getNode());
    stateDistribs[2]->protoInit();
    stateDistribs[3] = new CkHistogram(threadName,"Ready state",getNode());
    stateDistribs[3]->protoInit();
    stateDistribs[4] = new CkHistogram(threadName,"Active state",getNode());
    stateDistribs[4]->protoInit();

    allRtlThreads.append(this);
}

rtlThread::~rtlThread ()

{
    delete userHandle;
    allRtlThreads.remove(this);
    getNode()->unnameObject(this);
}

void rtlThread::cleanup ()

{
    struct rtl_cleanup_struct *cleanup = userHandle->cleanup;

    while (cleanup)
	{
	cleanup->routine(cleanup->arg);
	cleanup = cleanup->next;
	}
}

void rtlThread::body () {

    kroot(run)();
}

void rtlThread::kroot(run) ()

{
    getNode()->reinstateThread(this);
    void *exit_val = startAddr(startArg);
    getNode()->kdoor(threadExit)(exit_val);
}

void rtlThread::startAsr (cksigmask_t sigs)

{
    // startAsr() is called by the VRTOS on behalf of the killed
    // thread and stands for the low-level signal handler.
    // We use it to handle default signal cases.
    // RTL_SIGNAL_WAKEUP leads to a no-op at this point since
    // the VRTOS has obviously done the job.

    if (sigs & (1 << (RTL_SIGNAL_DELETE - 1)))
	{
	if (!tstStatus(RTL_CANCEL_DEFERRED))
	    getNode()->kdoor(threadExit)(PTHREAD_CANCELED);
	else
	    setStatus(RTL_CANCEL_PENDING);
	}
    else if (sigs & (1 << (RTL_SIGNAL_SUSPEND - 1)))
	getNode()->kdoor(threadSuspend)(userHandle);
}

void rtlThread::logThreadState (rtlThreadStates s)

{
    int olds = getThreadState();

    ITime dtLap = SxClock - lastStateChangeTime;

    if (olds >= 2 && olds <= 6)
	// <!> Depends on rtlThreadStates enum order.
	stateDistribs[olds - 2]->add(dtLap);

    lastStateChangeTime = SxClock;

    setThreadState(s);
}

void rtlThread::setReady () {

    logThreadState(rtlThreadREADY);
}

void rtlThread::setSuspended () {

    logThreadState(rtlThreadSUSPENDED);
}

void rtlThread::setPending () {

    logThreadState(rtlThreadSLEEPING);
}

void rtlThread::setRunning () {

    logThreadState(rtlThreadACTIVE);
}

void rtlThread::setDelayed () {

    logThreadState(rtlThreadDELAYED);
}

void rtlThread::setHalted () {

    logThreadState(rtlThreadHALTED);
}

void rtlThread::setDormant () {

    logThreadState(rtlThreadDORMANT);
}

void (*rtlThread::getStartAddr ())(...) {

    return (void (*)(...))startAddr;
}

void rtlThread::protoInit ()

{
    tstate->protoInit();
    CkThreadExportMessage tem(this,"rtl","rtlThread");
    protoExport(&tem,sizeof(tem));
}

void rtlThread::dynamicExpose ()

{
    TclList __tclist, _tclist, tclist;

    _tclist.set("prio");
    _tclist.append(getPri());
    tclist.append(_tclist);

    _tclist.set("ilevel");
    _tclist.append(getThreadIL());
    tclist.append(_tclist);

    _tclist.set("signals");
    __tclist.clear();
    cksigmask_t sigmask = getSignals();
    __tclist.append("DELETE");
    __tclist.append((int)(!!(sigmask & (1 << (RTL_SIGNAL_DELETE - 1))) ||
			  tstStatus(RTL_CANCEL_PENDING)));
    __tclist.append("WAKEUP");
    __tclist.append((int)(!!(sigmask & (1 << (RTL_SIGNAL_WAKEUP - 1)))));
    __tclist.append("SUSPEND");
    __tclist.append((int)(!!(sigmask & (1 << (RTL_SIGNAL_SUSPEND - 1)))));
    _tclist.append(__tclist);
    tclist.append(_tclist);

    _tclist.set("state");
    _tclist.append(stateArray[getThreadState()]);
    tclist.append(_tclist);

    if (tstStatus(CKPEND))
	{
	// Update the "wait channel" information using the
	// first pended synchronization object.
	_tclist.set("wchan");
	// If the thread is pending on a resource, add its name.
	_tclist.append(getPendedSynchro()->first()->getTypedName());
	tclist.append(_tclist);
	}
    
    if (tstStatus(CKDELAY))
	{
	_tclist.set("timeout");
	_tclist.append(getTimerCount());
	tclist.append(_tclist);
	}

    protoInfo(SXPROTO_OBJECT_INFO,tclist.get(),tclist.length());
}

void rtlThread::dynamicConfigure (StProtoInfoMessage *mbuf)

{
    TclListParser parser(mbuf->data); // ProtoInfo msg are null-terminated.
    const char *setting;

    while ((setting = parser.next()) != NULL)
	{
	if (strcmp(setting,"prio") == 0)
	    {
	    int prio = atoi(parser.next());

	    if (prio >= 0)
		getNode()->reniceThread(this,prio);
	    else
		CkManager::This->warning("RTLinux thread priority can't be negative");
	    }
	else if (strcmp(setting,"ilevel") == 0)
	    {
	    int ilevel = atoi(parser.next());
	    setThreadIL(ilevel);
	    }
	else if (strcmp(setting,"signals") == 0)
	    {
	    TclListParser _parser(parser.next());
	    const char *siginfo;

	    while ((siginfo = _parser.next()) != NULL)
		{
		TclListParser __parser(siginfo);
		const char *signame = __parser.next();
		int signo = RTL_SIGNAL_NULL;
		
		if (!strcmp(signame,"DELETE"))
		    signo = RTL_SIGNAL_DELETE;
		else if (!strcmp(signame,"WAKEUP"))
		    signo = RTL_SIGNAL_WAKEUP;
		else if (!strcmp(signame,"SUSPEND"))
		    signo = RTL_SIGNAL_SUSPEND;

		if (signo != RTL_SIGNAL_NULL && atoi(__parser.next()))
		    getNode()->kdoor(threadKill)(userHandle,signo);
		}
	    }
	}
}

void rtlThread::signalJoin (void *exitValue)

{
    joinMutex.lock(this,0);

    rtlThreadIterator it(joinCv);
    rtlThread *thread;

    while ((thread = it.next()) != NULL)
	thread->wakeUpArgs.joinValue = exitValue;

    joinCv.broadcast();
    joinMutex.unlock(this);
}

void *rtlThread::waitJoin (rtlThread *thread)

{
    thread->joinMutex.lock(this,0);
    thread->joinCv.wait(this,&thread->joinMutex,CK_INFINITE);
    // Note: the mutex is not unlocked since the owning thread does
    // not exist anymore.
    return wakeUpArgs.joinValue;
}

// rtlThread::setPeriodic() -- Make the thread periodic (or not).
// It is assumed that "period" has already been checked for sanity
// (i.e. >= 0). If zero, the thread is aperiodic.

void rtlThread::setPeriodic (hrtime_t _periodStart,
			     hrtime_t _period)
{
    periodStart = _periodStart;
    period = _period;
}

// rtlThread::schedulePeriodic() computes the next release point
// in the processor time line for a periodic thread and puts the
// thread to sleep until this point. The caller must reschedule.
// It should be noted that checking for deadline misses is quite
// outside Ck's scope since since it does not charge for execution
// time in the same way the target would do it.
// Anyway, this may help discovering spurious infinite loops.
// If you have unexpected warnings, just uncheck the deadline
// control option from the configuration panel, or raise the
// target Warp factor.

void rtlThread::schedulePeriodic ()

{
    hrtime_t now = getNode()->getHrTime();
    hrtime_t res = getNode()->getHrRes();

    if (periodStart - now > res)
	{
	ckticks_t sleepTicks = getNode()->ns2ticks(periodStart - now);
	getNode()->suspendThread(this,CKDELAY,sleepTicks);
	now = periodStart;
	}
    else if (now > periodStart &&
	     getNode()->testOption(RTL_WARN_MISSED_DEADLINE))
	CkManager::This->warning("last periodic deadline missed");

    // Compute the starting date of the next period.

    if (period > 0)
	{
	while (periodStart <= now)
	    periodStart += period;
	}
    else
	periodStart = HRTIME_INFINITY;
}

void rtlThread::pipNotify (CkSynchro *synch, CkThread *requestor)

{
    if (getNode()->testOption(RTL_WARN_PRIO_INVERSION))
	CkManager::This->warning("priority inversion detected and fixed (synch=%s, owner=%s, requestor=%s)",
				 synch->protoGetName(),
				 protoGetName(),
				 requestor->protoGetName());
}

// pipHold() is specialized for the RTLinux thread in order to get
// priority ceiling as a variation of priority inheritance.

int rtlThread::pipHold (CkSynchro *synch, CkThread *requestor)

{
    // The current thread object holds the <synch> resource
    // which is now reclaimed by the <requestor> thread;
    // proceed to a priority boost if applicable.

    int boostPri;

    if (synch->getCkId().getType() == RTLINUX_MUTEX_INODE)
	boostPri = ((rtlMutex *)synch)->getPrioCeiling();
    else
	boostPri = requestor->getPri();

    if (getNode()->testPri(boostPri,getPri()) < 0)
	{
	if (!tstStatus(CKBOOST))
	    {
	    setStatus(CKBOOST);
	    CkThread::pipSetPri(getPri());
	    pipNotify(synch,requestor);
	    }

	addOwnership(synch);
	getNode()->boostThread(this,boostPri);
	return 1; // Indicates that caller needs to reschedule.
	}

    return 0;
}

void *rtlLinuxThread::khide(doLinux) ()

{
    rtlNode *node = getNode();
    int (*initModule)(void) = (int (*)(void))node->getRootEntry();

    int rc  = initModule();

    if (rc)
	CkManager::This->fatal("init_module() returned with code %d",rc);

    for (;;)
	{
	node->kdoor(pendSynchBase)(&flag,CK_INFINITE);
	flag.reset();

	if (node->tstStatus(RTL_FIFO_PEND_HANDLER))
	    {
	    node->clrStatus(RTL_FIFO_PEND_HANDLER);
	    rtlFifo::kcout(runHandlers)(node);
	    }
	}

    return NULL;
}

static void *linuxThread (void *arg)

{
    rtlLinuxThread *linuxThread = (rtlLinuxThread *)arg;
    return linuxThread->khide(doLinux)();
}

rtlLinuxThread::rtlLinuxThread (rtlNode *_node) :
    rtlThread(-1,
	      &linuxThread,
	      this,
	      RTLTHREAD_STACKSIZE,
	      new rtl_thread_struct,
	      "linux-thread",
	      _node)
{
    userHandle->handle = getCkId();
    userHandle->cleanup = NULL;
}

void rtlLinuxThread::dynamicConfigure (StProtoInfoMessage *mbuf) {
    // Prevent any re-configuration.
    CkManager::This->warning("you may not change linux thread properties - sorry");
}
