/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Specialization of the CarbonKernel trace manager for the
 * RTLinux extended API.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation
#endif // __GNUG__
#include <config.h>
#include <errno.h>
#include "CkManager.h"
#include "rtlNode.h"
#include "rtlProto.h"
#include "rtlTrace.h"

static const char rtlThreadFamily[] = "Thread";
static const char rtlSchedFamily[] = "Scheduler";
static const char rtlMutexFamily[] = "Mutex";
static const char rtlCvFamily[] = "CondVar";
static const char rtlClockFamily[] = "Clock";
static const char rtlFifoFamily[] = "Fifo";
static const char rtlIpcFamily[] = "IPC";
static const char rtlIrqFamily[] = "IRQ";
static const char rtlSema4Family[] = "Semaphores";

CkTraceDef rtlTraceInterface::traceDefTable[] = {

    { "pthread_create", rtlThreadFamily }, // 0
    { "pthread_exit", rtlThreadFamily },
    { "pthread_kill", rtlThreadFamily },
    { "pthread_setcancelstate", rtlThreadFamily },
    { "pthread_setcanceltype", rtlThreadFamily },
    { "pthread_cancel", rtlThreadFamily },
    { "pthread_testcancel", rtlThreadFamily },
    { "pthread_join", rtlThreadFamily },
    { "pthread_yield", rtlThreadFamily },
    { "pthread_self", rtlThreadFamily },
    { "pthread_setschedparam", rtlThreadFamily }, // 10
    { "pthread_getschedparam", rtlThreadFamily },
    { "pthread_attr_init", rtlThreadFamily },
    { "pthread_attr_getschedparam", rtlThreadFamily },
    { "pthread_attr_setschedparam", rtlThreadFamily },
    { "pthread_attr_getdetachstate", rtlThreadFamily },
    { "pthread_attr_setdetachstate", rtlThreadFamily },
    { "pthread_attr_setstacksize", rtlThreadFamily },
    { "pthread_attr_destroy", rtlThreadFamily },
    { "pthread_make_periodic_np", rtlThreadFamily },
    { "pthread_wait_np", rtlThreadFamily }, // 20
    { "pthread_suspend_np", rtlThreadFamily },
    { "pthread_wakeup_np", rtlThreadFamily },
    { "pthread_delete_np", rtlThreadFamily },
    { "pthread_setfp_np", rtlThreadFamily },
    { "pthread_attr_getcpu_np", rtlThreadFamily },
    { "pthread_attr_setcpu_np", rtlThreadFamily },
    { "pthread_mutexattr_init", rtlMutexFamily },
    { "pthread_mutexattr_settype", rtlMutexFamily },
    { "pthread_mutexattr_setpshared", rtlMutexFamily },
    { "pthread_mutexattr_gettype", rtlMutexFamily }, // 30
    { "pthread_mutexattr_getpshared", rtlMutexFamily },
    { "pthread_mutexattr_destroy", rtlMutexFamily },
    { "pthread_mutexattr_setprotocol", rtlMutexFamily },
    { "pthread_mutexattr_setprioceiling", rtlMutexFamily },
    { "pthread_mutexattr_getprotocol", rtlMutexFamily },
    { "pthread_mutexattr_getprioceiling", rtlMutexFamily },
    { "pthread_mutex_lock", rtlMutexFamily },
    { "pthread_mutex_trylock", rtlMutexFamily },
    { "pthread_mutex_unlock", rtlMutexFamily },
    { "pthread_mutex_init", rtlMutexFamily }, // 40
    { "pthread_mutex_destroy", rtlMutexFamily },
    { "pthread_mutex_setprioceiling", rtlMutexFamily },
    { "pthread_mutex_getprioceiling", rtlMutexFamily },
    { "pthread_condattr_init", rtlCvFamily },
    { "pthread_condattr_setpshared", rtlCvFamily },
    { "pthread_condattr_getpshared", rtlCvFamily },
    { "pthread_condattr_destroy", rtlCvFamily },
    { "pthread_cond_init", rtlCvFamily },
    { "pthread_cond_destroy", rtlCvFamily },
    { "pthread_cond_wait", rtlCvFamily }, // 50
    { "pthread_cond_broadcast", rtlCvFamily },
    { "pthread_cond_signal", rtlCvFamily },
    { "pthread_cond_timedwait", rtlCvFamily },
    { "sched_get_priority_min", rtlSchedFamily },
    { "sched_get_priority_max", rtlSchedFamily },
    { "rtl_schedule", rtlSchedFamily },
    { "rtl_sched_lock", rtlSchedFamily },
    { "rtl_getschedclock", rtlClockFamily },
    { "rtl_setclockmode", rtlClockFamily },
    { "rtl_setclockhandler", rtlClockFamily }, // 60
    { "rtl_unsetclockhandler", rtlClockFamily },
    { "rtl_getbestclock", rtlClockFamily },
    { "clock_gettime", rtlClockFamily },
    { "clock_settime", rtlClockFamily },
    { "clock_getres", rtlClockFamily },
    { "clock_gethrtime", rtlClockFamily },
    { "time", rtlClockFamily },
    { "usleep", rtlClockFamily },
    { "nanosleep", rtlClockFamily },
    { "rtl_delay", rtlClockFamily }, // 70
    { "gethrtime", rtlClockFamily },
    { "gethrtimeres", rtlClockFamily },
    { "rtf_create", rtlFifoFamily },
    { "rtf_destroy", rtlFifoFamily },
    { "rtf_put", rtlFifoFamily },
    { "rtf_get", rtlFifoFamily },
    { "rtf_create_handler", rtlFifoFamily },
    { "rtf_create_rt_handler", rtlFifoFamily },
    { "rt_sem_init", rtlIpcFamily },
    { "rt_sem_destroy", rtlIpcFamily },	// 80
    { "rt_sem_post", rtlIpcFamily },
    { "rt_sem_wait", rtlIpcFamily },
    { "rt_sem_trywait", rtlIpcFamily },
    { "rt_task_delay", rtlIpcFamily },
    { "rt_task_ipc_init", rtlIpcFamily },
    { "rt_mq_init", rtlIpcFamily },
    { "rt_mq_destroy", rtlIpcFamily },
    { "rt_mq_send", rtlIpcFamily },
    { "rt_mq_receive", rtlIpcFamily },
    { "rtl_hard_enable_irq", rtlIrqFamily }, // 90
    { "rtl_hard_disable_irq", rtlIrqFamily },
    { "rtl_request_global_irq", rtlIrqFamily },
    { "rtl_free_global_irq", rtlIrqFamily },
    { "rtl_get_soft_irq", rtlIrqFamily },
    { "rtl_free_soft_irq", rtlIrqFamily },
    { "rtl_global_pend_irq", rtlIrqFamily },
    { "rtl_no_interrupts", rtlIrqFamily },
    { "rtl_restore_interrupts", rtlIrqFamily },
    { "sem_init", rtlSema4Family },
    { "sem_destroy", rtlSema4Family }, // 100
    { "sem_post", rtlSema4Family },
    { "sem_timedwait", rtlSema4Family },
    { "sem_wait", rtlSema4Family },
    { "sem_trywait", rtlSema4Family },
    { "sem_getvalue", rtlSema4Family }
};

int rtlTraceInterface::traceDefCount =
sizeof(rtlTraceInterface::traceDefTable) /
sizeof(CkTraceDef);

CkErrorDef rtlTraceInterface::errorDefTable[] = {

    { "ENOERR", 0 },
    { "EAGAIN", EAGAIN },
    { "ETIME", ETIME },
    { "EINVAL", EINVAL },
    { "ESRCH", ESRCH },
    { "EBUSY", EBUSY },
    { "EDEADLK", EBUSY },
    { "ETIMEDOUT", ETIMEDOUT },
    { "EOPNOTSUPP", EOPNOTSUPP },
    { "EINTR", EINTR },

    { "EAGAIN", -EAGAIN },
    { "ETIME", -ETIME },
    { "EINVAL", -EINVAL },
    { "ENODEV", -ENODEV },
    { "EBUSY", -EBUSY },
    { "EBADF", -EBADF },
    { "", 0 }
};

int rtlTraceInterface::errorDefCount =
sizeof(rtlTraceInterface::errorDefTable) /
sizeof(CkErrorDef);

rtlTraceInterface::rtlTraceInterface (rtlNode *_node) :
    CkTraceInterface(_node->getOS(),
		     traceDefTable,
		     traceDefCount,
		     errorDefTable,
		     errorDefCount) {
}
