#  This file is part of the CarbonKernel Simulation System.
#  Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
# 
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
# 
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  Author(s): rpm
# 
#  Contributor(s):

proc rtlThread:attach {context name private} {

    global $context.info
    set $context.info(present) yes
    return [list [list {RTLinux kernel} Threads $name] thread]
}

proc rtlThread:detach {context} {

    global $context.info
    unset $context.info
    catch { destroy $context }
}

proc rtlThread:show {context nid name} {

    if {[winfo exists $context]} {
	wm deiconify $context
	raise $context
    }

    rtlThread:pollInfo $context
}

proc rtlThread:event {context event} {

    switch $event {
	SimulationHeldEvent {
	    if {[catch { wm state $context } state] == 0} {
		if {$state != "withdrawn"} {
		    rtlThread:pollInfo $context
		}
	    }
	}
    }
}

proc rtlThread:update {context nid name info} {

    global $context.info

    set $context.info(wchan) n/a
    set $context.info(timeout) n/a

    # dispatch the information block contents
    # to the per-context data array...
    foreach e $info {
	set item [lindex $e 0]
	set value [lindex $e 1]
	if {$item == "wchan"} {
	    # special case for wchan - join pieces
	    set value [join $value]
	}
	set $context.info($item) $value
    }

    set $context.info(current) $info

    if {[winfo exists $context]} {
	# update signal information by hand -
	# other display updates will be triggered automatically
	# by changing cells from the data array.
	set sigfrm [$context.sw2 subwidget frame]

	foreach {signame sigstate} [set $context.info(signals)] {
	    if {$sigstate == 1} {
		$sigfrm.sig$signame select
	    } {
		$sigfrm.sig$signame deselect
	    }
	}
	return
    }

    # if the thread object has never been displayed before,
    # build a brand new toplevel for it...

    toplevel $context
    wm title $context "RTLinux thread: $name"
    wm resizable $context 0 0
    cascadeWindow $context
    wm protocol $context WM_DELETE_WINDOW \
	"wm withdraw $context"

    set sw1 $context.sw1
    set sw2 $context.sw2

    tixLabelFrame $sw1 -label Status \
	-labelside acrosstop -options {
	    label.padx 5
	}

    tixLabelFrame $sw2 -label Signals \
	-labelside acrosstop -options {
	    label.padx 5
	}

    pack $sw1 $sw2 -side top -fill both -expand yes

    # Status frame

    set statfrm [$sw1 subwidget frame]

    # State
    tixLabelEntry $statfrm.state -label State: -state disabled
    $statfrm.state subwidget entry configure \
	-textvariable $context.info(state)
    pack $statfrm.state -anchor e -pady 5 -padx 5

    # Timeout
    tixLabelEntry $statfrm.timeout -label Timeout: -state disabled
    $statfrm.timeout subwidget entry configure \
	-textvariable $context.info(timeout)
    pack $statfrm.timeout -anchor e -pady 5 -padx 5

    # Wait channel
    tixLabelEntry $statfrm.wchan -label Wchan: -state disabled
    $statfrm.wchan subwidget entry configure \
	-textvariable $context.info(wchan)
    pack $statfrm.wchan -anchor e -pady 5 -padx 5

    # Interrupt masking level
    tixControl $statfrm.ilevel \
	-label I-Mask: \
	-variable $context.info(ilevel) \
	-integer true \
	-min 0 \
	-max 32 \
	-allowempty false

    pack $statfrm.ilevel -anchor e -pady 5 -padx 5

    # Priority
    tixControl $statfrm.prio \
	-label Priority: \
	-variable $context.info(prio) \
	-integer true \
	-min -1 \
	-max 32 \
	-allowempty false

    pack $statfrm.prio -anchor e -pady 5 -padx 5

    # Signal frame

    set sigfrm [$sw2 subwidget frame]
    set i 0

    foreach {signame sigstate} [set $context.info(signals)] {
	checkbutton $sigfrm.sig$signame -text $signame \
	    -relief flat -anchor w \
	    -variable $context.info(sigstate,$signame)
	if {$sigstate == 1} {
	    $sigfrm.sig$signame select
	} {
	    $sigfrm.sig$signame deselect
	}
	set col [expr $i % 4]
	set row [expr $i / 4]
	grid $sigfrm.sig$signame -column $col -row $row
	incr i
    }

    tixButtonBox $context.bbox -orientation horizontal -relief flat -bd 0
    $context.bbox add update -text Update \
	-command "rtlThread:pollInfo $context"
    $context.bbox add apply -text Apply \
	-command "rtlThread:applyChanges $context"
    $context.bbox add cancel -text Close \
	-command "wm withdraw $context"
    pack $context.bbox -side bottom -fill x
}

proc rtlThread:pollInfo {context} {
    TkRequest $context GetObjectInfo all
}

proc rtlThread:applyChanges {context} {

    global $context.info

    set oldinfo [set $context.info(current)]
    set newprio [list prio [set $context.info(prio)]]
    set newilvl [list ilevel [set $context.info(ilevel)]]
    set sigstates {}

    foreach signame {DELETE WAKEUP SUSPEND} {
	lappend sigstates [list $signame [set $context.info(sigstate,$signame)]]
    }

    set newsigs [list signals $sigstates]

    # Spurious "round-robin" side-effect on thread scheduling is
    # prevented by never reinstateing an unmodified priority level.

    foreach config {newsigs newilvl newprio} {
	if {[lsearch -exact $oldinfo [set $config]] == -1} {
	    TkRequest $context ConfigureObject [set $config]
	}
    }
}

proc rtlThread:fillWaitList {context listbox} {

    global $context.info

    set lbox [$listbox subwidget listbox]
    $lbox delete 0 end

    foreach threadinfo [set $context.info(waiters)] {
	set name [lindex $threadinfo 2]
	$lbox insert end $name
    }
}

proc rtlThread:pickWaitList {context listbox} {

    global $context.info

    set lbox [$listbox subwidget listbox]
    set sel [$lbox curselection]

    if {$sel != {}} {
	set threadinfo [lindex [set $context.info(waiters)] $sel]
	set nid [lindex $threadinfo 0]
	set name [lindex $threadinfo 2]
	TkRequest $context InspectObject \
	    [list $nid {RTLinux kernel} Threads $name]
    }
}
