/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Testing RTL services:
 * - pthread_mutexattr_settype()
 * - pthread_mutexattr_setpshared()
 * - pthread_mutexattr_gettype()
 * - pthread_mutexattr_getpshared()
 * - pthread_mutexattr_init()
 * - pthread_mutexattr_destroy()
 * - pthread_mutex_lock()
 * - pthread_mutex_unlock()
 * - pthread_mutex_trylock()
 * - pthread_mutex_init()
 * - pthread_mutex_destroy()
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#include <ck/testlib.h>
#include <rtl/rtl.h>

static pthread_t pth1, pth2;

ckhandle_t handle1, handle2;

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

pthread_mutex_t umutex = PTHREAD_MUTEX_INITIALIZER;

void *test_thread1 (void *data)

{
    int n;

    /* Error cases: uninit mutex */
    TEST_ASSERT(pthread_mutex_lock(&umutex) == EINVAL);
    TEST_ASSERT(pthread_mutex_unlock(&umutex) == EINVAL);
    TEST_ASSERT(pthread_mutex_trylock(&umutex) == EINVAL);

    TEST_ASSERT(pthread_mutex_trylock(&mutex) == EBUSY);

    for (n = 0; n < 3; n++)
	{
	TEST_MARK();
	/* Lock already granted to init (lock()) or thread2 (last unlock()): should block. */
	TEST_ASSERT_OK(pthread_mutex_lock(&mutex));
	pthread_yield();
	TEST_MARK();
	TEST_ASSERT_OK(pthread_mutex_unlock(&mutex));
	}

    TEST_MARK();

    TEST_ASSERT_OK(pthread_mutex_destroy(&mutex));
    TEST_ASSERT(pthread_mutex_destroy(&umutex) == EINVAL); /* Error case: uninit mutex */

    TEST_CHECK_SEQUENCE(SEQ(handle1,1),
			SEQ(handle2,2),
			SEQ(handle1,2),
			SEQ(handle2,2),
			SEQ(handle1,2),
			SEQ(handle2,2),
			SEQ(handle1,2),
			END_SEQ);

    TEST_FINISH();

    return NULL;		/* Not executed */
}

void *test_thread2 (void *data)

{
    for (;;)
	{
	TEST_MARK();
	TEST_ASSERT_OK(pthread_mutex_unlock(&mutex));
	TEST_MARK();
	/* Lock already granted to thread1 by last unlock(): should block */
	TEST_ASSERT_OK(pthread_mutex_lock(&mutex));
	pthread_yield();
	}

    return NULL;
}

int init_module ()

{
    struct rtl_sched_param param1, param2;
    pthread_attr_t attr1, attr2;
    pthread_mutexattr_t mattr;
    int mflag;

    rtl_sched_lock(pthread_self());

    TEST_START(0);

    param1.sched_priority = 3;
    TEST_ASSERT_OK(pthread_attr_init(&attr1));
    TEST_ASSERT_OK(pthread_attr_setstacksize(&attr1,8192));
    TEST_ASSERT_OK(pthread_attr_setschedparam(&attr1,&param1));

    TEST_ASSERT_OK(pthread_create(&pth1,
				  &attr1,
				  test_thread1,
				  NULL));
    handle1 = pth1->handle;

    param2.sched_priority = 3;
    TEST_ASSERT_OK(pthread_attr_init(&attr2));
    TEST_ASSERT_OK(pthread_attr_setstacksize(&attr2,8192));
    TEST_ASSERT_OK(pthread_attr_setschedparam(&attr2,&param2));

    TEST_ASSERT_OK(pthread_create(&pth2,
				  &attr2,
				  test_thread2,
				  NULL));
    handle2 = pth2->handle;


    TEST_ASSERT_OK(pthread_mutexattr_init(&mattr));
    TEST_ASSERT(mattr.type == PTHREAD_MUTEX_DEFAULT);
    TEST_ASSERT(mattr.pshared == PTHREAD_PROCESS_SHARED);

    TEST_ASSERT_OK(pthread_mutexattr_settype(&mattr,PTHREAD_MUTEX_DEFAULT));
    TEST_ASSERT(pthread_mutexattr_settype(&mattr,-1) == EINVAL);

    TEST_ASSERT_OK(pthread_mutexattr_gettype(&mattr,&mflag));
    TEST_ASSERT(mflag == PTHREAD_MUTEX_DEFAULT);

    TEST_ASSERT_OK(pthread_mutexattr_setpshared(&mattr,PTHREAD_PROCESS_SHARED));
    TEST_ASSERT_OK(pthread_mutexattr_setpshared(&mattr,PTHREAD_PROCESS_PRIVATE));
    TEST_ASSERT(pthread_mutexattr_setpshared(&mattr,-1) == EINVAL);

    TEST_ASSERT_OK(pthread_mutexattr_getpshared(&mattr,&mflag));
    TEST_ASSERT(mflag == PTHREAD_PROCESS_PRIVATE);

    TEST_ASSERT_OK(pthread_mutex_init(&mutex,&mattr));
    TEST_ASSERT_OK(pthread_mutex_trylock(&mutex));

    TEST_ASSERT_OK(pthread_mutexattr_destroy(&mattr));

    rtl_sched_lock(NULL);

    return 0;
}
