/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Testing RTL services:
 * - pthread_cond_init()
 * - pthread_cond_destroy()
 * - pthread_cond_signal()
 * - pthread_cond_broadcast()
 * - pthread_cond_wait()
 * - pthread_cond_timedwait()
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#include <ck/testlib.h>
#include <rtl/rtl.h>

static pthread_t pth1, pth2, pth3;

ckhandle_t handle1, handle2, handle3;

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

pthread_mutex_t umutex = PTHREAD_MUTEX_INITIALIZER;

pthread_cond_t cond = PTHREAD_COND_INITIALIZER;

pthread_cond_t ucond = PTHREAD_COND_INITIALIZER;

static int event;

void *test_thread1 (void *data)

{
    struct timespec ts = { 0, 0 };
    int n;

    /* Error cases: uninit condition variable and/or mutex */
    TEST_ASSERT(pthread_cond_signal(&ucond) == EINVAL);
    TEST_ASSERT(pthread_cond_broadcast(&ucond) == EINVAL);
    TEST_ASSERT(pthread_cond_wait(&ucond,&mutex) == EINVAL);
    TEST_ASSERT(pthread_cond_wait(&cond,&umutex) == EINVAL);
    TEST_ASSERT(pthread_cond_wait(&ucond,&umutex) == EINVAL);
    TEST_ASSERT(pthread_cond_timedwait(&ucond,&mutex,&ts) == EINVAL);
    TEST_ASSERT(pthread_cond_timedwait(&cond,&umutex,&ts) == EINVAL);
    TEST_ASSERT(pthread_cond_timedwait(&ucond,&umutex,&ts) == EINVAL);
    TEST_ASSERT(pthread_cond_destroy(&ucond) == EINVAL);

    TEST_MARK();

    TEST_ASSERT_OK(pthread_mutex_lock(&mutex));

    while (!event)
	{
	TEST_MARK();
	TEST_ASSERT_OK(pthread_cond_wait(&cond,&mutex));
	}

    TEST_ASSERT_OK(pthread_mutex_unlock(&mutex));

    TEST_MARK();

    pthread_yield();

    TEST_MARK();

    TEST_ASSERT_OK(pthread_mutex_lock(&mutex));

    event = 0;

    while (!event)
	{
	TEST_MARK();
	ts = timespec_from_ns(gethrtime() + 1000000);
	TEST_ASSERT_OK(pthread_cond_timedwait(&cond,&mutex,&ts));
	}

    ts = timespec_from_ns(gethrtime() + 1000000);
    TEST_ASSERT(pthread_cond_timedwait(&cond,&mutex,&ts) == ETIMEDOUT);

    TEST_MARK();

    TEST_ASSERT_OK(pthread_mutex_unlock(&mutex));

    TEST_MARK();

    TEST_CHECK_SEQUENCE(SEQ(handle1,2),
			SEQ(handle2,2),
			SEQ(handle3,2),
			SEQ(handle1,1),
			SEQ(handle2,1),
			SEQ(handle1,2),
			SEQ(handle3,2),
			SEQ(handle1,2),
			END_SEQ);

    TEST_ASSERT_OK(pthread_cond_destroy(&cond));

    TEST_ASSERT(!ckObjectExists(cond.handle));

    TEST_ASSERT_OK(pthread_mutex_destroy(&mutex));

    TEST_ASSERT(!ckObjectExists(mutex.handle));

    TEST_FINISH();

    return NULL;		/* Not executed */
}

void *test_thread2 (void *data)

{
    TEST_MARK();

    TEST_ASSERT_OK(pthread_mutex_lock(&mutex));

    while (!event)
	{
	TEST_MARK();
	TEST_ASSERT_OK(pthread_cond_wait(&cond,&mutex));
	}

    TEST_ASSERT_OK(pthread_mutex_unlock(&mutex));

    TEST_MARK();

    pthread_exit(NULL);

    TEST_MARK();		/* Should not pass this mark */

    return NULL;
}

void *test_thread3 (void *data)

{
    TEST_MARK();

    TEST_ASSERT_OK(pthread_mutex_lock(&mutex));

    event = 1;

    TEST_ASSERT_OK(pthread_cond_broadcast(&cond));

    TEST_MARK();

    TEST_ASSERT_OK(pthread_mutex_unlock(&mutex));

    pthread_yield();		/* Force rescheduling */
    
    TEST_MARK();

    TEST_ASSERT_OK(pthread_mutex_lock(&mutex));

    event = 1;

    TEST_ASSERT_OK(pthread_cond_signal(&cond));

    TEST_MARK();

    TEST_ASSERT_OK(pthread_mutex_unlock(&mutex));

    pthread_exit(NULL);

    TEST_MARK();		/* Should not pass this mark */
}

int init_module ()

{
    struct rtl_sched_param param1, param2, param3;
    pthread_attr_t attr1, attr2, attr3;
    pthread_mutexattr_t mattr;

    rtl_sched_lock(pthread_self());

    TEST_START(0);

    param1.sched_priority = 2;
    TEST_ASSERT_OK(pthread_attr_init(&attr1));
    TEST_ASSERT_OK(pthread_attr_setstacksize(&attr1,8192));
    TEST_ASSERT_OK(pthread_attr_setschedparam(&attr1,&param1));

    TEST_ASSERT_OK(pthread_create(&pth1,
				  &attr1,
				  test_thread1,
				  NULL));
    handle1 = pth1->handle;

    param2.sched_priority = 2;
    TEST_ASSERT_OK(pthread_attr_init(&attr2));
    TEST_ASSERT_OK(pthread_attr_setstacksize(&attr2,8192));
    TEST_ASSERT_OK(pthread_attr_setschedparam(&attr2,&param2));

    TEST_ASSERT_OK(pthread_create(&pth2,
				  &attr2,
				  test_thread2,
				  NULL));
    handle2 = pth2->handle;

    param3.sched_priority = 1;
    TEST_ASSERT_OK(pthread_attr_init(&attr3));
    TEST_ASSERT_OK(pthread_attr_setstacksize(&attr3,8192));
    TEST_ASSERT_OK(pthread_attr_setschedparam(&attr3,&param3));

    TEST_ASSERT_OK(pthread_create(&pth3,
				  &attr3,
				  test_thread3,
				  NULL));
    handle3 = pth3->handle;

    TEST_ASSERT_OK(pthread_mutexattr_init(&mattr));
    TEST_ASSERT_OK(pthread_mutex_init(&mutex,&mattr));
    TEST_ASSERT_OK(pthread_cond_init(&cond,NULL));

    rtl_sched_lock(NULL);

    return 0;
}
