/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Testing RTL services:
 * - rtl_getschedclock()
 * - rtl_setclockmode()
 * - rtl_setclockhandler()
 * - rtl_unsetclockhandler()
 * - rtl_getbestclock()
 * - rtl_delay()
 * - clock_gettime()
 * - clock_settime()
 * - clock_getres()    [Depends on RtlTest configuration values]
 * - clock_gethrtime()
 * - time()
 * - gethrtime()
 * - gethrtimeres()    [Depends on RtlTest configuration values]
 * - usleep()
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#include <ck/testlib.h>
#include <rtl/rtl.h>

static pthread_t pth1;

static struct rtl_clock badclock, *uclockid = &badclock;

ckhandle_t handle1;

int oneshot_count = 0;

int periodic_count = 0;

void one_shot_handler (struct pt_regs *r)

{
    hrtime_t now;

    CK_TIME_LOCK();

    now = gethrtime();

    if (oneshot_count++ == 0)
	TEST_ASSERT(now == 234567)
    else
	TEST_ERROR("unexpected call of one_shot_handler()");

    CK_TIME_UNLOCK();
}

void periodic_handler (struct pt_regs *r)

{
    static hrtime_t base;
    static int n;
    hrtime_t now;

    CK_TIME_LOCK();

    now = gethrtime();

    if (!periodic_count)
	base = now;

    TEST_ASSERT(now == periodic_count * 777 + base);
    periodic_count++;

    CK_TIME_UNLOCK();
}

void *test_thread1 (void *data)

{
    struct timespec ts;
    clockid_t clockid;
    hrtime_t t0, t1;
    time_t secs;

    t0 = gethrtime();
    usleep(1000);
    t1 = gethrtime();

    CK_TIME_LOCK();

    clockid = rtl_getbestclock(0);
    TEST_ASSERT(clockid == rtl_getschedclock());
    TEST_ASSERT(gethrtime() == clock_gethrtime(clockid)); /* True under Ck */
    TEST_ASSERT(clock_gettime(uclockid,&ts) == EINVAL);
    TEST_ASSERT_OK(clock_gettime(clockid,&ts));
    TEST_ASSERT(ts.tv_sec * 1000000000 + ts.tv_nsec == clock_gethrtime(clockid));
    TEST_ASSERT_OK(clock_getres(clockid,&ts));
    TEST_ASSERT(ts.tv_sec == 0 && ts.tv_nsec == 838);
    TEST_ASSERT(gethrtimeres() == 838);
    TEST_ASSERT(time(NULL) == ts.tv_sec);
    TEST_ASSERT(time(&secs) == ts.tv_sec);
    TEST_ASSERT(secs == ts.tv_sec);
    ts.tv_sec = 2000;
    ts.tv_nsec = 8380;
    TEST_ASSERT_OK(clock_settime(clockid,&ts));
    ts.tv_sec = 0;
    ts.tv_nsec = 0;
    TEST_ASSERT_OK(clock_gettime(clockid,&ts));
    TEST_ASSERT(ts.tv_sec == 2000 && ts.tv_nsec == 8380);
    ts.tv_sec = 0;
    ts.tv_nsec = 0;
    TEST_ASSERT_OK(clock_settime(clockid,&ts));

    CK_TIME_UNLOCK();

    TEST_ASSERT_OK(rtl_setclockmode(clockid,RTL_CLOCK_MODE_ONESHOT,0));
    TEST_ASSERT_OK(rtl_setclockhandler(clockid,one_shot_handler));
    TEST_ASSERT_OK(clockid->settimer(clockid,234567));
    rtl_delay(300000);

    TEST_ASSERT(rtl_setclockhandler(clockid,one_shot_handler) == -EBUSY);
    TEST_ASSERT_OK(rtl_unsetclockhandler(clockid));

    TEST_ASSERT_OK(rtl_setclockhandler(clockid,periodic_handler));
    TEST_ASSERT_OK(rtl_setclockmode(clockid,RTL_CLOCK_MODE_PERIODIC,777));
    rtl_delay(7770);

    TEST_ASSERT(oneshot_count == 1);
    TEST_ASSERT(periodic_count == 10);

    TEST_FINISH();

    return NULL;		/* Not executed */
}

int init_module ()

{
    struct rtl_sched_param param1;
    pthread_attr_t attr1;

    rtl_sched_lock(pthread_self());

    TEST_START(0);

    param1.sched_priority = 2;
    TEST_ASSERT_OK(pthread_attr_init(&attr1));
    TEST_ASSERT_OK(pthread_attr_setstacksize(&attr1,16384));
    TEST_ASSERT_OK(pthread_attr_setschedparam(&attr1,&param1));

    TEST_ASSERT_OK(pthread_create(&pth1,
				  &attr1,
				  test_thread1,
				  NULL));
    handle1 = pth1->handle;

    rtl_sched_lock(NULL);

    return 0;
}
