/*
 * This file is part of the CarbonKernel Simulation System.
 * Copyright (C) 1997-2000 Realiant Systems.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * The original code is CarbonKernel - Real-time Operating System Simulator,
 * released April 15, 2000. The initial developer of the original code is
 * Realiant Systems (http://www.realiant.com).
 *
 * Description: Testing RTL/Message Queue services:
 * - rt_mq_init
 * - rt_mq_destroy
 * - rt_mq_send
 * - rt_mq_receive
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#include <ck/testlib.h>
#include <rtl/rtl.h>

static pthread_t pth1, pth2;

ckhandle_t handle1, handle2;

static rt_mq_t umq;

static rt_mq_t mq;

void *test_thread1 (void *data)

{
    int msg;

    TEST_ASSERT_OK(pthread_suspend_np(pth2));

    msg = 1;
    TEST_ASSERT_OK(rt_mq_send(&mq,(char *)&msg,RT_MQ_NORMAL,RT_WAIT_FOREVER));
    TEST_MARK();

    msg = 2;
    TEST_ASSERT_OK(rt_mq_send(&mq,(char *)&msg,RT_MQ_URGENT,RT_WAIT_FOREVER));
    TEST_MARK();

    TEST_ASSERT(rt_mq_send(&mq,(char *)&msg,RT_MQ_URGENT,RT_NO_WAIT) == -EAGAIN);
    TEST_MARK();

    TEST_ASSERT(rt_mq_send(&mq,(char *)&msg,RT_MQ_URGENT,RELATIVE_TIME(1000)) == -ETIME);
    TEST_MARK();

    pthread_wakeup_np(pth2);
    TEST_MARK();

    pthread_yield();
    TEST_MARK();

    usleep(1010);
    TEST_MARK();

    TEST_CHECK_SEQUENCE(SEQ(handle1,5),
			SEQ(handle2,4),
			SEQ(handle1,1),
			SEQ(handle2,1),
			SEQ(handle1,1),
			END_SEQ);

    TEST_ASSERT_OK(rt_mq_destroy(&mq));

    TEST_ASSERT(!ckObjectExists(mq.handle));

    TEST_FINISH();

    return NULL;		/* Not executed */
}

void *test_thread2 (void *data)

{
    int msg = -1;

    TEST_MARK();

    TEST_ASSERT_OK(rt_mq_receive(&mq,(char *)&msg,RT_WAIT_FOREVER));
    TEST_ASSERT(msg == 2);
    TEST_MARK();

    TEST_ASSERT_OK(rt_mq_receive(&mq,(char *)&msg,RT_WAIT_FOREVER));
    TEST_ASSERT(msg == 1);
    TEST_MARK();

    TEST_ASSERT(rt_mq_receive(&mq,(char *)&msg,RT_NO_WAIT) == -EAGAIN);
    TEST_MARK();

    TEST_ASSERT(rt_mq_receive(&mq,(char *)&msg,RELATIVE_TIME(1000)) == -ETIME);
    TEST_MARK();

    return NULL;
}

int init_module ()

{
    struct rtl_sched_param param1, param2;
    pthread_attr_t attr1, attr2;

    rtl_sched_lock(pthread_self());

    TEST_START(0);

    param1.sched_priority = 3;
    TEST_ASSERT_OK(pthread_attr_init(&attr1));
    TEST_ASSERT_OK(pthread_attr_setstacksize(&attr1,8192));
    TEST_ASSERT_OK(pthread_attr_setschedparam(&attr1,&param1));

    TEST_ASSERT_OK(pthread_create(&pth1,
				  &attr1,
				  test_thread1,
				  NULL));
    handle1 = pth1->handle;

    param2.sched_priority = 3;
    TEST_ASSERT_OK(pthread_attr_init(&attr2));
    TEST_ASSERT_OK(pthread_attr_setstacksize(&attr2,8192));
    TEST_ASSERT_OK(pthread_attr_setschedparam(&attr2,&param2));

    TEST_ASSERT_OK(pthread_create(&pth2,
				  &attr2,
				  test_thread2,
				  NULL));
    handle2 = pth2->handle;

    TEST_ASSERT(rt_mq_init(&mq,0,1) == -EINVAL);
    TEST_ASSERT(rt_mq_init(&mq,1,-1) == -EINVAL);
    TEST_ASSERT_OK(rt_mq_init(&mq,2,sizeof(int)));

    rtl_sched_lock(NULL);

    return 0;
}
